#!/usr/bin/perl

#est_pairwise_and_cement.pl

# This script performs all four steps of the orbits code, and displays
# the results of cement at each iteration.
# For usage: "./allstages.pl"
# PLM NOTE: estpchirp2m usually works best on PLMs with an exponent of 2.2,
#           while cement works on an exponent of 4.7. To use different
#           images for estpchirp2m, enter the prefix of the files to be 
#           used for estpchirp2m. (see usage)



use IO::Handle;

`rm prev_input*`;
`rm pork_params*`;
`rm newparam`;
`rm tmpparams`;
`rm dechirped*`;
`rm Parameters_pairwise`;
`rm parameters_pairwise`;

`touch parameters_pairwise`;

# number of steps is specified here, e.g. $steps = sprintf("4 8 7 6 5");
# does 8 steps at lowest resolution, 7 next, then 6, and finally 5 at max res

# $steps = sprintf("4 3 3 3 3"); # does 3 repetitions at each of 4 levels
# $steps = sprintf("4 4 3 2 1"); # does 4 then 3 then 2, then 1 (1 means none)

# here are some examples of what i've done in the past for eyetap video;
# just uncomment the one you want to try:
# $steps = sprintf("5 8 4 4 4 4"); #usually better because you more at low res
#$steps = sprintf("4 32 16 8 2");
#$steps = sprintf("4 4 4 2 1");
$steps = sprintf("3 32 8 2");
#$steps = sprintf("2 4 2");


$prefix="v";  # path as well as part of file before number
$is_jpg = 0;

$firstframe=$ARGV[0];
$lastframe=$ARGV[1];
$corner=$ARGV[2];
$filetype=$ARGV[3];

if(@ARGV == 5){
    $estprefix=$ARGV[4];
} else {
    $estprefix=$prefix;
}

if ($filetype =~ m/jpg/) {
   $is_jpg = 1;
   $postfix = ".pnm";  # we don're really know if these are col or gray
} elsif ($filetype =~ m/plm/) {
   $postfix = ".plm";
} elsif ($filetype =~ m/pgm/) {                                                
   $postfix = ".pgm"; 
} elsif ($filetype =~ m/pdm/) {                                                
   $postfix = ".pdm"; 
} elsif ($filetype =~ m/ppm/) {                                                
   $postfix = ".ppm";                                                       
} else {
  die "\nUsage: perl allstages.pl  START_RANGE  STOP_RANGE  CORNER  FILE_TYPE  EST_PREFIX\nSTART_RANGE: number denoting the first image to be processed. (eg. 1 for v0001.pgm)\nSTOP_RANGE: number denoting the last image to be processed.\nCORNER: corner in which to display image, numbered 1-4 starting with top left, -1 for auto.\nFILE_TYPE: one of jpg, plm, pdm, ppm, or pgm.\nEST_PREFIX: prefix of files to use for estpchirp2m. Default: same as used for cement.\n\n";  
}

# Dechirp the first image to get a file for cementing.
$counter = $firstframe;
printf("\n");

if ($is_jpg == 1) {
    $inputfile = sprintf("$prefix%04d.jpg", $counter);
    `djpeg -grayscale $inputfile > prev_input$postfix`;

    # jpegs are never turned into plms, so estprefix should always be the same as prefix
    `cp prev_input$postfix prev_est_input$postfix`;

} else {
    $inputfile = sprintf("$prefix%04d$postfix", $counter);
    $estinputfile = sprintf("$estprefix%04d$postfix", $counter);
    `cp $inputfile prev_input$postfix`;
    `cp $estinputfile prev_est_input$postfix`;
}

# We use pchirp2nocrop rather than cp to get a well-formed header for cement.
`pchirp2nocrop prev_input$postfix dechirped$counter$postfix 1.0000 0.0000 0.00000 0.0000 1.0000 0.0000 0.0000 0.0000`;


# The rest of the images get the full treatment.
for ($counter=$firstframe+1; $counter<=$lastframe; $counter++) {

   if ($is_jpg == 1) {
       $inputfile = sprintf("$prefix%04d.jpg", $counter);
       `djpeg -grayscale $inputfile > cur_input$postfix`;
       `cp cur_input$postfix cur_est_input$postfix`;
   } else {
       $inputfile = sprintf("$prefix%04d$postfix", $counter);
       $estinputfile = sprintf("$estprefix%04d$postfix", $counter);
       `cp $inputfile cur_input$postfix`;
       `cp $estinputfile cur_est_input$postfix`;
   }

   `estpchirp2m prev_est_input$postfix cur_est_input$postfix -steps $steps -outp pork_params$counter.txt`;

   # append the last line of the output from estpchir2p to allparams
   `tail -n 1 pork_params$counter.txt > newparam`;
   `cat parameters_pairwise newparam > tmpparams`;
   `cp tmpparams parameters_pairwise`;

   # integrate and cement the images
   #$reference = $counter-2;
   #print "reference: $reference\n";
   `pintegrate parameters_pairwise 0 > Parameters_pairwise`;
   $paramline = `tail -n 1 Parameters_pairwise`;
   printf("\ncounter: $counter paramline: $paramline\n");

   `pchirp2nocrop cur_input$postfix dechirped$counter$postfix $paramline`; 
   
   `cement cement_out$postfix dechirped*$postfix`;

   # Bring up a window to display the final cemented image.
   if ($counter == ($firstframe+1)) {

      # Figure out which corner of the screen the image should be displayed in
      # Offset is in line 2 and 3, so this should be quick
      $offsetfile = "dechirped$counter$postfix";
      open(OFFSETFILE, "< $offsetfile");
      <OFFSETFILE>; # First line, throw it away.
      $Moffset = <OFFSETFILE>; # M offset.     
      $Noffset = <OFFSETFILE>; # N offset.      

      $Moffset =~ s/[^\-0-9]//g;
      $Noffset =~ s/[^\-0-9]//g; 

      if ($corner == -1) {

          # Image is in the top left hand corner.
          if($Moffset >= 0 && $Noffset >= 0)    { $corner = 1; }

          # Image is in the top right hand corner.
          elsif ($Moffset >= 0 && $Noffset < 0) { $corner = 2; }

          # Image is in the bottom left hand corner.
          elsif ($Moffset < 0 && $Noffset >= 0) { $corner = 3; }

          # Image is in the bottom right hand corner.
          elsif ($Moffset < 0 && $Noffset < 0)  { $corner = 4; }
      }
    
      open(DISP, "| cement_display");
   }

   # Display the result of this iteration.
   print DISP "cement_out$postfix $corner";
   print DISP "\r";
   DISP->autoflush(1);

   # This iteration's first image becomes the next iteration's second image.
   `mv cur_input$postfix prev_input$postfix`;
   `mv cur_est_input$postfix prev_est_input$postfix`;
}

close(DISP);










