#!/usr/bin/perl -w
# use -w for useful debugging output scripts
##based on pchirp2nocrop.pl, but interpolates frames between existing frames


################
# USER OPTIONS

# set to 1 to use the existing parameters_pairwise.txt as interpolator
# or set to 0 so repeat frames
$USE_INTERPOLATION = 1; 

# number of new frames
$NUM_INTRA_FRAMES = 2;

###############

$INPUT_FILE="parameters_pairwise.txt";
$output_is_png = FALSE;
$argv_start_index = 0; # initialize counter as if no -png given in ARGV
$MSE_THRESH = 300;

if ($#ARGV < 0) {
  printf("Use: pinterpolate.pl v???.jpg # or use gash\n");
  printf("     pinterpolate.pl also autodetects jpg or ppm\n");
  printf("     pinterpolate.pl -png v???.jpg # outputs png with NaN alpha\n");
  die "     ERROR: YOU MUST SPECIFY SOME FILENAMES TO BE DECHIRPED$!\n";
}

if ($ARGV[1] eq "-png") {
  $output_is_png = TRUE;
  $argv_start_index = 1; # because index 0 is the -png input
}

@required_files = ("parameters_pairwise.txt");
#check to see if the required files exist
foreach $file (@required_files) {
   if (!-e $file) {
        printf("Missing file %s!\n\n",$file);
        exit -1;
   }
}


for ($counter=$argv_start_index; $counter<=$#ARGV+1-2; $counter++) {
  # $#ARGV is -1 if no arguments given; is 0 if 1 argument given, etc...
  $filea = $ARGV[$counter];
  $counterplusone = $counter+1;
  $fileb = $ARGV[$counterplusone];

  $periodpositiona = rindex($filea,"."); # right index, period from right
  $periodpositionb = rindex($fileb,".");
  $postfixa = substr($filea,$periodpositiona+1,3); # 2 char, e.g. JP or jp
  $postfixa =~ tr/[A-Z]/[a-z]/;
  #print $postfixa;
  $postfixb = substr($fileb,$periodpositionb+1,3);
  $postfixb =~ tr/[A-Z]/[a-z]/;
  #print $postfixb;
  #if it's JPEG, jpeg, JPG, or jpg:
    if (($postfixa eq "jpg") || ($postfixa eq "jpe")) {
      #printf("about to djpeg $filea\n");

      ##extract filesize 
      $sizestr = `rdjpgcom -verbose $filea`;
      ##print $sizestr;
      $sizestr =~ m/(JPEG image is )(\d*)(w \* )(\d*)/;
      $window_horiz_size = $2;
      $window_vert_size = $4;
      $ident_horiz_size = $window_horiz_size;
      $ident_vert_size  = $window_vert_size;

      printf("%d %d\n", $window_horiz_size, $window_vert_size);
      `djpeg $filea > deleteme_deletemea.pnm`;
      #$filea = "deleteme_deletemea.pnm";
    }
    else {
      `cp $filea deleteme_deletemea.pnm`;
      $sizestr = `pnmfile $filea`;
      print $sizestr;
      $sizestr =~ m/( raw, )(\d*)( by )(\d*)/;
      $window_horiz_size = $2;
      $window_vert_size = $4;
      $ident_horiz_size = $window_horiz_size;
      $ident_vert_size  = $window_vert_size;
      printf("%d %d\n", $window_horiz_size, $window_vert_size);
    }
  #if it's JPEG, jpeg, JPG, or jpg:
    if (($postfixb eq "jpg") || ($postfixb eq "jpe")) {
      #printf("about to djpeg $fileb\n");
      `djpeg $fileb > deleteme_deletemeb.pnm`;
      #$fileb = "deleteme_deletemeb.pnm";
    }
    else {
      `cp $fileb deleteme_deletemeb.pnm`;
    }
  $tmpcount = $counter+1;
  ##$tmpcount2= $counter+2;
  ##chop($pchirp_first=`head -$tmpcount $INPUT_FILE | tail -1`);
  ##chop($pchirp_second=`head -$tmpcount2 $INPUT_FILE | tail -1`);
  chop($pchirp_second=`head -$tmpcount $INPUT_FILE | tail -1`);
  $pchirp_second =~ m/(-{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* )/;
  $pchirp_second =$1;
  $pchirp_first="1.000 0.000 0.000 0.000 1.000 0.000 0.000 0.000";

  $iframecount = 1;
  while() {

    if($USE_INTERPOLATION == 0 ) { 
      ##form a sensible output filename
      ##appending digits will ls in correct order: v01.jpg->v011.jpg->v02.jpg
      ##so for now, dumb append of 1 to first of pair
      $fileout = substr($filea,0,$periodpositiona); # before period
      $fileout = $fileout.$iframecount;
      if ($output_is_png eq TRUE) {  # pchirp2nocrop.c will insert NaN into alpha
        $fileout = $fileout.".png"; # e.g. v000d.png, v001d.png, etc...
      }
      if ($output_is_png eq FALSE) { # pchirp2nocrop.c will reserve NaN in pnm
        $fileout = $fileout.".pnm"; # e.g. v000d.pnm, v001d.pnm, etc...
      }
      printf("No interpolation done, holding previous frame\n");
      `cp $filea $fileout`;
    }
    else {

    ##interpolate parameters for new frame
    ##$pinterp_location = (1-$iframecount/($NUM_INTRA_FRAMES+1));
    $pinterp_location = ($iframecount/($NUM_INTRA_FRAMES+1));
    printf("iframecount=%d, location=%f\n", $iframecount, $pinterp_location);
    printf("About to pinterp %s %s %f %s\n",$pchirp_first, $pchirp_second, 
           $pinterp_location, $MSE_THRESH);
    chop($pchirp_interp=`pinterp $pchirp_first $pchirp_second $pinterp_location  $MSE_THRESH`);
    printf("new params %s\n",$pchirp_interp);
  
    ##make a cement of the pair (in the case of linear interpolation)
    ##first generate the new parameters for each first,second frame, use
    ##interpolated frame as the identity
    ##`echo $pchirp_first > pork_parameters_interpd.txt`;
    ##`echo $pchirp_interp > pork_parameters_interpd.txt`;
    ##`echo $pchirp_second >> pork_parameters_interpd.txt`;
    ###`pintegrate pork_parameters_interpd.txt 1 > pork_Parameters_interpd.txt`;
  
    ##figure out how to chirp first frame if interp frame were ID
    `echo $pchirp_interp > pork_parameters_interpd.txt`;
    `pintegrate pork_parameters_interpd.txt 1 > pork_Parameters_interpd.txt`;
    ##$Parameters_first = `tail -1 pork_Parameters_interpd.txt`;
    $Parameters_first = `head -1 pork_Parameters_interpd.txt`;
    printf("ABOUT TO pchirp2nocrop deleteme_deletemea.pnm pork_imga.pnm %s\n",
            $Parameters_first);
    `pchirp2nocrop deleteme_deletemea.pnm pork_imga.pnm $Parameters_first`;
  
    ##figure out how to chirp second frame if interp frame were ID
    ##for now, just reapply the interpd params 
    ##(2 applications of interpd params should result in the original first frame   ## to second frame pairwise chirp, but only approximately gets it in practice
    ## because linear interpolation is poor so i think)
    ##better:  p1.E1(x)=E2(x)
    ##say p1' is the interpolated parameters, then
    ##p1 = p1'.p1'.e ,where e is error due to extrapolating from interpolated p1' 
    ##but that would be too easy, so rather get
    ##p1 = p1'.p1'' where p1'' is what E2(x) looks like if p1' were ID
    ##p1'' is what we want ,so solve for it like:
    ##                  p1 = p1'.p1''
    ##          p1^-1 . p1 = p1^-1 . p1'.p1''
    ##          p1^-1 . p1 = p1''
    ## and we know p''^-1 so we know p'' too
    ## intuitively: p'' = p1 . p1'^-1, means go all the way to p1 then rewind by
    ##                                 the p1'
  
    ##ez way,  apply p', that is 
    ##p1 = p1'.p1'.e  ignore e
    ##`pchirp2nocrop deleteme_deletemeb.pnm pork_imgb.pnm $pchirp_interp`;
  
    ## determine p1'^-1
    `echo $pchirp_interp > pork_pprime.txt`;
    chop($p_prime_inv=`pintegrate pork_pprime.txt 1 | head -1`);
    
    ## determine p1 . p1'^-1
  
    $pchirp_second =~ m/(-{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* -{0,1}\d.\d* )/;
    ##$pchirp_second =~ m/(-{0,1}\d.\d*)/;
    $pchirp_second ="$1";
    printf("NEW PCHIRP_SECOND=%s\n", $pchirp_second);
    
    `echo $pchirp_second > pork_p_double_prime.txt`;
    `echo $p_prime_inv >> pork_p_double_prime.txt`;
    chop($p_double_prime=`pintegrate pork_p_double_prime.txt 0 | tail -1`);
    printf("p double prime is %s\n", $p_double_prime);
    printf("ABOUT TO pchirp2nocrop deleteme_deletemeb.pnm pork_imgb.pnm %s\n",
           $p_double_prime);
    `pchirp2nocrop deleteme_deletemeb.pnm pork_imgb.pnm $p_double_prime`;
  
    ##cement output 
    ##`cement pork_imgout.pnm pork_imga.pnm pork_imgb.pnm`;
    ##cement a on the top so ID frame is kept if MSE too large (i.e. bizarre imgb warp hidden)
    ##`cement pork_imgout.pnm pork_imga.pnm pork_imgb.pnm`;
    `cement pork_imgout.pnm pork_imgb.pnm pork_imga.pnm`;
  
    ##form a sensible output filename
    ##appending digits will ls in correct order: v01.jpg->v011.jpg->v02.jpg 
    ##so for now, dumb append of 1 to first of pair
    $fileout = substr($filea,0,$periodpositiona); # before period
    $fileout = $fileout.$iframecount;
    if ($output_is_png eq TRUE) {  # pchirp2nocrop.c will insert NaN into alpha
      $fileout = $fileout.".png"; # e.g. v000d.png, v001d.png, etc...
    }
    if ($output_is_png eq FALSE) { # pchirp2nocrop.c will reserve NaN in pnm
      $fileout = $fileout.".pnm"; # e.g. v000d.pnm, v001d.pnm, etc...
    }
    printf("fileout is %s\n", $fileout);
    ##now extract the desired area to form new  file (from extract_3.pl)
    ##cooks up a porkchop.pnm
    $leftpad = $window_horiz_size/2;
    $rightpad = $window_horiz_size/2;
    $toppad = $window_vert_size/2;
    $bottompad = $window_vert_size/2;
    $output = `head -5 pork_imgout.pnm`;
    print $output."\n";
  
    $output =~ m/MOffset (-{0,1}\d*)/;
    $MOffset = $1;
  
    $output =~ m/NOffset (-{0,1}\d*)/;
    $NOffset = $1;
  
    printf("MOffset = %d, NOffset = %d\n", $MOffset, $NOffset);
  
    $dimensions = `head -4 pork_imgout.pnm | tail -1`;
    $dimensions =~ m/(\d*) (\d*)/;
    $N_dim = $1;
    $M_dim = $2;
  
    printf("M_dim = %d, N_dim = %d\n",  $M_dim, $N_dim) ;
    $centre_N = abs($NOffset) + $ident_horiz_size/2;
    $centre_M = abs($MOffset) + $ident_vert_size/2;
    printf("Centre is at (m,n) = (%d, %d)\n", $centre_M, $centre_N);
  
    #pad the image to ensure we crop ok
    `cp pork_imgout.pnm pork_pad_accumulator.pnm`;
    `pnmpad -white -l$leftpad -r$rightpad -t$toppad -b$bottompad pork_imgout.pnm > pork_padded.pnm`;
  
    $centre_N = abs($NOffset) + $ident_horiz_size/2 + $leftpad;
    $centre_M = abs($MOffset) + $ident_vert_size/2 + $toppad;
  
    #x,y as in pnmutils def, horiz
    $cutx = $centre_N - $window_horiz_size/2;
    $cuty = $centre_M - $window_vert_size/2;
  
    `pnmcut $cutx $cuty $window_horiz_size $window_vert_size pork_padded.pnm > pork_chop.pnm`;
  
    `cp pork_chop.pnm $fileout`;
    printf("Wrote %s\n",$fileout);
    }  #end else
  
    $iframecount++;
    last if( $iframecount > $NUM_INTRA_FRAMES );
  }  #end while
  
  $prefixa = substr($filea,0,$periodpositiona);
  $newfileout = $prefixa."0.".$postfixa;
  `cp $filea $newfileout`;
  printf("Wrote %s\n",$newfileout);
}
