#!/usr/bin/perl

# remember to install perl-tk e.g. apt-get install perl-tk
# remember to install libtime-hires-perl e.g. apt-get install libtime-hires-perl

# virtual_camera.pl was developed by Corey Manders and James Fung
# to efficiently display lightspace images using a variety of
# command-line options to plm2pnm 

################################################################
##################### Main #####################################
use Tk;
require Tk::HList;
require Tk::ProgressBar;
use Getopt::Std;
use Tk::WinPhoto;
use Time::HiRes qw( usleep ualarm gettimeofday tv_interval );
use Cwd;

##### initial settings to pass to plm2pnm to generate a first image
$k = 1;
$blocksize = 1;
$scalefactor = 1;
$first_image_present = 0;
$lut; 
$mean_shift =2;
$var_gamma = 2;
$max_dilation = 4;
$gnf_max_thresh = -1;
$gnf_spacial_thresh = 0;
$show_nastiness = 0;
$variable_blocksize =0;
$rotation = 0;
$go_simple_toggle =0;
  
parse_command_line();
@gui_dims=generate_first_image();
create_ui(@gui_dims); # create the gui
MainLoop();           # Power up the GUI!

################## End of Main ###################################
##################################################################


# generate first image is called to display an image entered
# on the command line, or if no image was entered, display the
# initial splash screen. This Subroutine returns the size of the
# appropriate canvas size to use.
sub generate_first_image{

   #generate 1st image
   $cmdstr = "";
   if( $lut eq "" && $input_filename ne"") {
      print("virtual camera: loading first plm\n");
      $cmdstr=sprintf("plm2pnm $input_filename -o $opt_o -m $mean_shift");
   }else {
      print("virtual camera: loading first plm\n");
      $cmdstr=sprintf("plm2pnm $input_filename -o $opt_o --lut $lut -m $mean_shift");
   }


   if ($input_filename ne ""){
      $first_image_present = 1;
      system($cmdstr);
      @ppm_dims = read_ppm_dims(".pork");
      if ($ppm_dims[0]==undef || $ppm_dims[0]<300){
         @gui_dims =(480,640);
      }else{
         @gui_dims = @ppm_dims;
      }
   }else{
      $first_image_present = 0;
      @gui_dims =(480,640);
   }
   return @gui_dims;
}


# parse command line sets inital values 
sub parse_command_line{

   # Check the input arguments
   getopts("l:o:");
   if ($opt_l){
      $lut = $opt_l;
      print("\nUsing lut $lut\n");
   }
   if ($opt_o){
      print("Outputting resulting pnm to $opt_o\n");
   }
   else{
      $opt_o = ".pork";
   }
   $input_filename = $ARGV[0];
}

# read the image parameters of a ppm image
# i.e. the first line after the comments.
# The subroutine returns the size (x,y) in
# an array with two elements
sub read_ppm_dims{
   $filename = pop;
   open(FILE,$filename);
   $found_params=0;

   #first line should be P6, etc.
   $filetype =<FILE>;

   #following lines are comments which
   #begin with '#'
   while(!$found_params){
      $line =<FILE>;
      if (!($line=~/#/)){ 
         $found_params = 1;
      }
   }
   @params=split(/ /, $line);
   return @params;
} 

# Regenerate the image to display and update the virtual_camera screen.
# This procedure is called after the 'generate new image' button is 
# pressed.
sub compute_new_image{
   $root = MainWindow->new;
   
   print "nastiness ->$show_nastiness\n"; 
   $root->Label( -text => 'Computing New Image')->pack();



   $progress_bar=$root->ProgressBar( -relief => 'ridge',
                                     -length => 400,
                                     -width  => 30,
                                     -from   => 0,
                                     -gap    => 0,
                                     -to     => 100,
                                     -colors => [0, 'yellow'],
                                    )->pack();
   $root->deiconify();

   $progress_bar->value(0);
   $progress_bar->update();

   $truevar_gamma =  10 ** $var_gamma;

   $cmdstr=sprintf("./plm2pnm $input_filename  -m $mean_shift -o $opt_o -k $k  -b $blocksize   -d $max_dilation -n $gnf_spacial_thresh -g $truevar_gamma");
   #$cmdstr=sprintf("./plm2pnm $input_filename  -m $mean_shift -o $opt_o -k $k  -b $blocksize  -g $truevar_gamma ");
   if( $lut ne "") {
      $cmdstr = $cmdstr." --lut $lut ";
   }
   if ($show_nastiness == 1){
      $cmdstr = $cmdstr." -w ";
   }
   if ($variable_blocksize == 1){
      $cmdstr = $cmdstr." -v ";
   }
   $i = 0;
   print "$cmdstr\n";
   open(DONE_FILE,">/tmp/.counter");
   print DONE_FILE "0";
   close(DONE_FILE);

   if (!defined($kidpid = fork())) {
       # fork returned undef, so failed
       die "cannot fork: $!";
   } elsif ($kidpid == 0) {
                # fork returned 0, so this branch is the child
       exec($cmdstr);
               # if the exec fails, fall through to the next statement
       die "can't exec $cmdstr";
    } else { 
                # fork returned neither 0 nor undef, 
                # so this branch is the parent
      while ($i < 100){
         open(DONE_FILE,"/tmp/.counter");
         $i = <DONE_FILE>; 
         $progress_bar->value($i);
         $progress_bar->update();
        close(DONE_FILE);
         usleep(10000);
      }
       waitpid($kidpid, 0);
    } 


   $root->destroy();
   $image->configure(-file => '.pork');
   @ppm_dims = read_ppm_dims($opt_o);
   $canvas->configure(-height => $ppm_dims[1],
                      -width  => $ppm_dims[0],
                     );
   $mystr = "rotate_photo_$rotation";
   print $mystr;
   $rotation = 0;
   eval($mystr);
#   $image->update();
}





# save the current imagespace picture
sub save_imagespace{
   my $entry_window = MainWindow->new;
   my $filename;

   # create a text entry gui to allow the user to enter a filename
   $entry_window->Label(-text => 'Save file as'
                       )->pack(-side => 'top');

   $t = $entry_window->Entry(-width => 30,
                             -textvariable =>\$filename
                            );
   $container = $entry_window->Frame()->pack(side => 'bottom');

   $container->Button(-text => 'Save',
                      -command => sub{system("cp .pork $filename");
                                          $entry_window->destroy})->pack(-side => 'left');
   $container->Button(-text => 'Cancel',
                      -command => sub{ $entry_window->destroy})->pack(-side => 'right');
   
   $t->pack(-expand => 1,
           -fill   => 'both');
}

#load a new lightspace image
sub load_lightspace{
    $filelist = MainWindow->new();
    $hlist = $filelist->Scrolled('HList',
                                  -height       => 30,
                                  -width        => 30,
                                  -drawbranch   => 1,
                                  -separator    => '/',
                                  -indent       => 15,
                                  -command      => \&show_or_hide_dir);
    $hlist->pack(-fill   => 'both',
                 -expand => 'y'
                );
    $open_folder_bitmap = $filelist->Bitmap('open',
                                              -file      =>Tk->findINC('openfolder.xbm')
                                           );
    $closed_folder_bitmap = $filelist->Bitmap('closed',
                                              -file      =>Tk->findINC('folder.xbm')
                                             );
    show_or_hide_dir(".");

}

# show or hide dir expands a directory in the HList. If
# a non-directory is double-clicked, a new lightspace image
# is loaded.
sub show_or_hide_dir{
   my $path = $_[0];
   if (! -d $path){

   # load the new plm and get rid of the gui
      $filelist->destroy();
      $input_filename = $path;
      if( $lut eq "") {
         $cmdstr=sprintf("./plm2pnm $path -m $mean_shift -o $opt_o $input_filename ");
      }else {
         $cmdstr=sprintf("./plm2pnm $path  -m $mean_shift -o $opt_o --lut $lut $input_filename");
      }
           
      #print("$cmdstr\n");
      system($cmdstr);
      @ppm_dims = read_ppm_dims($opt_o);
      print ("1 = $ppm_dims[0]  2= $ppm_dims[1]\n");
      $first_image_present =1;
      $image->configure(-file => "$opt_o");
          
      $canvas->configure(-height => $ppm_dims[1],
                         -width  => $ppm_dims[0],

                        );
#      top_repack();
      $image->update();
      $canvas->update();
 $canvas->configure(-scrollregion => [$canvas->bbox("all")]);

   }
   return if (! -d $path);
   if ($hlist->info('exists',$path)){
        
      $next_entry = $hlist->info('next',$path);
      if(!$next_entry || (index($next_entry, "$path/") == -1)){
         $hlist->entryconfigure($path, 
                                  -image => $open_folder_bitmap
                               );
           add_dir_contents($path);
      }else{
         $hlist->entryconfigure($path,
                                  -image => $closed_folder_bitmap
                               );
         $hlist->delete('offsprings',$path);
      }
   }else {
      die "'$path' is not in directory\n" if (! -d $path);
      $hlist->add($path, -itemtype   => 'imagetext',
                         -image       => $icons{"open"},
                         -text        => $path);
      add_dir_contents($path);
   }
}

# load a new lookup table
sub load_lut{
   $filelist = MainWindow->new();
    $hlist = $filelist->Scrolled('HList',
                                  -height       => 30,
                                  -width        => 30,
                                  -drawbranch   => 1,
                                  -separator    => '/',
                                  -indent       => 15,
                                  -command      => \&show_or_hide_lut_dir
                                );
    $hlist->pack(-fill   => 'both',
                 -expand => 'y');
    $open_folder_bitmap = $filelist->Bitmap('open',
                                              -file      =>Tk->findINC('openfolder.xbm')
                                           );
    $closed_folder_bitmap = $filelist->Bitmap('closed',
                                              -file => Tk->findINC('folder.xbm') 
                                             );
    show_or_hide_lut_dir(".");

}

# show or hide dir expands a directory in the HList. If
# a non-directory is double-clicked, a new lookup table 
# is loaded.
sub show_or_hide_lut_dir{
   my $path = $_[0];
   if (! -d $path){

     # load the new plm and get rid of the gui
     $filelist->destroy();
     $lut = $path;
 
   }
   return if (! -d $path);
   if ($hlist->info('exists',$path)){
        
      $next_entry = $hlist->info('next',$path);
      if(!$next_entry || (index($next_entry, "$path/") == -1)){
         $hlist->entryconfigure($path, 
                                   -image => $open_folder_bitmap
                               );
         add_dir_contents($path);
      } else{
         $hlist->entryconfigure($path,
                                  -image => $closed_folder_bitmap
                               );
         $hlist->delete('offsprings',$path);
      }
   }else {
      die "'$path' is not in directory\n" if (! -d $path);
      $hlist->add($path, -itemtype   => 'imagetext',
                         -image       => $icons{"open"},
                         -text        => $path);
       add_dir_contents($path);
   }
}


# used by load_lightspace and load lut to add directory contents
# to a visible HList
sub add_dir_contents{
   my $path = $_[0];
   my $oldcursor = $filelist->cget('cursor');
   $filelist->configure(-cursor => 'watch');
   $filelist->update();
   my @files = glob "$path/*";
   foreach $file (@files){
      $file =~ s|//|/|g;
      ($text = $file) =~ s|^.*/||g;
      if (-d $file){
         $hlist->add($file, -itemtype => 'imagetext',
                            -image => $closed_folder_bitmap, 
                            -text => $text
                    );
      }else{
         $hlist->add($file, -itemtype => 'text',
                            -text => $text
                    );
      }
   }
   $filelist->configure(-cursor => $oldcursor);
}
        

# draw_pixel puts a black pixel where the mouse is currently
sub draw_pixel{
   my $y_value = pop;
   my $x_value = pop;
  # $canvas->createOval($x_value,$y_value,$x_value,$y_value,-fill => 'black');
}

# update_mouse_coord is used to update the mouse position and
# corresponding pixel value on the canvas on the gui everytime
# the mouse is moved.
sub update_mouse_co_ord{
   $y_val = pop;
   $x_val = pop;

   $pstring = sprintf("pixel co-ord = %3.3d %3.3d",$x_val,$y_val);
   #@pix_value = (1,2,3);

   $x_val = int $x_val;
   $y_val = int $y_val;
   @pix_value = $image->get($x_val,$y_val);
   $vstring = sprintf("r=%3.3d g=%3.3d b=%3.3d",$pix_value[0],$pix_value[1],$pix_value[2]); 
   $pixel_co_ord->configure('text' => $pstring);
   $pixel_label->configure('text' => $vstring);  
}

# gamma correction will do typical imagespace gamma correction on
# an image
sub gamma_correct{
   $gamma =1;
   $gamma_window = MainWindow->new(); 
   $gamma_window->Scale(orient        => 'horizontal',
                        from          => 0,
                        to            => 20,
                        tickinterval  => 5 ,
                        resolution    => 1/10 ,
                        label         => 'gamma value',
                        width         => 5,
                        length        => 300,
                        variable      => \$gamma,     #k value
                        command       => sub{$image->configure(-gamma =>  $gamma)}
                       )->pack(side => 'top',
                               fill => 'x'
                              );
   $gamma_window->Button(text      =>   'Done',
                         command   =>   sub{$gamma_window->destroy()} 
                         )->pack(side =>'bottom');

}

sub print_image{
   $print_window = MainWindow->new;
   $print_command = "lpr";
   $rb_value = "printer";
   $filename = "picture.ps";
   $print_window->Label(-text    => 'Print the current image'
                       )->grid(-sticky     => "nsew",
                               -columnspan => 2);
  
   #first line of the print_window
   $print_window->Radiobutton(-text   => 'Printer',
                              -value  => 'printer',
                              -variable => \$rb_value)->grid(
      $print_window->Label(-text     => 'Print Command'),
      $print_window->Entry(-textvariable => \$print_command,
                           -width  => 40),
       
      -sticky => "nsew" ,
   );
   #second line of the print_window
      $print_window->Radiobutton(-text   => 'File     ',
                                 -value  => 'file',
                                 -variable => \$rb_value)->grid(
      $print_window->Label(      -text   => 'Filename '),
      $print_window->Entry(      -textvariable => \$filename,
                                 -width =>  40),
      -sticky => "nsew"
   );
  
   #third line of the print window
      $print_window->Button( -text    =>   'Print',
                             -command =>   [\&print_pic,$rb_value,$print_command,$filename]
                           )->grid(
      $print_window->Button(
                             -text    =>   'Don\'t Print',
                             -command =>   sub{$print_window->destroy}
                           ),
                   -sticky => "nsew",
                   -columnspan =>2 
                                 );
   $print_window->gridColumnconfigure(1, -weight => 1);
   $print_window->gridRowconfigure(1, -weight => 1);
   $print_window->bind('<Key-Return>',[\&print_pic,$rb_value,$print_command,$filename]);

  }
sub print_pic{
   $filename = pop;
   $print_command = pop;
   $rb_value = pop;
   
   ### print to a file or to the printer
   if ($rb_value eq "printer"){
      print "about to render postscript\n";
      $canvas->postscript(-file => "/tmp/ps.out",
                          -colormode => "gray"
                         );


      print "about to issue print comand";
      system("$print_command /tmp/ps.out");
      print "returned\n";

   }else{
      $canvas->postscript(-file => $filename
                          -colormode => "gray"
                         );
   } 
   $print_window->destroy;
}
sub rotate_photo_0;
sub rotate_photo_90{
    if ($rotation == 90) {return;}
    my $tmp = $image->Tk::Widget::image('create','photo');
    bless $tmp, 'Tk::Photo';
    $image->configure(-file => $opt_o);
    $image->update;
    my $width = $image->width;
    my $height = $image->height;
    for(my $x=0;$x<$width; $x++){
        my $curpix = $image->data(-from => $x, 0 ,$x+1,$height);
        $curpix = "{$curpix}";
        $tmp->put($curpix, -to => 0, $width -$x -1 );
    }
    $image->blank;
    $image->copy($tmp);
    print "height = $height width = $width\n"; 
    $image->configure(-height => $width,
                      -width => $height);
    $image->idletasks;
    $tmp->delete;
    print "height = $height width = $width\n"; 
    $canvas->configure(-height => $width ,
                      -width => $height );
    $height=$canvas->height;
    $width=$canvas->width;
    print "height = $height width = $width\n"; 
    $image->update;
    $canvas->update;
    $top->update;
    $image->redither;
    $rotation = 90;
}
sub rotate_photo_270{
    my $tmp = $image->Tk::Widget::image('create','photo');
    bless $tmp, 'Tk::Photo';
    if ($rotation == 270) {return;}
    $image->configure(-file => $opt_o);
    my $height = $image->height;
    my $width = $image->width;

    for(my $x=0;$x<$height; $x++){
        my $curpix = $image->data(-from => 0, $x ,$width,$x+1);
        $curpix =~ s/^{(.*)}$/$1/;
        $tmp->put($curpix, -to => $height - $x - 1, 0);
    }
    $image->blank;
    $image->copy($tmp);
    $image->configure(-height => $width,
                      -width => $height);
    $image->idletasks;
    $tmp->delete;
    $canvas->configure(-height => $width,
                      -width => $height);
    $image->update;
    $canvas->update;
    $top->update;
    $rotation = 270;
}
sub rotate_photo_180{
    my $tmp = $image->Tk::Widget::image('create','photo');
    bless $tmp, 'Tk::Photo';


    $tmp->copy($image, -subsample => -1,-1);
    $image->blank;
    $image->copy($tmp);
    $image->idletasks;
    $tmp->delete;
    $image->update;
    $canvas->update;
    $top->update;
    $rotation = ($rotation +180) % 360;
}
sub take_snap{
   my $filename = "screen.ps";
   $snap_window = MainWindow->new;
   $snap_window->Label(-text    => 'Take a Snapshot of the current Screen'
                       )->pack();
   $snap_frame=$snap_window->Frame; 
   $snap_frame->Label(-text => 'save as:')->pack(side => 'left');
   $snap_frame->Entry(-textvariable => \$filename,
                           -width  => 40)->pack(side => 'right');
   $snap_window->bind('<Key-Return>',[\&snap_pic,$filename]);
   $snap_frame->pack();
   $snap_window->Button( -text    =>   'Take Picture',
                          -command =>   [\&snap_pic,$filename]
                        )->pack(side => 'right');
   $snap_window->Button(
                         -text    =>   'Don\'t Take Picture',
                         -command =>   sub{$snap_window->destroy}
                        )->pack(side => 'left');
  
}
sub snap_pic{
   my $filename = pop;

   $image2 = $top->Photo(-format => 'Window', -data => oct($top->id)); 
   $image2->write('/tmp/pork.ppm', -format => 'PPM');
   #$canvas->postscript(-file => "ps.out");
   $snap_window->destroy;
}


# convert to the simpler interface */
sub go_simple{
   if ($go_simple_toggle == 0){
      #$canvas->CanvasBind('<Button1-Motion>' => "");
      #$canvas->CanvasBind('<Motion>' => "");
      $menu_bar->packForget();
      $center_row->packForget();
      $frame1->packForget();
      $frame4->packForget();
      $frame5->packForget();
      $frame2->packForget();
      $bar1->packForget();
      $bar2->packForget();
      $bar3->packForget();
      $go_simple_toggle = 1;
      $log_scroll6->pack(side => 'bottom',
                                 fill => 'x');
      $log_scroll_scale6->pack(side => 'bottom',
                               fill => 'x');
      $log_scroll5->pack(side => 'bottom',
                               fill => 'x');
   
      $log_scroll4->pack(side => 'bottom',
                                 fill => 'x');
      $log_scroll_scale4->pack(side => 'bottom',
                               fill => 'x');
   
      $log_scroll3->pack(side => 'bottom',
                         fill => 'x');
      $log_scroll2->pack(side => 'bottom',
                         fill => 'x');
      $log_scroll_scale2->pack(side => 'bottom',
                               fill => 'x');
      $log_scroll->pack(side => 'bottom',
                        fill => 'x');
      $canvas->pack(side => 'top');

      $top->update();
   }
  else{
     
     $log_scroll6->packForget();
     $log_scroll5->packForget();
     $log_scroll4->packForget();
     $log_scroll3->packForget();
     $log_scroll2->packForget();
     $log_scroll->packForget();
     $canvas_frame->packForget();
     $canvas->packForget(); 
     $menu_bar->pack('-side' => 'top',
                     '-fill' => 'x');
     $center_frame->pack(side => 'top',
                         fill => 'x');

     $bar1->pack(side => 'top',
                 fill => 'x');
     $center_row->pack(side => 'top',
                       fill => 'x'
                      );
     $canvas_frame->pack(side => 'top');
     $frame1->pack(side => 'left');
     $frame4->pack(side => 'left');
     $frame5->pack(side => 'left');
     $canvas->pack(side => 'left');
     $frame2->pack(side => 'left');
     $bar2->pack(side => 'bottom',
                 fill => 'x');
     $bar3->pack(side => 'bottom',
                 fill => 'x');
     $go_simple_toggle =0;
     $top->update();
   }
}
sub top_repack{
   my @kids = $top->packSlaves;
   foreach(@kids){
      $_->pack();
   }
}
################ Create the GUI #######################
sub create_ui{
   $y_dim = pop;
   $x_dim = pop;
   $top = MainWindow->new(-title  => 'Virtual Camera',
                          -height => 768,
                          -width  => 1024); #initialize the gui
   $top->packPropagate(0);
   $canvas_frame=$top->Frame; 
      if ($first_image_present){
         $image = $canvas_frame->Photo('toshow',
                               -file   => $opt_o 
                             );
         @ppm_dims = read_ppm_dims($opt_o);
         $hgt = $ppm_dims[0];
         $wdth =$ppm_dims[1];
         print "hgt = $hgt wdth = $wdth\n";
         $canvas =$canvas_frame->Scrolled('Canvas',
                                -height => 450,
                                -width  => 700 
                              # -confine   => 'false'
                              );
         $canvas->createImage(0, 
                              0, 
                                -anchor => nw,
                                -image => 'toshow'
                              );
         
         $canvas->configure(-scrollregion => [$canvas->bbox("all")]);


 
       }else {
          $image = $top->Photo('toshow',
                                -file      =>  'splash.ppm' 
                              );
          $canvas =$canvas_frame->Scrolled('Canvas',
                              #             -height => 525,
                              #             -width  => 800
                                           );
          $canvas->createImage(0, 
                               0, 
                               -anchor => nw,
                               -image => 'toshow'
                              );
       }
    
   ########## the menubar %%%%%%%%%%%%%%%%%
   $menu_bar = $top->Frame()->pack('-side' => 'top',
                                   '-fill' => 'x'
                                  );

   #create the "file" button
   $file_mb = $menu_bar->Menubutton('-text' =>   'File',
                                    '-relief' => 'raised',
                                    '-borderwidth' => 2,
                                    )->pack('-side' => 'left',
                                            '-padx' => 2
                                           );

   $file_mb->command( -accelerator => 'Ctrl+S',
                      -label      => 'Save current imagespace image',
                      -underline  =>0,
                      -command     => \&save_imagespace
                    );

   $top->bind('<Control-s>' => \&save_imagespace) ;

   $file_mb->command( -accelerator => 'Ctrl+L',
                      -label       => 'Load new lightspace image',
                      -underline   =>0,
                      -command     => \&load_lightspace 
                    );
   $top->bind('<Control-q>' => \&load_lightspace) ;

   $file_mb->command( -accelerator => 'Ctrl+T',
                      -label       => 'Load a new lookup table',
                      -underline   => 18,
                      -command     => \&load_lut 
                    );
   $top->bind('<Control-l>' => \&load_lut) ;

   $file_mb->command( -accelerator => 'Ctrl+P',
                      -label       => 'Print the current image',
                      -underline   => 0,
                      -command     => \&print_image
                    );

   $top->bind('<Control-p>' => \&print_image) ;
   $file_mb->separator();
   $file_mb->command( -accelerator => 'Ctrl+Q',
                      -label       => 'Quit',
                      -underline   => 0,
                      -command     => sub{exit(0)} 
                     );

   $top->bind('<Control-q>' => sub{exit(0)}) ;

   $view_mb = $menu_bar->Menubutton(text =>   'View',
                                    relief => 'raised',
                                    borderwidth => 2,
                                    )->pack(side => 'left',
                                            padx => 0
                                            );
   $view_mb->command(-label    => 'simple interface',
                     -command  => \&go_simple
                    );
   $view_mb->command(-label    => 'imagespace gamma',
                     -command  => \&gamma_correct
                    );
   $view_mb->command(-label    => 'rotate 90 degress',
                     -command  =>  \&rotate_photo_90
                    );
   $view_mb->command(-label    => 'rotate 180 degress',
                     -command  =>  \&rotate_photo_180
                    );
   $view_mb->command(-label    => 'rotate 270 degress',
                     -command  =>  \&rotate_photo_270
                    );
   $screensize_mb = $menu_bar->Menubutton(text        => 'Screensize',
                                          relief      => 'raised',
                                          borderwidth => 2,
                                         )->pack(side => 'left',
                                                 padx => 0
                                                );
   $screensize_mb->command(-label    => '640 x 480',
                           -command  => sub{$top->configure(-width => 640,
                                                            -height=> 480 
                                                           );
                                            top_repack();
                                            }
                          );

   $screensize_mb->command(-label    => '800 x 600',
                           -command  => sub{$top->configure(-width => 800,
                                                            -height=> 600
                                                           );
                                            top_repack();
                                            }
                          );
   $screensize_mb->command(-label    => '1024 x 768',
                           -command  => sub{$top->configure(-width => 1024,
                                                            -height=> 768 
                                                           );
                                            top_repack();
                                            }
                          );
   $screensize_mb->command(-label    => '1280 x 1024',
                           -command  => sub{$top->configure(-width => 1280,
                                                            -height=> 1024 
                                                           );
                                            top_repack();
                                            }
                          );



   $snap_button=$menu_bar->Button(-text    => 'Snapshot',
                     -command  => sub{ 
                                      $image2 = $top->Photo(-format => 'Window', -data => oct($top->id)); 
                                      $image2->write('/tmp/pork.ppm', -format => 'PPM');
                                      $tmp_window = MainWindow->new();
                                      $second_canvas =$tmp_window->Canvas( -height    => $top->height,
                                                             -width     =>  $top->width,
                                                            -confine   => 'false'
                                                           );
                                      $image9 = $tmp_window->Photo('toshow2',
                                                               -file   => "/tmp/pork.ppm" 
                                                             );
                                      system("pnmtops /tmp/pork.ppm > screenshot.ps");
                                      $second_canvas->createImage(0, 
                                                              0, 
                                                              -anchor => nw,
                                                              -image => 'toshow2' 
                                                             );
                                      $second_canvas->pack();
                                      #$second_canvas->postscript(-file => "/tmp/pork.ps",
                                      #                           -colormode => "gray");
                                      $tmp_window->destroy;
                      })->pack(side => 'right');

   
                      

   ############ the sliders ################
   # make a k scrollbar
   $bar1= $top->Scale(orient        => 'horizontal',
               from          => 0,
               to            => 16,
               tickinterval  => 2 ,
               resolution    => 1/100 ,
               label         => '(2^1/3)^i gain (k)',
               width         => 5,
               variable      => \$k,     #k value
              )->pack(side => 'top',
                      fill => 'x' 
                     );
   ############ the button ################
   # make a generate image button
   $center_row = $top->Frame;
   $button = $center_row->Button(text      =>   'Generate New Image',
                                 command   =>   [\&compute_new_image, $opt_o]
                                 )->pack(side =>'left');

   ########## the pixel labels ###########
   $pixel_co_ord =$center_row->Label(
                                    text      =>   'pixel co-ord = 0,0',
                                    relief    =>   'groove',
                                    width     =>   30,
                                    height    =>   1
                                    )->pack(side => 'left');
 
    $pixel_label =$center_row->Label(
                                    text      =>   'r=255 g=255 b=255',
                                    relief    =>   'groove',
                                    width     =>   30,
                                    height    =>   1
                                    )->pack(side => 'left');
    $center_frame = $center_row->Frame();
    $show_nast = $center_frame->Checkbutton(
                                          -text      =>  "show nastiness           ",
                                          -variable => \$show_nastiness
                                         )->pack(side => 'top');
    $var_block = $center_frame->Checkbutton(
                                          -text      =>  "use variable blocksizes",
                                          -variable => \$variable_blocksize
                                         )->pack(side => 'top');
    $center_frame->pack(side => 'left');
    $center_row->pack(side => 'top',
                      fill => 'x'
                     );

     
   $canvas_frame->pack(side => 'top'); 


   # make a mean scale-factor scrollbar
    $frame1 = $canvas_frame->Frame;
    $frame1 -> Label(-text => 'Mean')->pack();
    $frame1 -> Label(-text => 'Scale')->pack();
    $frame1 -> Label(-text => 'Factor')->pack();
    $frame1->Scale(orient       => 'vertical',
               from          => 1,
               to            => 10,
               tickinterval  => 1/2 ,
               resolution    => 1/50 ,
               length        => 300,    #in pixels
               width         => 5,
               variable      => \$mean_shift,     #scalefactor value
              )->pack(side   => 'left',
                      fill   => 'y'
                     );
    $frame1->pack(side => 'left');
    # GNF max threshold scrollbar
    $frame4 = $canvas_frame->Frame;
    $frame4->pack(side => 'left');

    # GNF spacial gradient scrollbar
    $frame5 = $canvas_frame->Frame;
    $frame5 -> Label(-text => 'GNF')->pack();
    $frame5 -> Label(-text => 'Grad')->pack();
    $frame5 -> Label(-text => 'Thresh')->pack();
    $frame5->Scale(orient       => 'vertical',
               from          => 0,
               to            => 1/10,
               tickinterval  => 1/100 ,
               resolution    => 1/500000 ,
               length        => 300,    #in pixels
               width         => 5,
               variable      => \$gnf_spacial_thresh,     #scalefactor value
              )->pack(side   => 'left',
                      fill   => 'y'
                     );
    $frame5->pack(side => 'left');

    $canvas->pack(-side=> 'left',
                  -fill=> 'both');

    $frame2 = $canvas_frame->Frame;
    $frame2 ->Label(-text => 'block')->pack(); 
    $frame2 ->Label(-text => 'size')->pack(); 

   # make a blocksize scrollbar
    $frame2->Scale(orient       => 'vertical',
               from          => 1,
               to            => 100,
               tickinterval  => 10 ,
               resolution    => 1 ,
               length        => 300,    #in pixels
               width         => 5,
               variable      => \$blocksize,     #blocksize value
                 )->pack(side => 'left',
                      fill => 'y');
    $frame2->pack(side => 'left');

   # make a variance gamma scrollbar
    $bar2= $top->Scale(orient       => 'horizontal',
               from          => 0,
               to            => 5,
               tickinterval  => 1,
               resolution    => 1/50 ,
               label         => 'variance gamma',
               length        => 300,    #in pixels
               width         => 5,
               variable      => \$var_gamma,     #blocksize value
                 )->pack(side => 'bottom',
                      fill => 'x');

   $canvas->pack(side=> 'left');
   # make a max dilation scrollbar
   $bar3= $top->Scale(orient       => 'horizontal',
               from          => 1,
               to            => 10,
               tickinterval  => 1,
               resolution    => 1/50,
               label         => 'max dilation',
               length        => 300,    #in pixels
               width         => 5,
               variable      => \$max_dilation,     #blocksize value
                 )->pack(side => 'bottom',
                      fill => 'x');



   ############ the canvas and corresponding ppm ###############
   $canvas->CanvasBind('<Button1-Motion>' => [\&draw_pixel,Ev('x'),Ev('y')]);
   $canvas->CanvasBind('<Motion>' => [\&update_mouse_co_ord,Ev('x'),Ev('y')]);
   $top->bind('<KeyPress-i>',\&go_simple);

   ############### this doesn't show up until go_simple #########
   $log_scroll = $top->Frame();
   $font_1 ='courier 16 bold';
      $log_scroll->Label(-text => '2', -font => $font_1)->form(-left => '%3');
      $log_scroll->Label(-text => '2.8',-font => $font_1)->form(-left => '%13');
      $log_scroll->Label(-text => '4', -font => $font_1)->form(-left => '%30');
      $log_scroll->Label(-text => '5.6', -font => $font_1)->form(-left => '%43');
      $log_scroll->Label(-text => '8', -font => $font_1)->form(-left => '%60');
      $log_scroll->Label(-text => '11', -font => $font_1)->form(-left => '%75');
      $log_scroll->Label(-text => '16', -font => $font_1)->form(-left => '%91');

   $log_scroll2 = $top->Frame();
   $log_scroll_scale2=$log_scroll2->Scale(-orient  => 'horizontal',
                                          -troughcolor => 'white',
                                          ##-sliderlength=> 50,
                                          -background  => 'black',
                                          -showvalue => 0
                                         );
   $font_2 ='courier 12 bold';
   $log_scroll3= $top->Frame();
      $log_scroll3->Label(-text => '1/500',-font => $font_2)->form(-left => '%0');
      $log_scroll3->Label(-text => '1/250',-font => $font_2)->form(-left => '%15');
      $log_scroll3->Label(-text => '1/125',-font => $font_2)->form(-left => '%30');
      $log_scroll3->Label(-text => '1/60',-font => $font_2)->form(-left => '%45');
      $log_scroll3->Label(-text => '1/30',-font => $font_2)->form(-left => '%60');
      $log_scroll3->Label(-text => '1/15',-font => $font_2)->form(-left => '%75');
      $log_scroll3->Label(-text => '1/4',-font => $font_2)->form(-left => '%90');
   $log_scroll4 = $top->Frame();
   $log_scroll_scale4=$log_scroll4->Scale(-orient    => 'horizontal',
                                          -troughcolor => 'white',
                                          -background  => 'black',
                                          -showvalue => 0
                                         );
      $font_3 ='courier 12 bold';
      $log_scroll5= $top->Frame();
      $log_scroll5->Label(-text => '1.0',-font => $font_3 )->form(-left => '%0');
      $log_scroll5->Label(-text => '1.5',-font => $font_3)->form(-left => '%18');
      $log_scroll5->Label(-text => '2.0',-font => $font_3)->form(-left => '%36');
      $log_scroll5->Label(-text => '2.5',-font => $font_3)->form(-left => '%54');
      $log_scroll5->Label(-text => '3.0',-font => $font_3)->form(-left => '%72');
      $log_scroll5->Label(-text => '3.5',-font => $font_3)->form(-left => '%90');

   $log_scroll6 = $top->Frame();
   $log_scroll_scale6=$log_scroll6->Scale(-orient  => 'horizontal',
                                          -troughcolor => 'white',
                                          -background  => 'black',
                                          -showvalue => 0
                                         );
 
}

