#ifndef _PNMIMAGEOFFSETABLE_H_
#define _PNMIMAGEOFFSETABLE_H_

#include <stdio.h>
#include "mat_util.h"

typedef struct {
/* private: */
  int mOffset;
  int nOffset;
  int xSize;
  int ySize;
  int depth;
  int certchans; //number of certainty channels
  int magic_number;
  int plm_field_1;
  float plm_field_2;
  Image *data;
} PNMImageOffsetable; 

#define PNM_SUCCESS 0
#define PNM_ERROR -1
#define PNM_NULL -2 /* PNM_NULL is returned if a pPNM parameter is NULL: */
#define PNM_FILE_OPEN_ERROR -3 /* PNM_FILE_ERROR includes all errors related to files other than failure to open them, including (especially) bad formats. */
#define PNM_FILE_ERROR -4
#define PNM_OUT_OF_MEMORY -5

/* Helper functions: 
   All return PNM_SUCCESS on success and PNM_XXXX (see above) on failure,
   unless otherwise noted. */

/* Always call PNM_Init* before any other helper functions.
   Also note that only InitFromFile initializes the actual data. */
int PNM_Init( PNMImageOffsetable * pPNM );
int PNM_InitFromFile( PNMImageOffsetable * pPNM, const char * filename);
int PNM_InitFromSize( PNMImageOffsetable * pPNM, int xSize, int ySize, int depth, int magic_number, int plm_field_1, float plm_field_2);

/* Always call PNM_Destroy when finished with a PNMImageOffsetable */
void PNM_Destroy( PNMImageOffsetable * pPNM );

int PNM_Copy( PNMImageOffsetable * pTarget, const PNMImageOffsetable * pSource );

int PNM_WriteToFile( const PNMImageOffsetable * pPNM, const char * filename);

/* These return the value in question on success, and a PNM_XXXX on failure */
int PNM_GetXSize( const PNMImageOffsetable * pPNM ) ;
int PNM_GetYSize( const PNMImageOffsetable * pPNM ) ;
int PNM_GetDepth( const PNMImageOffsetable * pPNM ) ;

/*  These have no way to detect failure, so don't pass a NULL or bad pointer! */
int PNM_GetMOffset( const PNMImageOffsetable * pPNM ) ;
int PNM_GetNOffset( const PNMImageOffsetable * pPNM ) ;

/* This returns NULL on failure, and assert fails on an invalid pixel. */
Image * PNM_Pixel( const PNMImageOffsetable * pPNM, int x, int y);

int PNM_SetMOffset( PNMImageOffsetable * pPNM, int mOffset);
int PNM_SetNOffset( PNMImageOffsetable * pPNM, int nOffset);

int parseHeaderFrom( FILE * fp, int *xSize, int *ySize, int *depth, int *mOffset, int *nOffset, int *magic_number, int *plm_field_1, float *plm_field_2 );

#endif  // #defined _PNMIMAGEOFFSETABLE_H_




