#include <stdio.h>
#include "util.h"
#include "mat_util.h"
#include "pinverse.h"

void
printmatrix(Var *mat,
	    const char *name)
{
  int m, n;
  for (m = 0; m < mat->M; m++)
    {
      for (n = 0; n < mat->N; n++)
	{
	  fprintf(stderr, "%s[%d][%d] = %f\n",
		  name, m, n, mat->data[mat->N*m+n]);
	}
    }
}

int boundingbox(Var *parameters,
		Data *ml,
		Data *nl,
		Data *mh,
		Data *nh)
{
  /* Straight reimplementation of boundingbox.m */
  static const float REALLY_BIG_NUMBER = 1.0e12;
  Var X;
  Var A, b, c, ones14, ones21;
  Var num, den;
  Var denTemp;
  Var Y;
  Var pinver;
  int m1;

  if ((parameters->M != 8) || (parameters->N != 1))
    {
      fprintf(stderr, "boundingbox: (M,N) dims must be (8,1)\n");
      return -1;
    }
  init_var( &X, "X", 2, 4, INTERNAL_VARIABLE, GREY_SCALE,NULL );
  X.data[0] = 0;
  X.data[1] = 0;
  X.data[2] = 1;
  X.data[3] = 1;
  X.data[4] = 0;
  X.data[5] = 1;
  X.data[6] = 0;
  X.data[7] = 1;
  init_var( &pinver, "pinver", 1, 8, INTERNAL_VARIABLE, GREY_SCALE,NULL );
  pinverse(parameters, &pinver);
  init_var( &A, "A", 2, 2, INTERNAL_VARIABLE, GREY_SCALE,NULL );
  A.data[0] = pinver.data[0];
  A.data[1] = pinver.data[1];
  A.data[2] = pinver.data[3];
  A.data[3] = pinver.data[4];
  init_var( &b, "b", 2, 1, INTERNAL_VARIABLE, GREY_SCALE ,NULL);
  b.data[0] = pinver.data[2];
  b.data[1] = pinver.data[5];
  init_var( &c, "c", 1, 2, INTERNAL_VARIABLE, GREY_SCALE ,NULL);
  c.data[0] = pinver.data[6];
  c.data[1] = pinver.data[7];
  init_var( &ones14, "ones14", 1, 4, INTERNAL_VARIABLE, GREY_SCALE ,NULL);
  ones14.data[0] = 1;
  ones14.data[1] = 1;
  ones14.data[2] = 1;
  ones14.data[3] = 1;
  init_var( &ones21, "ones21", 2, 1, INTERNAL_VARIABLE, GREY_SCALE ,NULL);
  ones21.data[0] = 1;
  ones21.data[1] = 1;
  init_var( &num, "num", 2, 4, INTERNAL_VARIABLE, GREY_SCALE ,NULL); 
  matrix_multiply(&A, &X, &num);
  for (m1 = 0; m1 < 4; m1++)
    {
      num.data[m1] += b.data[0];
      num.data[4+m1] += b.data[1];
    }
  
  /*  printmatrix(parameters, "params");
  printmatrix(&pinver, "pinver");
  printmatrix(&X, "X");
  printmatrix(&A, "A");
  printmatrix(&b, "b");
  printmatrix(&c, "c");
  printmatrix(&num, "num"); */
  init_var( &denTemp, "denTemp", 1, 4, INTERNAL_VARIABLE, GREY_SCALE ,NULL);
  matrix_multiply(&c, &X, &denTemp);
  add_matrix(&denTemp, &ones14, &denTemp);

  /*  printmatrix(&denTemp, "denTemp"); */
  init_var( &den, "den", 2, 4, INTERNAL_VARIABLE, GREY_SCALE ,NULL);
  matrix_multiply(&ones21, &denTemp, &den);
  
  /* printmatrix(&den, "den"); */
  init_var( &Y, "Y", 2, 4, INTERNAL_VARIABLE, GREY_SCALE ,NULL);
  for (m1 = 0; m1 < 8; m1++)
    Y.data[m1] = num.data[m1] / den.data[m1];

  *ml = REALLY_BIG_NUMBER;
  *mh = -REALLY_BIG_NUMBER;
  *nl = REALLY_BIG_NUMBER;
  *nh = -REALLY_BIG_NUMBER;
  
  /* printmatrix(&Y, "Y"); */

  for (m1 = 0; m1 < 4; m1++)
    {
      if (Y.data[m1] < *ml)
	*ml = Y.data[m1];
      if (Y.data[m1] > *mh)
	*mh = Y.data[m1];
      if (Y.data[4+m1] < *nl)
	*nl = Y.data[4+m1];
      if (Y.data[4+m1] > *nh)
	*nh = Y.data[4+m1];
    }
  Var_Destructor( &X );
  Var_Destructor( &pinver );
  Var_Destructor( &A );
  Var_Destructor( &b );
  Var_Destructor( &c );
  Var_Destructor( &ones14 );
  Var_Destructor( &ones21 );
  Var_Destructor( &denTemp );
  Var_Destructor( &num );
  Var_Destructor( &den );
  Var_Destructor( &Y ); 

  return TRUE;
}
