#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <bits/nan.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include "PNMImageOffsetable.h"
#include "mat_util.h"
#include "boundingbox.h"
#include "motion.h"
#include "pchirp2ia.h"
/*#include <X11/Xlib.h>*/
#include "simple_png_read.c"

/* global variables -- this is from motion.c */
/*Pixmap my_drawing;*/
/* from display_service.c */
/*Window win;*/
int VERBOSE = 0;
int compose_pixels(int magic_number, int depth,  int image_number,
                   Image *src, Image *dst, Image *alpha_dst, 
                   Image *src_masked, int certaint_flag, int nanmask_flag);

typedef struct {
  int xSize;
  int ySize;
  int depth;
  int mOffset;
  int nOffset;
  int magic_number;
  int plm_field_1;
  float plm_field_2;
} PNMHeaderInfo ;

void PNMPrintHeaderInfo( PNMHeaderInfo *pPNM )
{
  fprintf(stderr, "magic number %d\n", pPNM->magic_number);
  fprintf(stderr, "xSize=%d  ySize=%d\n",pPNM->xSize, pPNM->ySize);
  fprintf(stderr, "depth=%d\n",pPNM->depth);
  fprintf(stderr, "mOffsets=%d nOffset=%d\n", pPNM->mOffset, pPNM->nOffset);
  fprintf(stderr, "Field 1=%d\n", pPNM->plm_field_1 );
  fprintf(stderr, "Field 2=%f\n", pPNM->plm_field_2 );
}

void PNMHeaderInfo_Copy( PNMHeaderInfo * pTarget, const PNMHeaderInfo * pSource) {
    pTarget->xSize = pSource->xSize;
    pTarget->ySize = pSource->ySize;
    pTarget->mOffset = pSource->mOffset;
    pTarget->nOffset = pSource->nOffset;
    pTarget->depth = pSource->depth;
    pTarget->magic_number = pSource->magic_number;
    pTarget->plm_field_1 = pSource->plm_field_1;
    pTarget->plm_field_2 = pSource->plm_field_2;
}

void usage() {
  printf( "usage: cement [output .ppm filename] [input .ppm or .png filenames...]\n"
          "Input PPMs must come from pchirp2nocrop!\n" 
          "Cement expects offsets to be in the image headers.\n" 
          "e.g. if offsets are not present in one or more inputs,.\n"
          "the offsets for said one or more inputs are set to zero.\n"
          "and a WARNING is given.\n"
          "This version supports certaint\n"
          "This version supports -overlay function"
          "This version supports -nanmask function\n");
}

int is_png_filename(char *fn){
  char *fn_ext=fn+strlen(fn)-4;
  if (fn_ext < fn) return 0; /* Would have caused a crash on a file with less than 4 chars in name - Pouya 13/01/2003 */

  if ((fn_ext[0]=='.') && ((fn_ext[1]=='p')||(fn_ext[1]=='P'))
                       && ((fn_ext[2]=='n')||(fn_ext[2]=='N'))
                       && ((fn_ext[3]=='g')||(fn_ext[3]=='G')))
  { return 1; }
  
  return 0;
}

int main(int argc, char **argv) {
  char * outFile, * fn; 
  PNMHeaderInfo * headers;
  int i, x, y, z;
  FILE * fp;
  PNMHeaderInfo curInfo;
  int bbMinX = INT_MAX, bbMinY = INT_MAX, bbMaxX = INT_MIN, bbMaxY = INT_MIN;
  int width, height, depth;
  PNMImageOffsetable ppmOut, ppmIn, alpha_ppmOut, ppm_masked;
  Image * src, * dst, *alpha_dst, *src_masked;
  int argv_offset = 0; /*to shift argv index by one for parsing*/
  int overlay_flag = 0, certaint_flag = 0, nanmask_flag = 0;
  int dstX,dstY;
  int is_png_in=0;
  int is_png_out=0;
  int min_mOffset = 0, min_nOffset = 0; /* these are for output image */
                                        /* so output image has offset comments*/
  int current_image_number;

  if (argc < 3) {
    usage();
    exit(1);
  }

  if( !strcmp(argv[1], "-overlay") )
    {
      fprintf(stderr, "Cementing in overlay mode\n");
      overlay_flag = 1;
      argv_offset = 1;
    }
  if( !strcmp(argv[1], "-nanmask") )
    {
      fprintf(stderr, "Cementing in NANmask mode\n");
      nanmask_flag = 1;
      argv_offset++;
    }
  if( !strcmp(argv[1], "-certainted") )
    {
      fprintf(stderr, "Cementing certainted images\n");
      certaint_flag = 1;
      argv_offset++;
    }

  outFile = argv[1+argv_offset];
  is_png_out=is_png_filename(outFile);
  headers = ( PNMHeaderInfo * ) malloc( sizeof( PNMHeaderInfo ) * (argc - 1) );
  if( headers == NULL ) {
    perror( "Out of memory creating headers" );
    return 1;
  }
  
  /* Parse all the headers */
  for (i = 2+argv_offset; i < argc; i++) {
    /* current_image_number=1 means the first input image is now being processed*/ 
    current_image_number = 1+argv_offset;

    fn = argv[i];
    fp = fopen(fn, "rb");
    if (fp == NULL) {
      fprintf( stderr, "cement: Error opening file %s: %s\n", fn, strerror( errno ) );
      return 1;
    }
 
    // check for png's here, read header info...   
    is_png_in=is_png_filename(fn);
    
    if (is_png_in) { 
      fclose(fp);
      printf("reading png header from '%s'\n", fn);
      simple_png_read(fn, &curInfo.depth, NULL,
          &curInfo.xSize,&curInfo.ySize,
          &curInfo.nOffset,&curInfo.mOffset); /* reversed; lefthand coords */
      switch(curInfo.depth) {
        case 1: curInfo.magic_number = 5; break;
        case 3: curInfo.magic_number = 6; break;
      }
      curInfo.plm_field_1=0;
      curInfo.plm_field_2=0;
    } else {
//    fprintf(stderr,"WARNING: no offset detected; ASSUMING ZERO\n");
//    fflush(stderr);
//    as of yet we have no way to check if input has offsets or not
        curInfo.mOffset = 0;  // default to zero offset
        curInfo.nOffset = 0;  // default
      if ( parseHeaderFrom( fp, &curInfo.xSize,
              &curInfo.ySize,
              &curInfo.depth,
              &curInfo.mOffset,
              &curInfo.nOffset,
              &curInfo.magic_number,
              &curInfo.plm_field_1,
              &curInfo.plm_field_2
              ) != PNM_SUCCESS ) {
        fprintf( stderr, "cement: Error parsing PNM header from file %s\n", fn );
        return 1;
      }
      fclose(fp);
    }

    PNMHeaderInfo_Copy( &headers[ i - (2+argv_offset)], &curInfo );
  } /* for( all headers ) */
   
  /* Now we have all the files' information; Perform a consistency check */
  /* also check for most upper-left offset, which is written to the cement */
  /* oufile for incremental cement */
  min_mOffset = headers[0].mOffset;
  min_nOffset = headers[0].nOffset;
  for (i = 0; i < argc - (2+argv_offset); i++) {
    if (headers[i].depth != headers[0].depth) {
      perror( "cement: All files must be of the same depth (8 or 24 bit)\n" );
      return 1;
    }
    if( headers[i].mOffset < min_mOffset ) {
      min_mOffset = headers[i].mOffset;
    }
    if( headers[i].nOffset < min_nOffset ) {
      min_nOffset = headers[i].nOffset;
    }
  }
  
  fprintf(stderr, "depth: %d\n", headers[0].depth);

  /* Compute the overall bounding box of the headers */
  for (i = 0; i < argc-2+argv_offset; i++) {
    /* Update notion of bounding box */
    if (headers[i].mOffset < bbMinY)
      bbMinY = headers[i].mOffset;
    if (headers[i].nOffset < bbMinX)
      bbMinX = headers[i].nOffset;
    if ((headers[i].mOffset + headers[i].ySize - 1) > bbMaxY)
      bbMaxY = (headers[i].mOffset + headers[i].ySize - 1);
    if ((headers[i].nOffset + headers[i].xSize - 1) > bbMaxX)
      bbMaxX = (headers[i].nOffset + headers[i].xSize - 1);
  }

  printf( "cement: BBox was (%d, %d) to (%d, %d)\n", bbMinX, bbMinY, bbMaxX, bbMaxY );

  if( overlay_flag == 1 || nanmask_flag == 1 ){
    /* if we are doing overlay only, we know output image size is just the 
       size of the first frame, because cement, in overlayonly mode, will
       not extend size of image...only overlay onto it...  */
    width = headers[0].xSize-1;
    height = headers[0].ySize-1;
    depth = headers[0].depth;
    bbMinX = 0;
    bbMinY = 0;
  }
  else {
    width = bbMaxX - bbMinX + 1;
    height = bbMaxY - bbMinY + 1;
    depth = headers[0].depth;
  }

  free( headers );
  headers = NULL;

  fprintf(stderr, "depth: %d width: %d height: %d\n", depth, width, height);

  if ( PNM_InitFromSize( &ppmOut, width, height, depth, curInfo.magic_number, curInfo.plm_field_1, curInfo.plm_field_2) != PNM_SUCCESS ) {
    perror( "cement: Error allocating output image" );
    return 1;
  }
  if( certaint_flag == 1 ) {
    if ( PNM_InitFromSize( &alpha_ppmOut, width, height, depth, 
                           curInfo.magic_number, curInfo.plm_field_1, 
                           curInfo.plm_field_2) != PNM_SUCCESS ) {
      perror( "cement: Error allocating output alpha image" );
      return 1;
    }
  }
 
  PNM_Pixel(&ppmOut, 0, 0);

  if (curInfo.magic_number < 9) {
    /* If anyone knows how to efficiently init an array 
     * with elements > a byte, this can be optimized. */
    if( certaint_flag != 1 )
      for (i=0; i<width*height*depth; i++) {
         ppmOut.data[i] = (Image)255;
      }
    else {
      fprintf(stderr, "initializing for certaint\n");
      for (i=0; i<width*height*depth; i++) {
         ppmOut.data[i] = (Image)0;
         alpha_ppmOut.data[i] = (Image)0;
      }
    }
  } else if(curInfo.magic_number == 12 ) {
      for (i=0; i<width*height*depth; i++) {
         alpha_ppmOut.data[i] = 0;
      } 
  }
  else {
    for (i=0; i<width*height*depth; i++)
       ppmOut.data[i] = (Image)NaNImage;
    /*memset( PNM_Pixel( &ppmOut, 0, 0), NaNImage, width * height * depth);*/
  }

  /* Now just iterate over all the input images, putting them into place */
  for (i = 2+argv_offset; i < argc; i++) {
    // check here for png's, read into struct...
    if (is_png_filename(argv[i])) {
      unsigned char *image_a=NULL,*s; // with alpha info...
      unsigned char r,g,b,a;
      int k; 
      Image *d;

      simple_png_read(argv[i], 
          &ppmIn.depth, &image_a,
          &ppmIn.xSize,&ppmIn.ySize,
       /* &ppmIn.mOffset,&ppmIn.nOffset); */
          &ppmIn.nOffset,&ppmIn.mOffset); /* png format is left hand system */
      switch(ppmIn.depth) {
        case 1: ppmIn.magic_number = 5; break;
        case 3: ppmIn.magic_number = 6; break;
      }
      ppmIn.plm_field_1 = 0;
      ppmIn.plm_field_2 = 0;

      // now copy from image_a to ppmIn.data as type Image
      // convert alpha=0 to NaN
      if( certaint_flag == 1)
        ppmIn.data = (Image *)malloc(ppmIn.xSize*ppmIn.ySize*(ppmIn.depth+1)
                            *sizeof(Image));
      else 
        ppmIn.data = (Image *)malloc(ppmIn.xSize*ppmIn.ySize*ppmIn.depth*sizeof(Image));
      s=image_a;
      d=ppmIn.data;
      if (ppmIn.depth==1) {
        printf("Converting Greyscale_Alpha format to Grey+NaN format.\n");
        for (k=0 ; k < ppmIn.xSize*ppmIn.ySize ; k++) {
         g=*s++;
         a=*s++;
         if( certaint_flag == 1 ) {
           *(d++) = g;
           *(d++) = a;
         }
         else {
           if (a==0) { // transparent
             *(d++) = NaNChar;
           } else {    // opaque
             *(d++) = g;
           }
        }
       }
      } 
      else if (ppmIn.depth==3) {
        printf("Converting RGB_Alpha format to RGB+NaN format.\n");
        for (k=0 ; k < ppmIn.xSize*ppmIn.ySize ; k++) {
         r=*s++;
         g=*s++;
         b=*s++;
         a=*s++;
         if( certaint_flag == 1 ) {
             *(d++) = (Image)r;
             *(d++) = (Image)g;
             *(d++) = (Image)b;
             *(d++) = (Image)a;
/*
 *           if(*(d-1)==0.0)
 *             fprintf(stdout,"alpha is cast to 0\n");
 */
         }
         else {
           if (a==0) { // transparent
             *(d++) = NaNChar;
             *(d++) = NaNChar;
             *(d++) = NaNChar;
           } else {    // opaque
             *(d++) = r;
             *(d++) = g;
             *(d++) = b;
           }
         }
        }
      } else {
        fprintf(stderr, "Unsupported number of PNG channels detected (must be 1 or 3).\n");
      }
      free(image_a);
    } 
    else if ( PNM_InitFromFile( &ppmIn, argv[i]) != PNM_SUCCESS ) {
      fprintf( stderr, "cement: Error reading input PNM %s\n", argv[i] );
      return 1;
    }

    if( nanmask_flag == 1 && i == 2+argv_offset ) {
        PNM_InitFromFile( &ppm_masked, argv[i]);
    }

    for (y = 0; y < PNM_GetYSize( &ppmIn ); y++) {
      for (x = 0; x < PNM_GetXSize( &ppmIn ); x++) {

        dstX = x + PNM_GetNOffset( &ppmIn );
        dstY = y + PNM_GetMOffset( &ppmIn ); 
        if( (overlay_flag==1 || nanmask_flag == 1) &&
            (dstX > width-1 || dstY > height-1 || dstX < 0 || dstY < 0) 
          ) {
          //fprintf(stderr, "rejected (%d,%d)\n", dstX, dstY);
          continue;
        }
 
//fprintf(stderr, "IMAGE DEPTH=%d\n", ppmIn.depth);

    /* ---------- Check out here -------------*/
        /* for certaint images, they have the alpha values in the image
           array, so make sure we offset accordingly by changing the
           depth value (used as offset multiplier in PNM_Pixel)  */
        src = 0; /* just to avoid compiler warning */
        alpha_dst = 0;
        if( certaint_flag == 1)  {
          if( ppmIn.depth == 1) {
            ppmIn.depth = 2;
            src = PNM_Pixel( &ppmIn, x, y );
            ppmIn.depth = 1;
          }
          if( ppmIn.depth == 3 ) {
            ppmIn.depth = 4;
            src = PNM_Pixel( &ppmIn, x, y );
            ppmIn.depth = 3;
          }
          /* dst pixel location is still the same though */
          dst = PNM_Pixel( &ppmOut, x + PNM_GetNOffset( &ppmIn ) - bbMinX,
                       y + PNM_GetMOffset( &ppmIn ) - bbMinY);
          /* but make a new one for the alpha image variable */
          alpha_dst = PNM_Pixel( &alpha_ppmOut, 
                           x + PNM_GetNOffset( &ppmIn ) - bbMinX,
                       y + PNM_GetMOffset( &ppmIn ) - bbMinY);
        }
        else {
          src = PNM_Pixel( &ppmIn, x, y);
          dst = PNM_Pixel( &ppmOut, x + PNM_GetNOffset( &ppmIn ) - bbMinX,
                       y + PNM_GetMOffset( &ppmIn ) - bbMinY);
          if( nanmask_flag == 1) {
            src_masked = PNM_Pixel(&ppm_masked, 
                           x + PNM_GetNOffset( &ppmIn ) - bbMinX, 
                           y + PNM_GetMOffset( &ppmIn ) - bbMinY);
          }
        }

        /* do consistency checking */
        assert(depth==1 || depth==3); if( certaint_flag == 1 )
        assert(alpha_dst!=NULL); if( nanmask_flag  == 1 )
        assert(src_masked!=NULL);

        /* create output cement PIC dst and alpha (see function comments)*/
        compose_pixels(curInfo.magic_number, curInfo.depth, current_image_number, 
                       src, dst, alpha_dst, src_masked, certaint_flag, nanmask_flag);       

      }
    }
    // maali this is done by PNM_Destroy() below?
    //if (is_png_filename(argv[i])) { free(ppmIn.data); }
  PNM_Destroy( &ppmIn );
  }


  /* for certainted cement images, normalize all final values */
  if( certaint_flag == 1)
  {
    for( y=0 ; y<ppmOut.ySize  ; y++ ) {
      for( x=0 ; x<ppmOut.xSize; x++ ) {
        dst = PNM_Pixel( &ppmOut, x, y);
        alpha_dst = PNM_Pixel( &alpha_ppmOut, x, y);
        if( alpha_dst[0] != 0 ) {
          for (z=0 ; z < ppmOut.depth ; z++ ) {
            dst[z] /= alpha_dst[0];
          }
        } else {
          for (z=0 ; z < ppmOut.depth ; z++ ) {
            if( curInfo.magic_number == 12 ) 
              dst[z] = 128;
            else 
              dst[z] = (double)NaNChar;
          }
        }
      }
    }
  }

  /* Write out the final PPM */
  ppmOut.mOffset = min_mOffset;
  ppmOut.nOffset = min_nOffset;
  if( curInfo.magic_number == 12 ) ppmOut.magic_number = 0xA;
  fprintf(stderr, "ppmOut->depth=%d\n",ppmOut.depth);
  fprintf(stderr, "ppmOut->magic_number=%d\n",ppmOut.magic_number);
  if ( PNM_WriteToFile( &ppmOut, outFile) != PNM_SUCCESS ) {
    fprintf( stderr, "cement: Error writing file %s\n", outFile );
    return 1;
  }
  PNM_Destroy( &ppmOut );
  //PNM_Destroy( &ppmIn );
  fprintf(stderr, "About to return\n");

  return 0;
}

/* LAWS OF COMPOSITION */
/* If the image read in is a pgm, then the NaN is NaNChar rather than
 * NaNImage (or else the pgm wouldn't display right). 
 */
/* cement runs by going scanning through each input image one at a time.  For
 * each location in the current input image (called double *src[]), it
 * determines where this location corresponds to in the output image (called
 * double *dst[]).  These  pointers were determined by calling PNM_Pixel() with
 * appropriate coordinates and offsets.  They are passed into the function
 * below.  Doing the sum image by image makes the program faster and 
 * less memory intensive, than loading all images into memory.
 */
/* So, dst is a location in the output image and it is used here like an
 * an accumulator, where values from each image are added to dst.  Thus the first 
 * image adds something to dst[], then later, some other overlapping image might
 * add something to dst[].  At the end, depending on what composition is used, by
 * dst[] pixels may need to be normalized.  
 */
/* in the simple case (pnms, no cert, no nanmasking) dst is 
 * overwritten by src (meaning the topmost image pixel ends up being dst,
 * and so the cement looks like the input images stacked
 * In the cases the src in actually NaN, dst is left untouched, so it is
 * left at its initialized value (pure white currently)
 */
/* in the case that certainties are present, then the src[] value is first
 * weighted by the certaint for that pixel.  Also, another certainty image
 * (called alpha_dst) is used to keep a sum of all the certainties.  This sum
 * of certainties (alpha_dst) is then used afterwards to normalize the dst
 * output pixel appropriated (like the division on the weighted average).  For
 * instance, if a straight averaging was wanted to compose the output pixel,
 * then each image src[] would add its value to dst[] and add 1 to alpha_dst[].
 * If three images contribute to dst[], then alpha_dst[] would sum up to 3,
 * then later on the final value of dst[] would be dst[]=dst[]/alpha_dst[]
 */
/* in the case that NaN masks are present, the NaN mask tells us which
 * locations we are to set to NaN.  (i.e. if the nanmask is 255 at a location
 * we consider that location NaN, and if nanmask != 255, then we process the
 * pixel normally)
 */
/* there are currently several different values for NaN, as defined in
 * the mat_util.h header. For plms, they are able to properly use the NAN
 * defined it /usr/include/bits/nan.h so some special care must be taken to
 * recognize each case.  (todo: make ALL NaN be ISO99 NAN defined by
 * /usr/include/bits/nan.h
 */
/* magic_number here is used to determine which type of NaN to use */
/* depth is the number of color channels.  Certainty is not included */
/* image_number is used to determine which image in the sequence this is and
 * can be used to preferentially decide which image to show (top or bottom) */
/* src[] is the double * to the current src pixel
 * dst[] is the double * to the current dst pixel
 * alpha_dst[] is the double * to the current alpha pixel (if needed else NULL)
 * src_masked[] is the double * to the NaN mask image being used
 *   (if needed else NULL)
 */
int compose_pixels(int magic_number, int depth,  int image_number,
                   Image *src, Image *dst, Image *alpha_dst, 
                   Image *src_masked, int certaint_flag, int nanmask_flag) 
{ 
  int z;

  /* do all channels.  3 (RGB) or 1 depending on depth */
  for ( z = 0; z < depth; z++) 
  { 
     /* PLMC case =>  Certainty usage implied true */
     if( magic_number == 12 ) {
       //if( !isnan(src[depth]) && !isnan(src[z]))  no such thing as NaN in PC (img type)
       //note src[depth] is the certainty channel e.g. src[3] 
          dst[z] += src[z] * src[depth];
       if( z==0 ) {
         //if( isnan(src[depth]) )   
         //  alpha_dst[0] = 0; 
         //else
           alpha_dst[0] += src[depth];
       }
     }
     /* certainty case */
     else if( certaint_flag == 1 && src[depth] != 0 ) { 
       dst[z] += src[z] * src[depth];
       /* only sum alpha ONCE, per pixel,
          so below, alpha is summed properly on last pass */
       if( z==0 )alpha_dst[0] += src[depth];
     }
     /* NANmask rule, mask imgA=argv[x] by imgB=argv[x+1] */
     else if ( nanmask_flag == 1 ) {
       /* Write NAN into final image whereve NAN is in mask image */
       if ( magic_number < 9 && src[depth-1] == NaNChar ) {
          dst[z] = NaNChar;
       }
       else if( magic_number >= 9 && src[depth-1] == NaNImage ) {
          dst[z] = NaNImage;
       }
       /* otherwise, let the first image show thru */
       else if( image_number == 1) { 
          dst[z] = src_masked[z];
       }
     }
     /* use correct type of NaN depending upon image input */
     else if (magic_number < 9 && src[depth-1] != NaNChar && src[depth-1] != 0.0) {
       dst[z] = src[z];
     }
     else if (magic_number >= 9 && src[z] != NaNImage) {
          dst[z] = src[z];
     }
  }  /* end for */
}
