#include <stdio.h>
#include <stdlib.h>
#include "corners2d.h"
#include "corners2r.h"
#include "mat_util.h"
#include "motion.h"
/*#include <X11/Xlib.h>*/

/* global variables -- this is from motion.c */
/*Pixmap my_drawing;*/
/* from display_service.c */
/*Window win;*/
int VERBOSE = 0;

int
main(int argc, char **argv)
{
  char *inFileName;
  char *outFileName;
  FILE *ifp, *ofp;
  int c1;
  int M, N;
  double srcCorners[8], destCorners[8];
  Var pDechirp, pRechirp, pFinal;
  
  if (argc != 3)
    {
      fprintf(stderr, "Usage: corr2p <infile> <outfile>\n");
      fprintf(stderr, "  where <infile> has the following format:\n");
      fprintf(stderr, "    M N m1 n1 i1 j1 m2 n2 i2 j2 m3 n3 i3 j3 m4 n4 i4 j4\n");
      fprintf(stderr, "    (m1, n1)...(m4, n4) are points (in pixels) in the\n");
      fprintf(stderr, "      original image\n");
      fprintf(stderr, "    (i1, j1)...(i4, j4) are the corresponding points in\n");
      fprintf(stderr, "      destination image\n");
      return 1;
    }

  inFileName = argv[1];
  outFileName = argv[2];

  if ((ifp = fopen(inFileName, "r")) == NULL)
    {
      fprintf(stderr, "Error opening file \"%s\" for input, aborting...\n",
	      inFileName);
      return 1;
    }

  if ((ofp = fopen(outFileName, "w")) == NULL)
    {
      fprintf(stderr, "Error opening file \"%s\" for output, aborting...\n",
	      outFileName);
      fclose(ifp);
      return 1;
    }

  /*
   * This is hard-coded for the above file format; i.e.,
   * a fixed number of input points (because corners2d and
   * corners2p deal only with four input points).
   */
  
  if (fscanf(ifp, "%d", &M) == EOF)
    {
      fprintf(stderr, "Error reading M from input file\n");
      return 1;
    }

  if (fscanf(ifp, "%d", &N) == EOF)
    {
      fprintf(stderr, "Error reading N from input file\n");
      return 1;
    }

  for (c1 = 0; c1 < 4; c1++)
    {
      int m, n, i, j;

      if (fscanf(ifp, "%d %d %d %d", &m, &n, &i, &j) == EOF)
	{
	  fprintf(stderr, "Error reading coordinate set %d\n", c1);
	  return 1;
	}
      
      if ((m < 0) || (m > M) ||
	  (i < 0) || (i > M) ||
	  (n < 0) || (n > N) ||
	  (j < 0) || (j > N))
	{
	  fprintf(stderr, "Error: an input point in set %d was out of bounds\n", c1);
	  return 1;
	}

      srcCorners[2*c1] = (double) m / (double) M;
      srcCorners[2*c1+1] = (double) n / (double) N;
      destCorners[2*c1] = (double) i / (double) M;
      destCorners[2*c1+1] = (double) j / (double) N;
    }

  init_var(&pDechirp, "dechirp parameters", 8, 1,
	   INTERNAL_VARIABLE, GREY_SCALE,NULL);
  init_var(&pRechirp, "rechirp parameters", 8, 1,
	   INTERNAL_VARIABLE, GREY_SCALE,NULL);
  init_var(&pFinal, "final parameters", 8, 1,
	   INTERNAL_VARIABLE, GREY_SCALE,NULL);

  /*
   * kbrussell's observations:
   * Intuitively I think they should be reversed (i.e.,
   * call:
   *   corners2d(dechirpParameters, srcCorners);
   *   corners2r(rechirpParameters, destCorners);
   * but I'll leave that up to the master...
   */

  corners2d(pDechirp.data, destCorners);
  corners2r(pRechirp.data, srcCorners);
  
  pcompose(&pDechirp, &pRechirp, &pFinal);

  for (c1 = 0; c1 < 8; c1++)
    {
      fprintf(ofp, "%f ", pFinal.data[c1]);
    }

  fclose(ifp);
  fclose(ofp);
  
  return 0;
}

