#include "util.h"
#include "mat_util.h"
#include "derivatives.h"
#include <string.h>

/* This function computes the derivatives of 2 pictures */
/* ================================================================== */
int derivatives(const Var * E1, const Var * E2,
		Var * Dt,   /* (returned) */
		Var * Dm,   /* (returned) */
		Var * Dn)   /* (returned) */
{
  const int M = E1->M;
  const int N = E1->N;     
  const int chnls = E1->channels; 
  char lookup_name[255];
  Var tmp1, tmp2;
  
  strcpy(lookup_name,E1->lookup_table_name); 
  if( (M != E2->M) || (N != E2->N) || (M==0) || (chnls != E2->channels))
   {
     fprintf( stderr, "  ERROR (derivatives.c): Images not the same size or depth.\n ending routine\n");
     return FALSE;
   }
  init_var( &tmp1, NULL, M - 1, N - 1, INTERNAL_VARIABLE, chnls,lookup_name );
  init_var( &tmp2, NULL, M - 1, N - 1, INTERNAL_VARIABLE, chnls,lookup_name );

  /* - - memory for data in Dt, Dm, Dn - - */
  init_var( Dt, "Dt", M - 1, N - 1, INTERNAL_VARIABLE, chnls, lookup_name );
  init_var( Dm, "Dm", M - 1, N - 1, INTERNAL_VARIABLE, chnls, lookup_name );
  init_var( Dn, "Dn", M - 1, N - 1, INTERNAL_VARIABLE, chnls, lookup_name );

#if DEBUG
  print_var(E1, 7, 7);
  print_var(E2, 7, 7);
#endif

  /* - - Time derivative of image matrices - - */
  sub_matrix(E1 , E2 , Dt);

  /* - - Downward derivative of matrices (Dy of E1 and E2) - - */
  sub_rows(E1, &tmp1); 
  sub_rows(E2, &tmp2);
  add_matrix(&tmp1, &tmp2, Dm);

  /* - - Across derivative of matrices (Dx of E1 and E2) - - */
  sub_cols(E1, &tmp1);
  sub_cols(E2, &tmp2);  
  add_matrix(&tmp1, &tmp2, Dn);

  /* - - free temporary Variable memory - - */
  Var_Destructor( &tmp1 );
  Var_Destructor( &tmp2 );
  return TRUE;
}











