/* to compile on sun   gcc -c -I/usr/openwin/include display_service.c      */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "util.h"
#include "mat_util.h"
#include "display_service.h"
#include "powLookup.h"


/* yes they are all globals! */
Screen *screen_point;      /* pointer to screen                      */
Display *display;          /* pointer to the display                 */
/*extern*/ Window win;        /* window to be used for graphics         */
int      screen_num;       /* screen number                          */
Window   root;             /* root window of display                 */
XSizeHints size_hints;     /* size hint structure for window manager */

GC the_GC;                 /* graphics context to be used            */
GC copyGC;                 /* main GC                                */ 
XGCValues the_GC_values;   /* values in graphics context             */
extern GC lineGC;
			   /* Removed FILE declarations              */
Pixmap buff;               /* Pixmap used as drawing buffer          */
XEvent event;              /* structure to hold an XEvent            */
Pixmap workingDrawing;


/* - - - NEW - - - */
XSetWindowAttributes win_attr;
Colormap cmap;
XColor scolor,ecolor, white, yellow;
Visual *vis,*def_vis;
Visual *visual;
XVisualInfo *p_vis_info, VInfoTmpl;
int ii;
XGCValues gc_attr;
GC gc2, clear_gc;
GC gc24;
int depth;

XImage *picture;

extern GC white_GC;


XImage * picTL, * picTR, * picLL, * picLR,*pic;
int borderwidth = 10;
/* - - - - - - - - */


/* - - - - New - - - - */
#define INIT_NAME "start_graphics"     /* name of startup module */

#define E_NOOPEN    -1		/* error--unable to open X connection */
#define E_NOFILE    -2		/* error--unable to open data file    */
#define E_NOEVENTS  -3		/* error--no events in file           */
#define BORDER 2		/* size of window border              */

#define TRUE  1
#define FALSE 0

/* LUT for displaying PLM files. */
static double inverseLookup[256];
static unsigned char binarySearch(double value);
static float lastexp = -1;


/* ======================================================================== */
void start_graphics(Display **p_display,   /* address of display pointer */
		    GC *p_context)         /* pointer to graphics context  */
{
  /* - - - Open display - - - */
  if((display = XOpenDisplay(NULL)) == NULL) 
    {
      printf("%s: can't open display",INIT_NAME);
      exit(E_NOOPEN);
    }
  
  /* - - - Set some useful globals - - - */
  screen_num = DefaultScreen(display);
  root = DefaultRootWindow(display);
  screen_point = XScreenOfDisplay(display,screen_num);

  /* - - - VISUAL STUFF - - - */
  VInfoTmpl.depth = 24;
  VInfoTmpl.screen = screen_num;

  /* note from shawn: using TrueColor rather than DirectColor */
#ifdef __cplusplus
  VInfoTmpl.c_class = TrueColor;
#else
  VInfoTmpl.class = TrueColor;
#endif

  
  p_vis_info = XGetVisualInfo(display, (VisualScreenMask | VisualDepthMask |
                                        VisualClassMask), &VInfoTmpl, &ii);

  /* - - - Check if 24 bit visual found - - - */
  if (ii < 1)
    {
      printf("  Could not get 24bit TrueColor Visual.\n");
      printf("  Trying 8 bit.");
      VInfoTmpl.depth = 8;
      p_vis_info = XGetVisualInfo(display,
				  (VisualScreenMask | VisualDepthMask |
				   VisualClassMask), &VInfoTmpl, &ii);      
      if(ii < 1)
	{
	  printf("  Could not get 8 bit TrueColor.\n  EXITING\n");
	  exit(1);
	}
      depth = 8;
    }
  else
    depth = 24;
 
  vis = p_vis_info->visual;
  def_vis = XDefaultVisual(display,screen_num);

  if (vis->visualid == def_vis->visualid)
   cmap = XDefaultColormap(display,screen_num);
  else
   {
     /* note from Shawn: use AllocNone not AllocAll for TrueColor Visuals */
     cmap = XCreateColormap(display, root, vis, AllocNone);
     if(cmap == DefaultColormap(display, screen_num))
       printf("All I got is the crummy DEFAULT CMAP!\n");
     fflush(stdout);
   }
  win_attr.colormap = cmap;
  win_attr.backing_store = Always;
  win_attr.backing_planes = AllPlanes;
  /* - - - - - - - - - - - - - - - - - */

  SetupGC();   /* this creates a GC that is the depth of the ROOT WINDOW !*/

  *p_display = display;
  *p_context = the_GC; 
}


/* ===================================================================== */
Pixmap create_display_panel(char * win_title,
			    int X, int Y, int width, int height)
{
  char * dataTL, * dataTR, * dataLL, * dataLR;
  Pixmap drawing;

  /* ====== Creat XImage structures to hold image data ===== */
  dataTL = (char *)calloc( (height * width), sizeof(unsigned long));
  picTL = XCreateImage(display, vis, 
		       depth,  
		       ZPixmap, 0,
		       dataTL,
		       width, height,
		       BitmapPad(display),
		       0);
  dataTR = (char *)calloc( (height * width), sizeof(unsigned long));
  picTR = XCreateImage(display, vis, 
		       depth,  
		       ZPixmap, 0,
		       dataTR,
		       width, height,
		       BitmapPad(display),
		       0);
  dataLL = (char *)calloc( (height * width), sizeof(unsigned long));
  picLL = XCreateImage(display, vis, 
		       depth,  
		       ZPixmap, 0,
		       dataLL,
		       width, height,
		       BitmapPad(display),
		       0);
  dataLR = (char *)calloc( (height * width), sizeof(unsigned long));
  picLR = XCreateImage(display, vis, 
		       depth,  
		       ZPixmap, 0,
		       dataLR,
		       width, height,
		       BitmapPad(display),
		       0);

  /* ==== create window ==== */
  drawing = create_buffered_window((width*2)+ borderwidth,
				   (height*2) + borderwidth,
				   win_title, X, Y);
  return drawing;
}
Pixmap create_display_panel_single(char * win_title,
				   int X, int Y, int width, int height)
{
  char * data;
  Pixmap drawing;

  /* ====== Creat XImage structures to hold image data ===== */
  data = (char *)calloc( (height * width), sizeof(unsigned long));
  pic = XCreateImage(display, vis, 
		     depth,  
		     ZPixmap, 0,
		     data,
		     width, height,
		     BitmapPad(display),
		     0);

  /* ==== create window ==== */
  drawing = create_buffered_window((width)+ borderwidth,
				   (height) + borderwidth,
				   win_title, X, Y);
  return drawing;
}

/* ===================================================================== */
int update_image(Var * Img, Pixmap drawing, int position)
{
  int x, y, index;
  unsigned long pixel_color, background_color, temp_pix_col;
  int is_a_plm = 0;  
  int exp;
  float exponent = Img->plm_field_2;
  char lutfilename[256];      /* name of the lut file */

  if( Img->var_type != IMAGE)
    {
      printf("  Can't display non-character images yet.  RETURNING\n");
      return FALSE;
    }

  if (Img->magic_number == 9 || Img->magic_number == 10) {
   //temporarily commented out to see log lightspace
    is_a_plm = 1;
    
    if ( exponent >= 0) {
      lastexp = exponent;      
      exp = (int)(rint(exponent*10.0));
      sprintf(lutfilename, "%s",Img->lookup_table_name); 
      readlut(lutfilename);
    }
  }

  background_color =  ((unsigned long)255 << 16) +
                       ((unsigned long)0 << 8) +
		       ((unsigned long)0); 
  
  /* - - Dump Img->image into XImage structure, picture - - */
  index = 0;
  
  for(y=0; y < Img->M; y++)
    {
      for(x=0; x < Img->N; x++, index++)
	{
	  if(Img->image[index] == NaNImage)
	    pixel_color = background_color;
	  
	  else if(is_a_plm) {
	    temp_pix_col = (unsigned long)binarySearch(Img->image[index]);
	    pixel_color = (temp_pix_col << 16) +
	                   (temp_pix_col << 8) +
			      (temp_pix_col);
	  }
	  else
	    pixel_color = ((unsigned long)Img->image[index] << 16) +
	                    ((unsigned long)Img->image[index] << 8) +
			      ((unsigned long)Img->image[index]); 

	  if(position == 1)
	    XPutPixel(picTL, x, y, pixel_color);
	  else if(position == 2)
	    XPutPixel(picTR, x, y, pixel_color);
	  else if(position == 3)
	    XPutPixel(picLL, x, y, pixel_color);
	  else if(position == 4)
	    XPutPixel(picLR, x, y, pixel_color);
	}
    }

  /* ==== create a 24 bit dummy GC to use as argument to XPutImage ==== */
  gc24 = XCreateGC(display, drawing, GCFunction, &the_GC_values);
  update_graphics();
  
  /* ==== display XImage in window ===== */
  if(position == 1)
    XPutImage(display, drawing, gc24, picTL, 
	      0, 0, 0, 0, Img->N, Img->M);
  else if(position == 2)
    XPutImage(display, drawing, gc24, picTR, 
	      0, 0, Img->N + borderwidth, 0, Img->N, Img->M);
  else if(position == 3)
    XPutImage(display, drawing, gc24, picLL, 
	      0, 0, 0, Img->M + borderwidth, Img->N, Img->M);
  else if(position == 4)
    XPutImage(display, drawing, gc24, picLR, 
	      0, 0, Img->N+borderwidth, Img->M+borderwidth, Img->N, Img->M);
  update_graphics();

  return TRUE;
}

/* ===================================================================== */
int update_image_single(Var * Img, Pixmap drawing)
{
  int x, y, index;
  unsigned long pixel_color, background_color, temp_pix_col;
  int is_a_plm = 0;
  int exp;
  float exponent = Img->plm_field_2;
  char lutfilename[50];      /* name of the lut file */

  if( Img->var_type != IMAGE)
    {
      printf("  Can't display non-character images yet.  RETURNING\n");
      return FALSE;
    }

  if (Img->magic_number == 9 || Img->magic_number == 10) {
    is_a_plm = 1;

    if ( exponent >= 0) {      
      lastexp = exponent;      
      exp = (int)(rint(exponent*10.0));
      sprintf(lutfilename, "%s",Img->lookup_table_name); 
      readlut(lutfilename);
    }
  }

  background_color =  ((unsigned long)255 << 16) +
                       ((unsigned long)0 << 8) +
		       ((unsigned long)0); 
  
  /* - - Dump Img->image into XImage structure, picture - - */
  index = 0;
  for(y=0; y < Img->M; y++)
    {
      for(x=0; x < Img->N; x++, index++)
	{
	  if(Img->image[index] == NaNImage)
	    pixel_color = background_color;
	  
	  else if(is_a_plm) {
	    temp_pix_col = (unsigned long)binarySearch(Img->image[index]);
	    pixel_color = (temp_pix_col << 16) +
	                    (temp_pix_col << 8) +
			      (temp_pix_col);
	  }
	  else
	    pixel_color = ((unsigned long)Img->image[index] << 16) +
	      ((unsigned long)Img->image[index] << 8) +
	      ((unsigned long)Img->image[index]);


	  XPutPixel(pic, x, y, pixel_color);
	}
    }
  /* ==== create a 24 bit dummy GC to use as argument to XPutImage ==== */
  gc24 = XCreateGC(display, drawing, GCFunction, &the_GC_values);
  update_graphics();
  
  /* ==== display XImage in window ===== */
  
    XPutImage(display, drawing, gc24, pic, 
	      0, 0, 0, 0, Img->N, Img->M);
  update_graphics();

  return TRUE;
}


/* ===================================================================== */
int display_image(Var * Img, char * win_title, int X, int Y)
{
  XImage * picture;
  Pixmap drawing;
  char * data;
  int index;
  int x, y;
  unsigned long pixel_color, background_color, temp_pix_col;
  int is_a_plm = 0;
  int exp;
  float exponent = Img->plm_field_2;
  char lutfilename[50];      /* name of the lut file */

  if( Img->var_type != IMAGE)
    {
      printf("  Can't display non-character images yet.  RETURNING\n");
      return FALSE;
    }

  if (Img->magic_number == 9 || Img->magic_number == 10) {
    is_a_plm = 1;
    
    if ( exponent >= 0) {      
      lastexp = exponent;      
      exp = (int)(rint(exponent*10.0));
      sprintf(lutfilename, "%s",Img->lookup_table_name); 
      readlut(lutfilename);
    }
  }

  background_color =  ((long)255 << 16) +
	                ((long)0 << 8) +
			  ((long)0); 

  /* - - Creat XImage structure to hold image data - - */
  data = (char *)calloc( (Img->M * Img->N), sizeof(unsigned long));
  picture = XCreateImage(display, vis, 
			 depth,  
                         ZPixmap, 0,
			 data,
			 Img->N, Img->M,
                         BitmapPad(display),
                         0);

  /* - - Dump Img->image into XImage structure, picture - - */
  index = 0;
  for(y=0; y < Img->M; y++)
    {
      for(x=0; x < Img->N; x++, index++)
	{
	  if(Img->image[index] == NaNImage)
	    pixel_color = background_color;
	  
	  else if(is_a_plm) {
	    temp_pix_col = (unsigned long)binarySearch(Img->image[index]);
	    pixel_color = (temp_pix_col << 16) +
	                    (temp_pix_col << 8) +
			      (temp_pix_col);
	  }
	  else
	    pixel_color = ((unsigned long)Img->image[index] << 16) +
	      ((unsigned long)Img->image[index] << 8) +
	      ((unsigned long)Img->image[index]);

	  XPutPixel(picture, x, y, pixel_color);
	}
    }


  /* - - create window - - */
  drawing = create_buffered_window(Img->N, Img->M,
				   win_title, X, Y);
  /* - - create a 24 bit dummy GC to use as argument to XPutImage - - */
  gc24 = XCreateGC(display, drawing, GCFunction, &the_GC_values);

  update_graphics();
  
  /* - - display XImage in window - - */
  XPutImage(display, drawing, gc24, picture, 
	    0, 0, 0, 0, Img->N, Img->M);
  update_graphics();

  return TRUE;
}


/* ===================================================================== */
Pixmap create_buffered_window(unsigned int width,
			      unsigned int height,
			      char * win_name,
			      int posX, int posY)  
{
  SetSizeHints(width,height);
  
  /* N.B. start_graphics must be called before this */
  win = XCreateWindow(display, root, posX, posY,
                      size_hints.width,
                      size_hints.height,
                      BORDER,
                      p_vis_info->depth,
                      InputOutput, vis,
                      CWColormap | CWEventMask |
                      CWBackingStore | CWBorderPixel,
                      &win_attr);

/*  win = XCreateSimpleWindow(display, root, posX, posY,
			    size_hints.width,
			    size_hints.height,
			    BORDER,
			    BlackPixel(display,screen_num),
			    WhitePixel(display,screen_num));

*/

  /* - - - set properties for window manager - - - */
/*  XSetStandardProperties(display, win, win_name, "current frames", None,
			 0,NULL,&size_hints); */
  XSetStandardProperties(display, win, win_name, "current frames", None,
			 NULL,0,&size_hints);
  
  /* - - - select events of interest and map window - - - */
  XSelectInput(display,win,
	       (ExposureMask | ButtonPressMask |
		VisibilityChangeMask | KeyPressMask));
  XMapWindow(display,win);  /* creates window */



  /* - - - create blanked out Pixmap buffer for user - - - */
  workingDrawing =  XCreatePixmap(display, win, width, height,
				  depth);  
  ClearPixmap(workingDrawing, width, height);
  buff = XCreatePixmap(display, win, width, height,
		       depth);
  ClearPixmap(buff, width, height);
  XSynchronize(display, TRUE); /* Allow other processes to share CPU */

  XNextEvent(display, &event);
  return(buff); 
}

void ClearPixmap(Pixmap p, int w, int h)
{
  GC gc;
  
  gc = XCreateGC(display, win, None, &the_GC_values);
  XSetGraphicsExposures(display, gc, False);
  XSetForeground(display, gc, WhitePixel(display, screen_num));
  XFillRectangle(display, p, gc, 0, 0, w,  h);
}

void update_graphics(void)
{
  XCopyArea(display, buff, win, gc24, 0, 0,
	    size_hints.width,size_hints.height, 0, 0);
  XFlush(display);

}

void SetSizeHints(int width, int height)
{
  size_hints.x = 200;   
  size_hints.y = 50;
  size_hints.min_width = width;
  size_hints.min_height = height;
  size_hints.max_width = width; 
  size_hints.max_height = height;
  size_hints.width = width;
  size_hints.height = height;
  size_hints.flags = USSize|USPosition|PMinSize|PMaxSize;
}

void SetupGC()
{
  the_GC_values.foreground = BlackPixel(display, screen_num);
  the_GC_values.function = GXcopy; /* GXequiv this is for black on white */
  the_GC_values.line_width = 4;
  the_GC_values.cap_style = CapRound;

  the_GC = XCreateGC(display, root,
		     (GCFunction | GCForeground | GCLineWidth | GCCapStyle),
		     &the_GC_values);

  copyGC = XCreateGC(display, root, None, &the_GC_values);


  XSetGraphicsExposures(display, the_GC, False);
  XSetGraphicsExposures(display, copyGC, False);
}

/*
bacon (e.g. had some trouble with order of arguments, might have been
wrong all along and never noticed on the hp735)
*/
/* find . -type f -print | xargs fgrep -i xsetstandard
./X11/Xutil.h:extern XSetStandardProperties(
./X11/Xutil.h:extern void XSetStandardColormap(
*/

void wait_for_any_key_in_Xwindow(Display *display/*, Window win*/)
{
  while(1) {
    XEvent event;

    XNextEvent( display, &event );
    if( event.type == KeyPress ) {
      char buffer_return[10];
      KeySym keysym_return;
      int len;

      for( len=0; len<10; len++ ) buffer_return[len] = '\0';
      len = XLookupString((XKeyEvent*)&event,
			  buffer_return, 10, 
			  &keysym_return, 
			  NULL );			    /* &status_in_out ); */
	  
      fprintf(stderr, "len:%d buff:%s keysym:%d \n", 
	      len, buffer_return, (int)keysym_return );

      return;
    }
  }
}

/* ===================================================================== */
/* ===================================================================== */
/* ===================================================================== */
/* This is used for converting plm images to pnms for display. This 
 * slows down the code significantly, so for fast operation, do not 
 * use the display option which plm images. 
 */

static unsigned char binarySearch(double value)
{
    int low=0;
    int high=0x100;
    int middle=0x7f;
    
    while(1)
    {
        if(high==middle || low==middle)
            return(middle);

        if(value>=inverseLookup[middle])
        {
            low=middle;
            middle=((high-low)>>1)+middle;
        }
        else
            if(value<inverseLookup[middle])
            {
                high=middle;
                middle=((high-low)>>1)+low;
            }
    }
    return(middle);
}



/* ===================================================================== */
/* ===================================================================== */
/* ===================================================================== */
/* Readlut - reads a given lookup table. */

void readlut(char *lutfilename)
{
    // some variables used in reading a text file lut
    FILE *lutfile;
    int lutsize=256;
    float p;
    int i;

    fprintf(stderr,"display_service.c: readlut: %s\n",lutfilename);
    
    if(lutfilename !=NULL && strlen(lutfilename) >0) 
    {
	/* if LUT filename is given */
        if ((lutfile = fopen(lutfilename, "r")) == NULL)
	{
            fprintf(stderr, "Unable to open %s.\n", lutfilename);
	    fclose(lutfile);
            exit(EXIT_FAILURE);
	} 
	else 
	{
	    fscanf(lutfile, "%f", &p);
	    for (i=0; i<lutsize; i++){
	      fscanf(lutfile, "%lf", &(inverseLookup[i]));
//	      fprintf(stderr,"%d-->%lf\n",i,inverseLookup[i]);
	    }
		   
	    if (i != lutsize) {
	      fprintf(stderr, "Incomplete read of LUT. Numread: %d\n", i);
	      fclose(lutfile);
	      exit(1);
	    }
	  }
	fclose(lutfile);
//	      fprintf(stderr,"----------\n",i,inverseLookup[i]);
    }
    else
    {
	/* if LUT filename is not given */
	for (i=0; i<lutsize; i++)
		inverseLookup[i] = powLookupTable[i];
    }
}
