/*
   not yet for distribution (copyright notice, etc. and all that...

   Video orbits, "seamless hockney", "painting with looks"...
   assembling images by treating them like pieces of a jigsaw puzzle
   with overlapping pieces.  Matlab/fmex translated to c
   Steve Mann, 1989-1995
   see also PV-wave function "smooth"
*/

#include "mat_util.h"
#include "filt2d.h"

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<string.h>

#define FALSE 0
#define TRUE 1

extern int VERBOSE;

void filt2d(Var * C, Var * image)
{
  int i;
  int row, col;
  int crow, ccol;
  float gsum;
  float ** gweights;
  Image ** xelbuff;
  Image * outputrow;
  float g;
  int crowso2, ccolso2;
  Image x;

  /* predefined values */
  float cmaxval = 0.0;      /* 18.0;*/
  float maxval = 255.0;
  Var * ANS;

  cmaxval = C->max_value;

  if(VERBOSE)
    {
      printf("Filtering %s . . . ", image->name);
      fflush(stdout);
    }

  /* - - Allocate space for answer - - */
  ANS = (Var *)calloc(1, sizeof(Var));
  init_var(ANS, image->name, image->M, image->N,
           IMAGE, GREY_SCALE,C->lookup_table_name);

  /* - - Allocate memory for weighting factors - - */
  gweights = (float **)calloc(C->M, sizeof(float *));
  for(i=0; i < C->M; i++)
    {
      gweights[i] = (float *)calloc(C->N, sizeof(float));
    }

  ccolso2 = C->N / 2;
  crowso2 = C->M / 2;
  
  /* - - Set up normalizing weights - - */
  gsum = 0.0;
  for ( crow = 0; crow < C->M; ++crow )
    for ( ccol = 0; ccol < C->N; ++ccol )
      {
	gsum += gweights[crow][ccol] =
	  ( (float)C->image[crow*C->N + ccol]  * 2.0 / cmaxval - 1.0 );
      }
  if(gsum < 0.9 || gsum > 1.1)
    {
      fprintf(stderr, "Biased  convolution matrix.  gsum:%f\n", gsum);
    }
  

  /* - - Allocate memory for 1 convolution matrix's worth of rows,
         plus a row output buffer - - */
  xelbuff = (Image **)calloc(C->M, sizeof(Image*));
  for(i=0; i < C->M; i++)
    {
      xelbuff[i] = (Image *)calloc(image->N, sizeof(Image));
    }
  outputrow = (Image *)calloc(image->N, sizeof(Image));

  /* - - convolution - -*/
  for (row = 0; row < (C->M - 1); ++row )
    {
      for(col = 0; col<image->N; ++col)
	xelbuff[row][col] = image->image[row*image->N + col];
      if( row < crowso2)
	{
	  /* write 1 output row */
	  for(col = 0; col < image->N; ++col)
	    ANS->image[row*image->N + col] = xelbuff[row][col];
	}
    }

  for( ; row < image->M; ++row)
    {
      for(col = 0; col<image->N; ++col)
	xelbuff[row % C->M][col] = image->image[row*image->N + col];
      
      for(col = 0; col<image->N; ++col)
	{
	  if( (col < ccolso2) || (col >= (image->N - ccolso2)) )
	    {
	      outputrow[col] = xelbuff[(row - crowso2) % C->M][col];
	    }
	  else
	    {
	      gsum = 0.0;
	      for ( crow = 0; crow < C->M; ++crow )
		for ( ccol = 0; ccol < C->N; ++ccol )
		  {
		    x = xelbuff[(row+1+crow) % C->M][col-ccolso2+ccol];
		    gsum += /*(((x) & 0xffc00) >> 10)*/ 
		      (float)x * gweights[crow][ccol];
		  }
	      if ( gsum < 0.0 ) g = 0.0;
	      else if ( gsum > maxval ) g = maxval;
	      else g = gsum + 0.5;
	      outputrow[col] = (Image)g;
	    }
	}
      for(col = 0; col < image->N; ++col)
	ANS->image[(row-1)*image->N + col] = outputrow[col];
    }

  /* - - write the remainder - - */
  for (; row < image->M + crowso2; ++row )
    {
      for(col = 0; col < image->N; ++col)
	ANS->image[(row-1)*image->N + col] =
	  xelbuff[(row - crowso2) % C->M][col];
    }


  /* - - Copy answer to original image - - */
  copy_image(ANS, image);
  
  /* - - Free OLD image - - */
  free(ANS->image);
  free(ANS);

  for(i=0; i<C->M; i++)
    free(xelbuff[i]);
  free(xelbuff);

  for(i=0; i<C->M; i++)
    free(gweights[i]);
  free(gweights);

  free(outputrow);

  if(VERBOSE)
    {
      printf("Done\n");
      fflush(stdout);
    }
}







