/*
   not yet for distribution (copyright notice, etc. and all that...

   Video orbits, "seamless hockney", "painting with looks"...
   assembling images by treating them like pieces of a jigsaw puzzle
   with overlapping pieces.  Matlab/fmex translated to c
   Steve Mann, 1989-1995
   see also PV-wave function "smooth"
*/

#include "mat_util.h"
#include "filt2ddsamp.h"

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<string.h>
#include<assert.h>
#include "basic.h"

#define FALSE 0
#define TRUE 1

extern int VERBOSE;

/*****filt2ddsamp******** 
*  Convolves filter C with image Image and downsamples by dx across and
*  dy wide.  Result is placed in Image 
*/

/* globals */
static int Crows, Ccols, ccolso2, crowso2;
static float gweightmult;
static float ** gweights;
extern int NANINPUT;

void filt2ddsampinit( const Var * C ) {
  int i, j, index = 0;

  Crows = C->M;
  Ccols = C->N;
  /* - - convolution - - */
  ccolso2 = Ccols / 2;
  crowso2 = Crows / 2;
  gweightmult = 2.0 / C->max_value;
  /* - - Allocate memory for weighting factors - - */
  /* - - in such a way so that you index them [row][col] - - */
  /* - - Set up normalizing weights - - */
  /* - - gweights is what we convolve with - - */
  gweights = (float **)malloc(Crows * sizeof(float *));
  for(i=0; i < Crows; i++) {
    gweights[i] = (float *)malloc(Ccols * sizeof(float));
    for ( j = 0; j < Ccols; ++j, index++ )
	gweights[i][j] = (float)C->image[index] * gweightmult - 1.0;
  }   
} 

void filt2ddsampdone() {
  int i;
  for(i=0; i<Crows; i++)
    free(gweights[i]);
  free(gweights);
}

void filt2ddsamp( Var * Img, int dx, int dy) {
  int xelbuffrow, i ,j,k,l,Imgi,Imgj;
  const int Irows = Img->M,
            Icols = Img->N,
            ANSrows = Irows / dy,
            ANScols = Icols / dx;
  Image * curANSPixel;
  Var ANS;

  if(VERBOSE)
    {
      printf("Filtering %s . . . ", Img->name);
      fflush(stdout);
    }
  /* - - Allocate space for answer - - */
  init_var(&ANS, Img->name, ANSrows, ANScols, IMAGE, GREY_SCALE,Img->lookup_table_name);
  
  /* load up xelbuff with proper rows */
  curANSPixel = ANS.image;
  for( i=Imgi=0 ; i<ANSrows && Imgi < crowso2 ; i++,Imgi+=dy ) {
    for( j=Imgj=0 ; j<ANScols ; j++,Imgj+=dx ) {
      /* replace border cases with the same pixel values */
      *(curANSPixel++) = Img->image[Imgi*Icols+Imgj]; 
    }
  }
  for( ; Imgi < Irows - crowso2; i++,Imgi+=dy ) {
     for( j=Imgj=0 ; j<ANScols && Imgj < ccolso2; j++,Imgj+=dx ) {
       /* replace border cases with the same pixel values */
       *(curANSPixel++) = Img->image[Imgi*Icols+Imgj]; 
     }
     for( ; Imgj < Icols - ccolso2; Imgj+=dx,j++ ) {
       for( k=-crowso2 ; k<=crowso2 ; k++ ) { /* for each row in filter */
         xelbuffrow = (Imgi + k )*Icols; /*calc row offset to op on */
         for( l=-ccolso2 ; l<=ccolso2 ; l++ ) { /*each col in filt row */

	   /* This shouldn't be floor. */ 
          if( Img->image[xelbuffrow+Imgj+l] > (NaNChar-1) && NANINPUT ) {
            *(curANSPixel) = NaNChar;
			goto funkygoto;
          } else
          //*(curANSPixel) += floor(gweights[k+crowso2][l+ccolso2] * 
          *(curANSPixel) += gweights[k+crowso2][l+ccolso2] * 
                             Img->image[xelbuffrow+Imgj+l];
//	   fprintf(stderr, "orig = %g ==> new = %g\n",  Img->image[Imgi*ANScols + Imgj],*curANSPixel);

         } /* for each col in filt row */
       } /* for each row in filter */
funkygoto:
       curANSPixel++;
     } /* for cols */
     for( ; j<ANScols; j++,Imgj+=dx ) 
      *(curANSPixel++) = Img->image[Imgi*Icols+Imgj]; 
  } /* for rows */
  
  for( ; i<ANSrows ; i++,Imgi+=dy ) 
    for( j=Imgj=0 ; j<ANScols ; j++,Imgj+=dx ) 
      /* replace border cases with the same pixel values */
      *(curANSPixel++) = Img->image[Imgi*Icols+Imgj]; 

  free(Img->image);
  Img->image = ANS.image;
  Img->M = ANSrows;
  Img->N = ANScols;

  //#ifdef DEBUG
  if(VERBOSE) 
    {
      printf("Done\n");
      fflush(stdout);
    }
  //#endif
}
