/*
   early initial distribution

   Video orbits, "painting with looks"...
   assembling images by treating them like pieces of a jigsaw puzzle
   with overlapping pieces.                    Steve Mann, 1989-1995
   http://wearcam.org/orbits
   (from matlab/fmex/cmex, etc)
*/


/* to compile on a Sun  gcc -c -I/usr/openwin/include main.c      */

#include "util.h"
#include "mat_util.h"
#include "display_service.h"  /* Include and prototype for X11 service */
#include "ppm.h"              /* protos for ppm handlers */
#include "est_pseudo.h"
#include "pchirp2ia.h"
#include "motion.h"
#include "filt2d.h"
#include "par.h"   
#include "filt2ddsamp.h"

#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<string.h>
#include<sys/time.h>
#include<unistd.h>
#include<sys/stat.h>
#include<getopt.h>
#include<ctype.h>
/* the above 2 are used bo stat() to see if downfilter.pgm present */

/*
   this is rough development code, not meant for distribution
   technology transfer is best done by looking at the paper, TR338, rather
   than using this code

   steve@media.mit.edu
   http://wearcam.org/orbits/orbits.html 
 */

/* - - extern globals shared with X11 service routines - - */
Display *my_display;
GC lineGC;
extern int borderwidth;

/* ---- toggle switches ----- */
int DETECT = FALSE;
int IS_MOVIE = FALSE;
int IS_FRAMES = FALSE;
int COLOR_INPUT = FALSE;
int COLOR_CHAN = 1; /* DEFAULTS TO GREEN */
int USE_STEPS = FALSE;
int USE_DISPLAY = FALSE;
int USE_FILTER = FALSE;
int OUT_MSE = FALSE;
int OUT_PAR = FALSE;
int VERBOSE = FALSE;
int RECORD_HIST = FALSE;
int HOLD_AT_END = FALSE;
int GOT_LOOKUP_TABLE = FALSE;
int ExpectASceneChange = FALSE;
/* used in motion.c */
extern int NANINPUT;

FILE * par_file_ptr = NULL;
FILE * diff_file_ptr = NULL;

/* - - Prototypes - - */
void process_frame(Var * frameA,double scale_factor);
double estimate_k(Var * frameA, Var * frameB);
void posixify(int argc, char * argv[]);
void create_steps(Var **, Var *, int);
void downsample(Var *, int, int,int);
void get_next_image(char *, int, Var *);
void detect_cut(double, double, int);
void get_next_frame(FILE *, Var *, int, int);
FILE * load_dat_movie(int, char *, int, int);
void write_variable(Var *, double *, FILE *, FILE *);
void print_usage(char *argv);
int load_image(char *, Var *);
void get_green_channel(Var * image);

struct timeval tp;
int starttime,endtime;

int main(int argc, char * argv[])
{
  /***********************************/
  /***         Variables           ***/
  /***********************************/
  /* ---- Accessory ---- */
  int i, j, filecount = 0;
  int level = 0;
  int * iterations=NULL;
  int steps = 1;
  int SERIES_NEEDED = 2;
  char c;
  char lookup_table[255];
  Pixmap my_drawing;
  struct stat filestatbuf;

  /* ---- Movie related ---- */
  int movieWidth, movieHeight;
  FILE * movie_file = NULL;

  /* ---- Image related ---- */
  char * imageNameA = NULL, * imageNameB = NULL;
  char * nextimage;
  char * filterName = NULL;
  int filterWidth = 3, filterHeight = 3;
  int StartFrameNum = 0, EndFrameNum = 1;
  int imagenum = 0;
  double est_k;

  /* ---- image buffers ---- */
  Var ** frameA, ** frameB;
  Var * filter = NULL;
  Var * down_filter = NULL;

  /* ---- estimation statistics ----- */
  double * diff_index;           /* an array of MSE values */
  Var * parameters = NULL;       /* Nth parameters returned by est_pchirp2r  
                                    - the "orbit" params.  Var is defined in mat_util.h 
                                    - it is a named N*M matrix 
                                    - 'parameters' will be a 1*8 column vector 
                                 */
  char * diff_file = NULL, * par_file = NULL;
  double DiffThresh = 1500.0;
  EstimatorParameters *estParams;
  estParams = (EstimatorParameters *) calloc(1, sizeof(EstimatorParameters));

  /***********************************/
  /****    Parse command line     ****/
  /***********************************/
  /* changed to getopt by Corey 8/19/2002 */
  opterr = 0; /* prevent weird options from causing errors */

  posixify(argc,argv);
  while(1) {
     int this_option_optind = optind ? optind : 1;
     int option_index = 0;
     static struct option long_options[] = {
                   {"pout", 0, 0, 'P'},
                   {"outMSE", 1, 0, 'M' },
                   {"outp", 1, 0, 'O'},
                   {"display", 0, 0, 'd'},
                   {"filter", 1, 0, 'f'},
                   {"verbose", 0, 0, 'v'},
                   {"detect", 0, 0, 'D'},
                   {"threshold", 1, 0, 't'},
                   {"lut", 1, 0, 'l'},
                   {"hold", 0, 0, 'H'},
                   {"help", 0, 0, 'h'},
                   {"rechist", 1, 0, 'r'},
                   {"nanmask", 0, 0, 'n'},
                   {"notranslation", 0, 0, 'T'},
                   {"nozoom", 0, 0, 'Z'},
                   {"constrainedzoom", 0, 0, 'C'},
                   {"constrainedrotation", 0, 0, 'R'},
                   {"noshear", 0, 0, 'S'},
                   {"nochirp", 0, 0, 'c'},
                   {"steps", 1, 0, 0},
                   {"pin", 1, 0, 0},
                   {"movie", 1, 0, 0},
                   {"frames", 1, 0, 0},
                   {"inp", 1, 0, 'p'},
                   {0, 0, 0, 0}
     };
     c = getopt_long (argc, argv, "p:PM:l:O:df:vDt:hHnTZCRS",
                      long_options, &option_index);
     if (c == -1) break;
     
     switch(c){
     case 'P':
	fprintf(stderr,"using pout\n");
        OUT_PAR = TRUE;
        par_file = (char *)calloc(200, sizeof(char));
        par_file = strcpy(par_file, "stdout");
        break;
     case 'M':
        OUT_MSE = TRUE;
        diff_file = (char *)calloc(200, sizeof(char));
        diff_file = strcpy(diff_file, optarg);
        break;
     case 'O':
        OUT_PAR = TRUE;
        par_file = (char *)calloc(200, sizeof(char));
        par_file = strcpy(par_file, optarg);
        break;
     case 'd':
        fprintf(stderr,"using display\n");
        USE_DISPLAY = TRUE;
        DETECT = TRUE;
        break;
     case 'f':
        filterName = (char *)calloc(200, sizeof(char));
        filterName = strcpy(filterName, optarg);
        USE_FILTER = TRUE;
        break;
     case 'v':
        VERBOSE = TRUE;
        break;
     case 'D':
        DETECT = TRUE;
        break;
     case 't':
        DiffThresh = atoi(optarg);
        break;
     case 'H':
        HOLD_AT_END = TRUE;
        break;
     case 'h':
        print_usage(argv[0]);
        exit(0);
        break;
     case 'r':
        RECORD_HIST = TRUE;
        ExpectASceneChange = atoi(optarg);
        break;
     case 'n':
        fprintf(stderr, "Treating 255 as NaN\n");
        NANINPUT = TRUE;
        break;
     case 'T':
        estParams->NO_TRANSLATION = 1;
        break;
     case 'Z':
        estParams->NO_ZOOM = 1;
        break;
     case 'C':
        if (estParams->ONLY_ROTATION){
           printf("Error: -constrainedzoom and -constrainedrotation are ");
           printf("mutually exclusive.\n");
           exit(1);
        }
        estParams->YES_ISOTROPIC = 1; 
        break;
     case 'R':
        if (estParams->YES_ISOTROPIC) {
             printf("Error: -constrainedzoom and -constrainedrotation are ");
             printf("mutually exclusive.\n");
             exit(1);
        }
        estParams->ONLY_ROTATION = 1;
        break;
     case 'S':
        estParams->NO_SHEAR = 1;
        break;
     case 'c':
        estParams->NO_CHIRP = 1;
        break;
     case 'l':
        strcpy(lookup_table,optarg); 
        printf("using lookup_table %s\n",lookup_table);
	GOT_LOOKUP_TABLE = TRUE;
        break;
     case 'p':
        {
          char *filename = optarg;
          FILE *fp;
  
          if ((fp = fopen(filename, "r")) == NULL){
             printf("Error opening input param file\n");
             exit(1);
          }
  
          parameters = (Var *) calloc(1, sizeof(Var));
          init_var(parameters, "parameters", 1, 8, INTERNAL_VARIABLE, 1,NULL);
          for (j = 0; j < 8; j++){
             if (fscanf(fp, "%lf", &parameters->data[j]) == EOF){
                printf("Premature EOF reading input param file\n");
                if( fp != NULL ) fclose(fp);
                exit(1);
             }
          }
          if( fp != NULL ) fclose(fp);
        }
        break;

     case 0:

        /* check for 'steps' option */
        if (strcmp(long_options[option_index].name,"steps")==0){
           steps = atoi(optarg);
           if(steps < 0) steps = -steps;
           iterations = (int *)calloc(steps, sizeof(int));

           for(j=0; j < steps; j++){
           /* the following, and all similar arg-checking bits,
              should display an error and terminate if the
              if-clause fails */
              if(optind < argc && strchr(argv[optind],'-')==NULL)
                    iterations[j] = atoi(argv[optind]);
              else {
                 fprintf( stderr, "argc and argv don't match!\n" );
                 exit(1);
              }
              optind++;
           }
           USE_STEPS = TRUE;
        }

        /* check for 'pin' */
        if (strcmp(long_options[option_index].name,"pin")==0){
           /* - - Create space for parameters - - */
           parameters = (Var *)calloc(1, sizeof(Var));
           init_var(parameters, "parameters", 1, 8, INTERNAL_VARIABLE, 1,NULL);
          /* i.e. create a 1*8 column vector of signed floating-point values 
             (INTERNAL_VARIABLE) named "parameters" to store the orbit parameters 
           */
          /* the next line is for optarg conformity */
	  if (argv[0][0]=='n') argv[0][0]='-';
          parameters->data[0] = atof(optarg);
          for(j=1; j < 8; j++) {
             if(optind < argc ) {
		  if (argv[optind][0]=='n'){
			  fprintf(stderr,"found an n in %s\n",argv[optind]);
			  argv[optind][0]='-';
		  }
                  parameters->data[j] = atof(argv[optind]);
             }
             else{
                fprintf( stderr, "not enought input parameters given!\n" );
                exit(1);
             }
             optind++;
          }
        }

        /* check for 'movie' */
        if (strcmp(long_options[option_index].name,"movie")==0){
           
	   movieWidth = atoi(optarg);
	   if(optind < argc && strchr(argv[optind],'-')==NULL) 
             movieHeight = atoi(argv[optind]);
            
           if(++optind < argc && strchr(argv[optind],'-')==NULL)
	      StartFrameNum = atoi(argv[optind]);
           if(++optind < argc && strchr(argv[optind],'-')==NULL)
	      EndFrameNum = atoi(argv[optind]);
           optind++;

      	   SERIES_NEEDED = 1;
	   IS_MOVIE = TRUE;
        }

        /* check for 'frames' */
        if (strcmp(long_options[option_index].name,"frames")==0){

           movieWidth = atoi(optarg);
           if(optind < argc && strchr(argv[optind],'-')==NULL)
              movieHeight = atoi(argv[optind]);


           if(++optind < argc && strchr(argv[optind],'-')==NULL)
              StartFrameNum = atoi(argv[optind]);
           if(++optind < argc && strchr(argv[optind],'-')==NULL)
              EndFrameNum = atoi(argv[optind]);
           optind++;

           SERIES_NEEDED = 1;
           IS_FRAMES = TRUE;
        }

        break;

      } /* - end switch - */

  } /* - end parse command line - */

  /* at this point, all command line options should be parsed,
     what needed now is to extract the two files which should 
     be at optind
   */
  if (!GOT_LOOKUP_TABLE)
  { 
      printf("No lookup table given. Using DEFAULT\n");
      lookup_table[0]='\0'; 

      
  }

  while (filecount < SERIES_NEEDED && optind < argc) {
     if(filecount == 0){
        imageNameA = (char *) calloc(200, sizeof(char));
        imageNameA  = strcpy(imageNameA, argv[optind]);
        optind++;
     }
    else{
	imageNameB = (char *) calloc(200, sizeof(char));
	imageNameB  = strcpy(imageNameB, argv[optind]);
        optind++;
    }
    filecount++;
  }  
  fprintf(stderr,"input file 1 = %s input file 2 = %s\n",imageNameA,imageNameB);
  /****************************************/
  /***** Check for bad input arguments ****/
  /****************************************/
  /* - - if too few files print usage - - */
  
  if(filecount <= 1)
    {
      print_usage(argv[0]);
      exit(0);
    }
  /* - - if no steps specified default - - */
  if(iterations == NULL)
    {
      iterations = (int *)calloc(1, sizeof(int));
      iterations[0] = 2;
      steps = 1;
    }
  /* - - Check frame range - - */
  if( (StartFrameNum > EndFrameNum) || StartFrameNum < 0 || EndFrameNum < 0)
    {
      fprintf(stderr, "ERROR: Bad frame range specified.\n");
      exit(1);
    }
  
  if( DETECT == TRUE || OUT_MSE == TRUE ) 
    diff_index = (double *)malloc( 10 * sizeof( double ) );
  else
    diff_index = NULL;
  /* - - Get dimensions of images from a file - - */
  if(!IS_MOVIE && !IS_FRAMES)
    get_image_dimensions(imageNameA, &movieWidth, &movieHeight); 

  /* - - open output files if user requests - - */
  if(OUT_PAR == TRUE)
    {
      if (!strcmp(par_file,"stdout"))
        {
          par_file_ptr = stdout;
        }
      else
        {
          par_file_ptr = fopen(par_file, "w");
        }
      if(par_file_ptr == NULL)
        printf("\n\nWARNING: Could not open parameters file for writing.\n\n");
    }

  if(OUT_MSE == TRUE)
    {
      if (!strcmp(diff_file,"stdout"))
        {
          diff_file_ptr = stdout;
        }
      else
        {
          diff_file_ptr = fopen(diff_file, "w");
        }
      if(diff_file_ptr == NULL)
        printf("\n\nWARNING: Could not open diff file for writing.\n\n");
    }

  /***************************************/
  /**** Allocate memory for variables  ****/
  /***************************************/
  

  if(USE_FILTER)
    {
      filter = (Var *)calloc(1, sizeof(Var));
      get_image_dimensions(filterName, &filterWidth, &filterHeight);     
      init_var(filter, "filter", filterWidth, filterHeight,
	       IMAGE, GREY_SCALE,NULL);
    }
  if(USE_STEPS)
    {
      down_filter = (Var *)calloc(1, sizeof(Var));
      init_var(down_filter, "down filter", 3, 3,
	       IMAGE, GREY_SCALE,NULL);
	/* down_filter = a 3*3 matrix of images used to build
		the shrunken images used in the initial steps */
    }

/* We are about to create an array of [steps] versions
	of both the first and the second input images */

  frameA = (Var **)calloc(steps, sizeof(Var*));
  frameB = (Var **)calloc(steps, sizeof(Var*));
 
  for(i=0; i < steps; i++)
    {
      frameA[i] = (Var *)calloc(1, sizeof(Var));
      frameB[i] = (Var *)calloc(1, sizeof(Var));
      init_var(frameA[i], "frame A",
	       movieHeight, movieWidth,
	       IMAGE, GREY_SCALE,lookup_table);
      init_var(frameB[i], "frame B",
	       movieHeight, movieWidth,
	       IMAGE, GREY_SCALE,lookup_table);
    }

  /************************************/
  /****  Load various image types  ****/
  /************************************/
  if(IS_FRAMES)
    {
      imagenum = StartFrameNum;
      nextimage = (char *)calloc(200, sizeof(char));
      sprintf(nextimage, "%s%04d.pgm", imageNameA, imagenum);
   
      load_image(nextimage, frameA[steps-1]);
      imagenum++;

      sprintf(nextimage, "%s%04d.pgm", imageNameA, imagenum);
      load_image(nextimage, frameB[steps-1]);
      imagenum++;
    }
  else if(IS_MOVIE)
    {
      movie_file = load_dat_movie(StartFrameNum, imageNameA,
				  movieWidth, movieHeight);
      get_next_frame(movie_file, frameA[steps-1], movieWidth, movieHeight);
      get_next_frame(movie_file, frameB[steps-1], movieWidth, movieHeight);
    }
  else
    {
 	/* the input image itself is always the last to be processed, 
		so it is stored in the last (steps-1) position in the
		Frame{A,B} array - the rest of the array will be filled,
		if necessary, with downfiltered images. */
      load_image(imageNameA, frameA[steps-1]);
      load_image(imageNameB, frameB[steps-1]);

      if(frameA[steps-1]->channels == RGB24)
	get_green_channel(frameA[steps-1]);

      if(frameB[steps-1]->channels == RGB24)
	get_green_channel(frameB[steps-1]);     

      if (COLOR_INPUT)
	{
	  for (i=0; i<(movieHeight*movieWidth); i++)
	    {
	      frameA[steps-1]->image[i] = frameA[steps-1]->image[3*i+COLOR_CHAN];
	      frameB[steps-1]->image[i] = frameB[steps-1]->image[3*i+COLOR_CHAN];
	      realloc (frameA[steps-1]->image,movieHeight*movieWidth);
	      realloc (frameB[steps-1]->image,movieHeight*movieWidth);
	      frameA[steps-1]->channels=GREY_SCALE;
	      frameB[steps-1]->channels=GREY_SCALE;


	    }
	}

    }
  
 

  /**************************************/
  /**** filter and downsample images ****/ 
  /**************************************/
  if(USE_STEPS)
    {
/* usage should indicate that 'downfilter.pgm' must be accessible */
/* check to make sure 'downfilter.pgm' is in the current working directory */
      stat("downfilter.pgm", &filestatbuf);
      if (!S_ISREG(filestatbuf.st_mode)) {
/* if downfilter.pgm is not present, then substitute it with default value */
	 int downfilter_index;
	      
/* these values are already defined in init_var, but just to make sure they have the right values */
	 down_filter->M = 3;
	 down_filter->N = 3;
	 down_filter->max_value = 180;
	 down_filter->channels = GREY_SCALE;
	 down_filter->var_type = IMAGE;

/* now fill-in down filter with default values */	 
	 down_filter->image = (Image *) calloc(3*3*GREY_SCALE, sizeof(Image));
	 for (downfilter_index = 0; downfilter_index < 3*3*GREY_SCALE; downfilter_index++)
	 {
	    down_filter->image[downfilter_index] = (Image)100;
	 }
      }
      else {
/* if downfilter.pgm presents, then load it from current working directory */
         load_image("downfilter.pgm", down_filter);
      }
      create_steps(frameA, down_filter, steps);
      create_steps(frameB, down_filter, steps);
    }
  if (frameA[level]->magic_number == 10){
      est_k=estimate_k(frameA[steps-1],frameB[steps-1]);
      fprintf(stderr, "estimated value of k = %g\n",est_k);
      for (i = 0 ; i< steps;i++){
         process_frame(frameA[i],1);
         process_frame(frameB[i],1);
      }
  }
  /*****************************/
  /**** filter first images ****/
  /*****************************/
  if(USE_FILTER)
    {
      load_image(filterName, filter);
      for(j=0; j < steps; j++)
	{
	  filt2d(filter, frameA[j]);
	  filt2d(filter, frameB[j]);
	}
    }
  
  /************************/
  /**** Start graphics ****/
  /************************/
  if(USE_DISPLAY == TRUE)
    {
      /* - - Initialize the X Window - - */
      start_graphics(&my_display, &lineGC);

      /* - - Create a display window large enough for the last image - - */ 
      my_drawing = create_display_panel("Current frames",
/*					400, 400, */
					0, 0,
					frameA[steps-1]->N, frameA[steps-1]->M);
    }
 
  /***********************************/
  /****  Run estimator on frames  ****/
  /***********************************/
  if(IS_FRAMES || IS_MOVIE)
    {
      for(i=StartFrameNum; i <= EndFrameNum; )  /* step through frames */
	{
          if(VERBOSE)
            printf("Comparing %d and %d\n", imagenum-2, imagenum-1 );
	  for(level = 0; level < steps; level++)
	    {
	      if(VERBOSE)
		printf("Starting Step %d . . . \n", level+1);
	      if(USE_DISPLAY)
		ClearPixmap(my_drawing, 2*movieWidth+borderwidth,
			    2*movieHeight+borderwidth);
	       
	      parameters = find_motion_parameters(frameA[level], frameB[level],
						  parameters,
						  0, iterations[level],
						  diff_index,
						  USE_DISPLAY, VERBOSE, i,
						  estParams,
						  my_drawing );
              if( DETECT == TRUE ) {
  	        detect_cut(*diff_index, DiffThresh, i);
              }
	    }
	  write_variable(parameters, diff_index, par_file_ptr, diff_file_ptr);
	  if(HOLD_AT_END)
	    {
	      fprintf(stderr, "\n PRESS RETURN TO CONTINUE\n\n");
	      fscanf(stdin, "%c", &c);
	    }
	  if(++i < EndFrameNum)
	    {
	      ++imagenum;
	      if(IS_FRAMES)
		get_next_image(imageNameA, imagenum-1, frameA[steps-1]);
	      else if(IS_MOVIE)
		get_next_frame(movie_file, frameA[steps-1], movieWidth, movieHeight);

	      if(USE_STEPS)
		{
		  create_steps(frameA, down_filter, steps);
		}
	      if(USE_FILTER)
		{
		  for(j=0; j<steps; j++)
		    filt2d(filter, frameA[j]);
		}
	      /* - - reset parameters for next frame pair - - */
              Var_Destructor( parameters );
	      parameters = NULL;
	    }
	  else
	    break;


          if(VERBOSE)
            printf("Comparing %d and %d\n", imagenum-2, imagenum-1 );
	  for(level = 0; level < steps; level++)
	    {
	      if(VERBOSE)
		printf("Starting Step %d . . . \n", level+1);
	      if(USE_DISPLAY)
		ClearPixmap(my_drawing, 2*movieWidth+borderwidth,
			    2*movieHeight+borderwidth);
	      parameters = find_motion_parameters(frameB[level], frameA[level],
						  parameters,
						  0, iterations[level],
						  diff_index,
						  USE_DISPLAY, VERBOSE, i,
						  estParams,
					          my_drawing );
              if( DETECT == TRUE ) 
  	        detect_cut(*diff_index, DiffThresh, i);
	    }
	  write_variable(parameters, diff_index, par_file_ptr, diff_file_ptr);
	  if(HOLD_AT_END)
	    {
	      fprintf(stderr, "\n PRESS RETURN TO CONTINUE\n\n");
	      fscanf(stdin, "%c", &c);
	    }
	  if(++i < EndFrameNum)
	    {
	      ++imagenum;
	      if(IS_FRAMES)
		get_next_image(imageNameA, imagenum-1, frameB[steps-1]);
	      if(IS_MOVIE)
		get_next_frame(movie_file, frameB[steps-1], movieWidth, movieHeight);

	      if(USE_STEPS)
		{
		  create_steps(frameB, down_filter, steps);	      
		}
	      if(USE_FILTER)
		{
		  for(j=0; j<steps; j++)
		    filt2d(filter, frameB[j]);
		}
	      /* - - reset parameters for next frame pair - - */
	      parameters = NULL;
	    } 
          else
            break;
	}  
    }
  else
    {
      for(level = 0; level < steps; level++)
	{
	  if (!iterations[level])
	    {
	      if (VERBOSE) 
		printf ("Skipping step %d\n",level+1);
	      continue;
	    }
	  if(VERBOSE)
	    printf("Starting Step %d . . . \n", level+1);
	  if(USE_DISPLAY)
	    ClearPixmap(my_drawing, 2*movieWidth+borderwidth,
			2*movieHeight+borderwidth);

	  parameters = find_motion_parameters(frameA[level],
					      frameB[level],
					      parameters,
					      0, iterations[level],
					      diff_index,
					      USE_DISPLAY,
					      VERBOSE, 0,
					      estParams,
 					      my_drawing );


	  if( DETECT == TRUE ) {
            if (*diff_index > DiffThresh ) { 
  	      fprintf(stdout,
	  	      "main.c: images in DIFFERENT ORBIT; MSE = %f\n",
		      *diff_index);
            } else {
	      fprintf(stdout,
                      "main.c: images in SAME ORBIT; MSE = %f\n",
                      *diff_index);
            }
            /* the above makes it no longer necessary to do:
               bash$ estpchirp2m 2>&1 | sed
               or
               bash$ (estpchirp2m > /dev/null) 2>&1 | sed ...

               now all you need to do is:
               estpchirp2m -steps 4 8 4 2 1 porka.pgm porkb.pgm -display | sed s/"main.c\: images in SAME ORBIT\; MSE ="//
            */

          } 

	  /*********************************************************************/

	  write_variable(parameters, diff_index, par_file_ptr, diff_file_ptr);
	  
	}
    }

  /* - - Free variables and images - - */
  for(i=0; i < steps; i++)
    {
      free(frameA[i]->image);
      free(frameB[i]->image);
    }
  free(frameA);   free(frameB);
  free( diff_index );
  if( par_file_ptr != NULL ) fclose(par_file_ptr);
  if( diff_file_ptr != NULL ) fclose(diff_file_ptr);
  if(HOLD_AT_END)
    {
      printf("\nHOLD: Press any key in Xwindow to end.\n\n");
      wait_for_any_key_in_Xwindow(my_display/*,win*/);
    }
  return 0;
}



/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
void create_steps(Var ** frameA, Var * down_filter, int steps)
{
  int i = 0;
  int magic_number = frameA[steps-1]->magic_number;
  float plm_field_2 = frameA[steps-1]->plm_field_2;

  filt2ddsampinit( down_filter );
  for(i=steps-1; i >= 1; i--)
    {
      if(frameA[i-1]->image != NULL)
	{
	  free(frameA[i-1]->image);
	}
      init_var(frameA[i-1], "frame A",
	       frameA[i]->M, frameA[i]->N,
	       IMAGE, GREY_SCALE,frameA[i]->lookup_table_name);

      frameA[i-1]->magic_number = magic_number;
      frameA[i-1]->plm_field_2 = plm_field_2;

      copy_image(frameA[i], frameA[i-1]);

      /* - - filtering and downsampling - - */
      filt2ddsamp( frameA[i-1],2, 2); 
    }

  filt2ddsampdone();
}



/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
void downsample(Var * image, int dX, int dY,int VERBOSE)
{
  int x,y,i,xx,yy;
  int subM,subN;
  Image * newImage;

  subM = image->M / dY;
  subN = image->N / dX;

  /* - - Create space to hold downsampled image - - */
  newImage = (Image *)calloc(subM*subN, sizeof(Image)); 
  i = 0;
  for(yy=y=0; y < image->M && yy < subM; y = y + dY,yy++)
    {
      for(xx=x=0; x < image->N && xx < subN; x = x + dX,xx++)
	{
	  newImage[i] = image->image[y*image->N + x];
	  i++;
	}
    }
  
  /* - - free old image memory and point to new memory - - */
  free(image->image);
  image->image = newImage;
  
  /* - - Change dimentions of image - - */
  image->M = image->M / dY;
  image->N = image->N / dX;
  
  if (VERBOSE)
    printf("New Image Size M: %d, N: %d\n", image->M, image->N);
 
}


/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
void detect_cut(double diff_index, double DiffThresh, int i)
{
  if(diff_index > DiffThresh)
    {
      fprintf(stderr,
	      "Scene change between frame %d and %d. MSE:%f\n\n",
	      i, i+1,  diff_index);
    }
  
}


/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
void get_next_image(char * imageNameA, int imagenum, Var * frameA)
{
  char nextimage[200];
  sprintf(nextimage, "%s%04d.pgm", imageNameA, imagenum);
printf("getting image %s\n", nextimage );
  load_image(nextimage, frameA);
}  


/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
int load_image(char * filenameA, Var * frameA)
{
  int ret;

/* Both this function and load_ppm_file should display the filename in any
	error output.  Also, there is no point in this function having
	a return value if it's just going to terminate the program
	on failure */

  ret = load_ppm_file(filenameA, frameA);
  
  /* - - if good increment num_got - - */
  if(ret >= 0)
    {
      return(1);
    }
  else
    {
      printf("\n Unable to load image: %s\nExiting\n\n", filenameA);
      free(frameA->image);
      exit(1);
    }
  return(0);
}


/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
FILE * load_dat_movie(int i, char * moviedir, int width, int height)
{
  FILE * file_ptr;
  char * datafilename;
  char * datafilesuffix = "data";

  datafilename = (char *)calloc(250, sizeof(char));
  sprintf(datafilename, "%s/%s", moviedir, datafilesuffix);

  file_ptr = fopen(datafilename, "r");
  if(file_ptr == NULL)
    {
      printf("ERROR: could not open movie file.\n EXIT(1)\n"); 
      exit(1);
    }
  else 
    {
      if(COLOR_INPUT)
	fseek(file_ptr, 3*i*width*height, SEEK_SET);
      else
	fseek(file_ptr, i*width*height, SEEK_SET);
      return(file_ptr);
    }
}


/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
void get_next_frame(FILE * current_file_location, Var * frame, 
	       int movieWidth, int movieHeight)
{
  int i;
  int pork; /* ignore red */
  int bacon; /* ignore blue */

  if(COLOR_INPUT)
    {
      fprintf(stderr,"colorin so looking at green channel only\n");
      for(i = 0; i< (movieWidth*movieHeight); i++)
	{
	  /* fseek(current_file_location, 1, SEEK_CUR);*/
	  fread(&pork, 1, 1, current_file_location);  /* ignore red*/
	  fread(&frame->image[i], 1, 1, current_file_location);
	  /* fseek(current_file_location, 2, SEEK_CUR);*/
	  fread(&bacon, 1, 1, current_file_location); /* ignore blue*/
	}
    }
  else
    fread(frame->image, 1, (movieWidth*movieHeight), current_file_location);

}
	       

/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
void write_variable(Var * parameters, double * diff_index,
		    FILE * par_file, FILE * diff_file)
{
  int i;
  
  if(OUT_PAR)
    {
      for(i = 0; i < 8; i++)
	{
	  fprintf(par_file, "%f ", parameters->data[i]);
	}
      fprintf(par_file, "\n"); /* , parameters->data[i]); */
    }
  if(OUT_MSE)
    {
      fprintf(diff_file, "%f ", diff_index[0]);
      fprintf(diff_file, "\n"); 
    }
}



/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
/* Extracts the green channel from an RGB image, so that estpchirp2m can     */
/* work on that channel alone.                                               */
/* Assumes that image has 3 channels.                                        */
/* ========================================================================= */
void get_green_channel(Var * image)
{
  int i, j;
  Image * green = (Image *)malloc(image->N*image->M*sizeof(Image));
  if(green == NULL)
    fprintf(stderr, "ERROR (main.c): could not allocate memory in get_green_channel.\n");

  if(image->channels != 3)
    fprintf(stderr, "ERROR (main.c): image must have 3 channels.\n");

  for(i=1, j=0; i < image->N*image->M*3; i+=3, j++)
    green[j] = image->image[i];

  /* Can get rid of the original image. */
  free(image->image);
  image->image = green;
  image->channels = GREY_SCALE;
}


/* ========================================================================= */
/* ========================================================================= */
/* ========================================================================= */
void print_usage(char *progname)
{
  char* image_required = "[inputA.pgm] [inputB.pgm]";

  char* movie_required = "[dat dir] -movie [width height startframe endframe]";

  char * frames_movie = "[prefix] -frames [width height startframe endframe]";

  printf("%s - estimates the 8-parameters (pchirp)\n",progname);
  printf("useful for ``seamless hockney'' (combining multiple images of\n");
  printf("the same scene by treating them like a jigsaw puzzle with\n");
  printf("overlapping pieces) and also for orbit change detection\n");
  printf("similar to detection of cuts or scene changes\n\n");
  printf("Steve Mann, 1989 to 1994 (translated from Matlab/fmex, etc.)\n");
  printf("see http://www-white.media.mit.edu/~steve/orbits/\n\n");
  printf("usage: (image pgm files)\n   %s %s [options]\n",
	 progname, image_required);
  printf("usage: (multiple pgm files)\n   %s %s [options]\n",
	 progname, frames_movie);
  printf("usage: (DAT raw data file)\n   %s %s [options]\n",
	 progname, movie_required);

  printf("\noptions:\n");
  printf("   --steps <N> <A1 A2 . . . An>\n");
  printf("                    - Specifies the # of steps and # of\n");
  printf("                      iterations to run at each step\n");
  printf("   --display         - Display images in a window.\n");
  printf("   --verbose         - Print numerical results to STDOUT.\n");
  printf("   --hold            - Holds images on screen after final estimation.\n");
  printf("   --filter [pgm]    - Prefilter images before estimation.\n");
  printf("   --colorin [r,g,b] - Specifies color input.  Uses channel specified.\n");
  printf("   --detect          - Signal scene changes. (Default threshold 1500)\n");
  printf("   --threshold [N]   - Set threshold for scene change detection.\n");
  printf("   --pin [p1..p8]    - Specify initial parameters on command line.\n");
  printf("   --inp [file]      - Read initial parameters from specified text file.\n");
  printf("   --outp [file]     - Write parameters to specified text file.\n");
  printf("   --pout            - Write parameters to standard output.\n");
  printf("   --outMSE [file]   - Write MSE results from frame differencing.\n");
  printf("   --notranslation   - Do not allow the images to shift positions\n");
  printf("                      translation between the input images \n");
  printf("   --nozoom          - Do not allow the scale of the input images\n");
  printf("                      to vary\n");
  printf("   --constrainedzoom - Equalize the X and Y components of the zoom\n");
  printf("                      factor between images\n");
  printf("   --constrainedrotation - Constrains the estimator to only use\n");
  printf("                      pure rotations (zoom X and Y factors are\n");
  printf("                      equalized, and shear is regularized)\n");
  printf("    *** NOTE that -constrainedzoom and -constrainedrotation are\n");
  printf("    *** mutually exclusive. (-constrainedzoom is a subset of\n");
  printf("    *** -constrainedrotation)\n");
  printf("   --noshear         - No shear allowed between two images\n");
  printf("   --nochirp         - No perspective correction done\n");
  printf("   --nanmask         - ignores to NaN pixels that are 255\n");
  printf("\n");
}

/* process frame is used to pre-process frames in a consistent manner.
   process_frame multiplies every element of an image by scale_value
 */
void process_frame(Var * frameA,double scale_factor){
   double max_val=0,max_val2=0;
   double min_val=1000000000,min_val2=1000000000;
   double best_max, best_min;
   int i,j;

   for (i= 0;i<frameA->M;i++)
      for (j = 0; j<frameA->N;j++){
         if (frameA->image[i*frameA->N+j] > max_val) 
		 max_val =frameA->image[i*frameA->N+j];	
         if (frameA->image[i*frameA->N+j] < min_val) 
		 min_val =frameA->image[i*frameA->N+j];	
	 frameA->image[i*frameA->N+j]= 
	         255+(scale_factor*(frameA->image[i*frameA->N+j] ));
       }
	   
     fprintf(stderr,"max val = %g \n",max_val);
     fprintf(stderr,"min val = %g \n",min_val);
     fprintf(stderr,"log(max val) = %g \n",log(max_val));
     fprintf(stderr,"log(min val) = %g \n",log(min_val));
  //   fprintf(stderr,"bringing values into the range 0 to 255\n");
     /* the following code normalizes thet log lightspace values to fall between 0 and 255 */
  /*
     for (i= 0;i<frameA[level]->M;i++)
        for (j = 0; j<frameA[level]->N;j++)
	   frameA[level]->image[i*frameA[level]->N+j]= 255/(best_max-best_min)*((frameA[level]->image[i*frameA[level]->N+j]+.01)-best_min);
	   //frameA[level]->image[i*frameA[level]->N+j]= ((frameA[level]->image[i*frameA[level]->N+j]+.01)-best_min);
        
     for (i= 0;i<frameB[level]->M;i++)
        for (j = 0; j<frameB[level]->N;j++)
           frameB[level]->image[i*frameB[level]->N+j]= 255/(best_max-best_min)*((frameA[level]->image[i*frameB[level]->N+j]+.01)-best_max);
           //frameB[level]->image[i*frameB[level]->N+j]= ((frameA[level]->image[i*frameB[level]->N+j]+.01)-best_max);
   */
  }

double estimate_k(Var * frameA, Var * frameB){
   double sum1=0,sum2=0;
   int i,j;
   for (i= 0;i<frameA->M;i++)
      for (j = 0; j<frameA->N;j++){
         sum1+=frameA->image[i*(frameA->N)+j];
      }
   for (i= 0;i<frameB->M;i++)
      for (j = 0; j<frameB->N;j++)
         sum2+=frameB->image[i*frameB->N+j];
   return sum1/sum2;
}
void posixify(int argc, char * argv[]){
   int i;
   char * newstring;
   for (i = 1;i<argc;i++){
      if(argv[i][0] == '-' && isdigit(argv[i][1]))
	  argv[i][0] = 'n';
      if(strlen(argv[i])>2 && argv[i][0]=='-' && argv[i][1] !='-'){
         //fprintf(stderr,"changing %s to -%s\n",argv[i],argv[i]);
         newstring = (char *)malloc(strlen(argv[i])*sizeof(char)+2*sizeof(char));
         sprintf(newstring,"-%s",argv[i]);
         argv[i] = strdup(newstring);
      }
   }
}

