/*  
   not yet for distribution (copyright notice, etc. and all that...

   Video orbits, "seamless hockney", "painting with looks"...
   assembling images by treating them like pieces of a jigsaw puzzle
   with overlapping pieces.  Matlab/fmex translated to c
   Steve Mann, 1989-1995

*/

/* interatively estimates the parameters of the phirp2 motion group */
/* to compile on a Sun: gcc -c -I/usr/openwin/include motion.c      */

#include "mat_util.h"
#include "est_pseudo.h"
#include "display_service.h"
#include "motion.h"
#include "pchirp2ia.h"
#include <math.h>

#define MEAN_SQR_ERROR 0
#define CONTINUITY 1

/*this is define in main.c. It is a toggle switch*/
int NANINPUT;


/* - - going to E1 from E2 - - */
/* - - this version returns an array of parameters - - */
Var * find_motion_parameters(Var * E1,             /* image A */
			     Var * E2,             /* image B */
			     Var * start_par,      /* starting parameters */
			     int DETECT_TYPE,      /* switch */
			     int iterations,
			     double * diff_index,  /* returned */
			     int USE_DISPLAY,      /* switch */
			     int USE_TEXT,         /* switch */
			     int frame_num,
			     EstimatorParameters *estParams,
                             Pixmap my_drawing )
{
  Var * parN;
  Var * p_comp;
  Var * parNprev;
  Var * ImageOut;
  Var * ImageIn;
  Var * diff_image;
  int iter,j,i;
  int magic_number = E1->magic_number; 
  
 /* temporary experiment */
  

  
  fprintf(stderr,"magic number = %d\n",magic_number);
  /* *************************************************** */
  /* ***  Squash input images to free up 255 for NaN *** */
  /* *************************************************** */
  if( !NANINPUT) {
    fprintf(stderr, "motion.c: squashing the image\n");
    chval(E1, NaNImage, NaNImage-1);
    chval(E2, NaNImage, NaNImage-1);
  //  chval(E1, NaNChar, NaNImage-1);  // wasn't doing anything anyway
  //  chval(E2, NaNChar, NaNImage-1);  // because now it's 512 not 255
  }
  else {
    chval(E1, NaNChar, NaNImage); // changes 255 to 512 (the new NaN)
    chval(E2, NaNChar, NaNImage);
  }

  if(iterations < 1)
    iterations = 1;

  /* ****************************************** */
  /* **** check that images have same size **** */
  /* ****************************************** */
  if( (E1->M != E2->M) || (E1->N != E2->N) ) {
      printf("  ERROR: motion.c\n  Images must be same size.\n");  
      return(0);
    }

  /* ***************************************************************** */
  /* *** initialize matrix that will hold in, out, and diff images *** */
  /* ***************************************************************** */
  ImageOut = (Var *)malloc( sizeof(Var));
  ImageIn = (Var *)malloc( sizeof(Var));
  diff_image = (Var *)malloc( sizeof(Var));
  
  init_var(ImageOut, "working OUTPUT", E1->M, E1->N, IMAGE, GREY_SCALE,E1->lookup_table_name);
  init_var(ImageIn, "working INPUT", E1->M, E1->N, IMAGE, GREY_SCALE,E1->lookup_table_name);
  init_var(diff_image, "difference image", E1->M, E1->N, IMAGE, GREY_SCALE,E1->lookup_table_name);
  diff_image->data = (Data *)calloc( E1->M *  E1->N, sizeof(Data));

  ImageOut->magic_number = magic_number;
  ImageIn->magic_number = magic_number;
  diff_image->magic_number = magic_number;

  /* ************************************************ */
  /* *** display reference and comparison: E1, E2 *** */
  /* ************************************************ */
  if(USE_DISPLAY) {
    update_image(E1, my_drawing, 1);
    update_image(E2, my_drawing, 2);
  }

  /****************************************/
  /**** get starting pchirp parameters ****/
  /****************************************/
  parN = (Var *)malloc( sizeof(Var));
  init_var(parN, "Nth parameters", 8, 1, 
           INTERNAL_VARIABLE, GREY_SCALE,E1->lookup_table_name);

  p_comp = (Var *)malloc(sizeof(Var));
  init_var(p_comp, "composite parameters", 8, 1, 
           INTERNAL_VARIABLE, GREY_SCALE,E1->lookup_table_name);

  parNprev = (Var *)malloc( sizeof(Var));
  init_var(parNprev, "previous parameters", 8, 1, 
           INTERNAL_VARIABLE, GREY_SCALE,E1->lookup_table_name);

  /* ===== use the supplied starting parameters if possible ==== */
  if((start_par == NULL) || (start_par->data == NULL)) {
    pseudo_estimate( p_comp, E2, E1 );
  } else {
      copy_data(start_par, p_comp);
      fprintf(stderr,"motion.c: using the supplied parameters\n");
  }

  /* Regularize parameters according to the user's preferences */
  if (estParams->NO_TRANSLATION)
    p_comp->data[2] = p_comp->data[5] = 0.0;

  if (estParams->NO_ZOOM)
    p_comp->data[0] = p_comp->data[4] = 1.0;

  if (estParams->ONLY_ROTATION)
    {
      /* estimating a theta for the given transformation matrix,
	 to turn it into a pure rotation matrix */
      double theta = ((p_comp->data[1] - p_comp->data[3]) /
		      (p_comp->data[0] - p_comp->data[4]));
      /* a11 */
      p_comp->data[0] = cos(theta);
      /* a12 */
      p_comp->data[1] = -sin(theta);
      /* a21 */
      p_comp->data[3] = sin(theta);
      /* a22 */
      p_comp->data[4] = cos(theta);
    }

  if (estParams->YES_ISOTROPIC)
    {
      double normZoom = sqrt(p_comp->data[0] * p_comp->data[4]);
      p_comp->data[0] = p_comp->data[4] = normZoom;
    }

  if (estParams->NO_SHEAR)
    {
      p_comp->data[1] = p_comp->data[3] = 0.0;
    }

  if (estParams->NO_CHIRP)
    {
      p_comp->data[6] = p_comp->data[7] = 0.0;
    }



  /* ********************************************************** */
  /* ***  do default first iteration apply parameters to E1 *** */
  /* ********************************************************** */
  pchirp2(p_comp, E2, ImageIn, 1, -1, E1->M, E1->N, 1); 
  if(USE_DISPLAY)
    {
      update_image(ImageIn, my_drawing, 4);
    }

  /* ******************************************** */
  /* *** Calculate difference for first image *** */
  /* ******************************************** */
  if( diff_index != NULL ) 
    diff_index[0] = calculate_difference(E1, ImageIn,
				         diff_image, 
				         MEAN_SQR_ERROR, USE_DISPLAY);
  if(USE_DISPLAY)
    update_image(diff_image, my_drawing, 3);


  /* **************************************** */  
  /* *** display results of 1st iteration *** */
  /* **************************************** */  
  if(USE_TEXT)
    {
      printf("\nTransition: %d     Iteration: %s\n", frame_num+1, "FIRST");
      print(p_comp, 1, 1, 8,1);
      /* lf -> f ww 20 Aug 1999 */
      if( diff_index != NULL )
        printf("diff_index: %f\n", diff_index[0]);
      printf("-------------------------------------------------\n");
    }

  /* *************************** */
  /* *** main iteration loop *** */
  /* *************************** */
  copy_image(ImageIn, ImageOut);
       
  for(iter = 1; iter < iterations;  iter++)
    {
      if(USE_TEXT)
	printf("Transition: %d     Iteration: %d\n", frame_num+1, iter+1);


      
      /* ===== Calculate p_comp parameters from parN ans prev p_comp ===== */ 
      copy_data(p_comp, parNprev);
      clear_data(parN);

      pseudo_estimate( parN, ImageOut, E1 );
  
      clear_data(p_comp);

      /* Now we modify parN according to the estimator parameters. */
      if (estParams->NO_TRANSLATION)
	{
	  parN->data[2] = parN->data[5] = 0.0;
	}

      if (estParams->NO_ZOOM)
	{
	  parN->data[0] = parN->data[4] = 1.0;
	}

      if (estParams->ONLY_ROTATION)
	{
	  /* estimating a theta for the given transformation matrix,
	     to turn it into a pure rotation matrix */
	  double theta = ((parN->data[1] - parN->data[3]) /
			  (parN->data[0] - parN->data[4]));
	  /* a11 */
	  parN->data[0] = cos(theta);
	  /* a12 */
	  parN->data[1] = -sin(theta);
	  /* a21 */
	  parN->data[3] = sin(theta);
	  /* a22 */
	  parN->data[4] = cos(theta);
	}

      if (estParams->YES_ISOTROPIC)
	{
	  double normZoom = sqrt(parN->data[0] * parN->data[4]);
	  parN->data[0] = parN->data[4] = normZoom;
	}

      if (estParams->NO_SHEAR)
	{
	  parN->data[1] = parN->data[3] = 0.0;
	}

      if (estParams->NO_CHIRP)
	{
	  parN->data[6] = parN->data[7] = 0.0;
	}

      pcompose(parN, parNprev, p_comp);
      if(USE_TEXT)
	print(p_comp, 1, 1, 8,1);

      /* ===== Apply composite parameters to E1n ===== */
      pchirp2(p_comp, E2, ImageOut, 1, -1, E1->M, E1->N, 1); 
      
      /* ===== Display image ===== */
      if(USE_DISPLAY)
	{
	  update_image(ImageOut, my_drawing, 4);
	}
      
      /* ===== difference calculation ===== */
      if( diff_index != NULL ) {
        diff_index[0] = calculate_difference(E1, ImageOut,
					      diff_image, 
					      MEAN_SQR_ERROR, USE_DISPLAY);
        if(USE_TEXT)
	  printf("diff_index: %f\n", diff_index[0]);
      }
      
      /* ===== Display diff_image if desired ===== */
      if(USE_DISPLAY)
	{
	  update_image(diff_image, my_drawing, 3);
	}
      if(USE_TEXT)
	printf("-------------------------------------------------\n");
    } /* for */

  Var_Destructor( parN ); free(parN);
  Var_Destructor( parNprev ); free(parNprev);
  Var_Destructor( diff_image ); free( diff_image );
  Var_Destructor( ImageOut ); free(ImageOut);
  Var_Destructor( ImageIn );   free(ImageIn);
  return(p_comp);
} 

/* ==================================================================== */
void pcompose(Var * A, Var * B, Var * Pout)
{
  /* - - realloc mem for 1 more element in A and B and Pout- - */
  A->data = (Data *) realloc(A->data, 9*sizeof(Data));
  B->data = (Data *) realloc(B->data, 9*sizeof(Data));
  Pout->data = (Data *) realloc(Pout->data, 9*sizeof(Data));

  /* - - change the matrix sizes - - */
  A->M = 3;  B->M = 3;  Pout->M = 3;
  A->N = 3;  B->N = 3;  Pout->N = 3;

  /* - - set last elements of A and B to 1 - - */
  A->data[8] = 1.0;
  B->data[8] = 1.0;
  Pout->data[8] = 0.0;

  /* - - multiply the matricies together - - */
  matrix_multiply(A, B, Pout);

  /* - - divid Pout by its last element if not too small - - */
  div_mat_by_constant(Pout, (Data) Pout->data[8], Pout);

  /* - - realloc A, B, Pout and change their sizes - - */
  A->data = (Data *) realloc(A->data, 8*sizeof(Data));
  B->data = (Data *) realloc(B->data, 8*sizeof(Data));
  Pout->data = (Data *) realloc(Pout->data, 8*sizeof(Data));
  A->M = 8;  B->M = 8;  Pout->M = 8;
  A->N = 1;  B->N = 1;  Pout->N = 1; 
}


/* =================================================================== */
double calculate_difference(Var * A, Var * B, Var *ANS,
			    int diff_type, int USE_DISPLAY)
{
  double diff_index = 0;

  switch(diff_type)
    {
    case MEAN_SQR_ERROR:
      {
	sub_matrix(B, A, ANS);   /* ans in data fork */
	ANS->var_type = INTERNAL_VARIABLE;
        diff_index = meansquared( ANS );
	break;
      }
    case CONTINUITY:
      {
	/* additional difference methods to be added */
	break;
      }
    default:
      {
	break;
      }
    }

  /* - - Convert the diff_image to uchar if USE_DISPLAY - - */
  if(USE_DISPLAY)
    {
      absval(ANS);  
      data_to_image(ANS);
      ANS->var_type = IMAGE;
    }  

  return(diff_index);
}











