/* - - This file provides support for the PAR file format - - */
#include <stdio.h>
#include "mat_util.h"
#include "par.h"

void save_var_to_par(Var * Input, char * file_name,
		    int data_type, char * comments)
{
  if(Input->var_type == IMAGE)
    {
      printf("found data in image\n");
      save_par((void *)Input->image, file_name, 
	       Input->N, Input->M, 
	       data_type, comments);
    }
  else if(Input->data == NULL)
    {
      printf("found data in data\n");
      save_par((void *)Input->data, file_name, 
	       Input->N, Input->M, 
	       data_type, comments); 
    }
  else if(Input->data == NULL)
    {
      printf("found data in data\n");
      save_par((void *)Input->data, file_name, 
	       Input->N, Input->M, 
	       data_type, comments); 
    }
}


/* =================================================================== */
int save_par(void * data, char * file_name,
	     int width, int height,
	     int data_type, char * comments)
{
  FILE * file_ptr;

  /* - - get file started  - - */
  file_ptr = fopen(file_name, "w");
  if(file_ptr == NULL)
    {
      printf("\n  ERROR: unable to create par file for writing\n");
      return(-1); /* returned NULL, changed to 0 */
    }

  /* - - Write par header info - - */
  write_par_header(file_ptr, comments, width, height, data_type);

  /* - - put the data - - */
  write_data(file_ptr, data,  width, height, data_type);

  fclose(file_ptr);
  return(0); /* to indicate successful */
}


/* =================================================================== */
void write_par_header(FILE * file_ptr,
		     char * comments,
		     int width, int height,
		     int data_type)
{
  char * magic_number = NULL;
  int bits_per_sample;
  unsigned long depth;

  /* - - color depth of frame - - */
  switch(data_type)
    {
    case CHARACTER:
      {
	depth = 256 - 1;
	bits_per_sample = CHAR_BITS;
	break;
      }
    case UNSIGNED_SHORT:
      {
	depth = 65536 - 1;       /* - max value for a u_short 2^16 - */
	bits_per_sample = SHORT_BITS;
	break;
      }
    case UNSIGNED_LONG:
      {
	/* depth = 2147483648 - 1; */
        depth = 2147483647;        /* - max value for a u_long 2^31 - */
	bits_per_sample = LONG_BITS;
	break;
      }
    case FLOAT:
      {
	depth = 1;               /* - range from 0.0 to 1.0 - */
	bits_per_sample = FLOAT_BITS;
	break;
      }
    case DOUBLE:
      {
	depth = 1;               /* - range from 0.0 to 1.0 - */
	bits_per_sample = DOUBLE_BITS;
	break;
      }
    default:
      {
	/* depth = 2147483648  - 1;  */
        depth = 2147483647;
	bits_per_sample = LONG_BITS;
	break;
      }
    }  

  /* - - Write file-type number - - */ 
  sprintf(magic_number, "P%d\n", data_type);
  fprintf(file_ptr, magic_number); 

  /* - - put comments - - */  /* !!!!!! break up long comments */
  fprintf(file_ptr, "# %s\n", comments);  

  /* - - All par extentions should go here in file- - */
  fprintf(file_ptr,"%s bits_per_sample %d\n", PAR_EXTENSION, bits_per_sample); 


  /* - - width, height of frame - - */
  fprintf(file_ptr, "%d %d\n", width, height); 
  
  fprintf(file_ptr, "%d\n", (int)depth); 

  /* !!!!!!! add other scene info here */
  
}


/* =================================================================== */
void write_data(FILE * file_ptr,
	       void * data,
 	       int width, int height,
	       int data_type)
{
  switch(data_type)
    {
    case CHARACTER:
      {
	fwrite(data, 1, width*height, file_ptr);
	break;
      }
    case UNSIGNED_SHORT:
      {
	fwrite(data, sizeof(unsigned short), width*height, file_ptr);
	break;
      }
    case UNSIGNED_LONG:
      {
	fwrite(data, 4, width*height, file_ptr);
	break;
      }
    case FLOAT:
      {
	fwrite(data, 4, width*height, file_ptr);
	break;
      }
    case DOUBLE:
      {
	fwrite(data, 8, width*height, file_ptr);
	break;
      }
    }
/*	(double *)data;*/  
/*  fwrite(&buf[index], 2, 1, myfile);   short */ 
  
}



