#include <stdlib.h>
#include <math.h>
#include "PNMImageOffsetable.h"
#include "mat_util.h"
#include "boundingbox.h"
#include "motion.h"
#include "pchirp2ia.h"
#include "simple_png_write.c"  /* simple example of png file writing */

/* define mynan is 255 */
#define MyNaN NaNChar

int VERBOSE = 0;

void usage() {
  printf( "usage: pchirp2nocrop [input .ppm filename] [output .ppm filename]\n"
          "                     [8 pchirp parameters]\n"
          "       this version support -pnmnanpass option to allow nan to\n"
          "       pass thru unchanged\n");
}

double round(double x) {
  return floor(x + 0.5);
}

void preFilterImage(const PNMImageOffsetable * pPNM) {
  int y, x, z;
  Image * ptr;
  for (y = 0; y < PNM_GetYSize( pPNM ); y++) {
    for (x = 0; x < PNM_GetXSize( pPNM ); x++) {
      ptr = PNM_Pixel( pPNM, x, y);
      for (z = 0; z < PNM_GetDepth( pPNM ); z++) {
        if (ptr[z] == NaNImage) 
	  ptr[z] = ptr[z] - 1;
      }
    }
  }
}

/* Channels go from 0..(depth-1) */
int copyChannelToVar(const PNMImageOffsetable * pPNM, int whichChannel, const Var * pVar) {
  int x, y;
  Image * srcPtr;

  if ( whichChannel < 0 || whichChannel >= PNM_GetDepth( pPNM ) )
    return -1;
  if ( pVar->var_type != IMAGE)
    return -1;
  if (pVar->channels != GREY_SCALE)
    return -1;
  if ((PNM_GetXSize( pPNM ) != pVar->N) ||
      (PNM_GetYSize( pPNM ) != pVar->M))
    return -1;
  for (y = 0; y < PNM_GetYSize( pPNM ); y++) {
    for (x = 0; x < PNM_GetXSize( pPNM ); x++) {
      srcPtr = PNM_Pixel(pPNM, x, y);
      pVar->image[y*PNM_GetXSize( pPNM )+x] = srcPtr[whichChannel];
    }
  }
  return 0;
}

/* Channels go from 0..(depth-1) */
int copyChannelFromVar(const Var * pVar,
		   int whichChannel,
		   const PNMImageOffsetable * pPNM)
{
  Image * destPtr;
  int x, y;

  if ((whichChannel < 0) || (whichChannel >= PNM_GetDepth( pPNM )))
    return -1;
  if (pVar->var_type != IMAGE)
    return -1;
  if (pVar->channels != GREY_SCALE)
    return -1;
  if ((PNM_GetXSize( pPNM ) != pVar->N) ||
      (PNM_GetYSize( pPNM ) != pVar->M))
    return -1;
  for (y = 0; y < PNM_GetYSize( pPNM ); y++) {
    for (x = 0; x < PNM_GetXSize( pPNM ); x++) {
      destPtr = PNM_Pixel( pPNM, x, y);
      destPtr[whichChannel] = pVar->image[y*PNM_GetXSize( pPNM )+x];
    }
  }
  return 0;
}

 int pnmnanpass_flag = 0;
int main(int argc, char **argv) {
  char * inFile, * outFile;
  int i, j, z, errCode;
  Var params, imageIn, p_trans, pout, imageOut;
  PNMImageOffsetable ppm;
  Data ml, mh, nl, nh;
  int mOut, nOut, M, N;
  PNMImageOffsetable ppmOut;

  int pngAction=0, *pngBuffer;
	short int *pngBufferGS;
  int g_to_ga, rgb_to_rgba, retval;
  int argv_offset = 0;
  char *outExt;

  if (argc != 11 && argc!= 12) {
    usage();
    return 1;
  }
 
  if( !strcmp( argv[1], "-pnmnanpass") ) {
    fprintf(stderr, "pchirp2nocrop: operating in NANpass mode\n");
    argv_offset++;
    pnmnanpass_flag = 1;
  }
  
  inFile = argv[1+ argv_offset];
  outFile = argv[2 + argv_offset];
  outExt = outFile+strlen(outFile)-4;
  if ((outExt[0]=='.') && ((outExt[1]=='p')||(outExt[1]=='P')) 
                       && ((outExt[2]=='n')||(outExt[2]=='N')) 
		       && ((outExt[3]=='g')||(outExt[3]=='G')))
  {
	  pngAction=1;
	  fprintf(stderr, "PNG output format request detected.\n");
  }
  
  init_var( &params, "params", 8, 1, INTERNAL_VARIABLE, GREY_SCALE,NULL );
  for (i = 0; i < 8; i++)
    params.data[i] = atof(argv[3+i+argv_offset]);

  errCode = PNM_InitFromFile( &ppm, inFile );
  if ( errCode != PNM_SUCCESS ) {
    fprintf( stderr, "pchirp2nocrop: Error %d reading file '%s': ", errCode, inFile );
    perror( "" );
    return 1;
  }

  preFilterImage( &ppm);

  /* Compute bounding box of outgoing image */

  if (boundingbox(&params, &ml, &nl, &mh, &nh) < 0) {
    perror( "pchirp2nocrop: Error computing bounding box" );
    return 1;
  }
  fprintf( stderr, "BBox: (%f, %f) to (%f, %f)\n", ml, nl, mh, nh );
  
  /* Set up everything but the data */
  M = PNM_GetYSize( &ppm );
  N = PNM_GetXSize( &ppm );
  ml = round(M*ml) / (double)M;
  mh = round(M*mh) / (double)M;
  nl = round(N*nl) / (double)N;
  nh = round(N*nh) / (double)N;
  mOut = M * (mh - ml);
  nOut = N * (nh - nl);
  
  init_var( &imageIn, "imageIn", M, N, IMAGE, GREY_SCALE,NULL );
  init_var( &imageOut, "imageOut", mOut, nOut, IMAGE, GREY_SCALE,NULL );

  /* translate by Ml and Nl in the inverse space */
  init_var( &p_trans, "p_trans", 1, 8, INTERNAL_VARIABLE, GREY_SCALE,NULL );
  init_var( &pout, "pout", 1, 8, INTERNAL_VARIABLE, GREY_SCALE,NULL );
  p_trans.data[0] = 1;
  p_trans.data[1] = 0;
  p_trans.data[2] = ml;
  p_trans.data[3] = 0;
  p_trans.data[4] = 1;
  p_trans.data[5] = nl;
  p_trans.data[6] = 0;
  p_trans.data[7] = 0;
  pcompose(&params, &p_trans, &pout);
  Var_Destructor( &params );
  Var_Destructor( &p_trans );
  fprintf( stderr, "final params: " );
  for (j = 0; j < 8; j++)
      fprintf( stderr, "%f ", pout.data[j] );
  fprintf( stderr, "\n" );

  if ( PNM_InitFromSize( &ppmOut, nOut, mOut, PNM_GetDepth( &ppm ), ppm.magic_number, ppm.plm_field_1, ppm.plm_field_2) != PNM_SUCCESS ) {
    perror( "pchirp2nocrop: Error initializing output PPM" );
    return 1;
  }
  PNM_SetMOffset( &ppmOut, ml*M+PNM_GetMOffset( &ppm) );
  PNM_SetNOffset( &ppmOut, nl*N+PNM_GetNOffset( &ppm) );

  // if png, alloc png buffer 
  if (pngAction) {
    pngBuffer = (int*)malloc(mOut*nOut*sizeof(int));
    pngBufferGS = (short int *)pngBuffer;
    memset(pngBuffer, 0, mOut*nOut*sizeof(int));
    g_to_ga     = (PNM_GetDepth( &ppm ) == 1);
    rgb_to_rgba = (PNM_GetDepth( &ppm ) == 3);
  } else {
    pngBuffer   = NULL;
    pngBufferGS = NULL;
    g_to_ga = rgb_to_rgba = 0;
	}

  for (z = 0; z < PNM_GetDepth( &ppm ); z++) {
    fprintf( stderr, "Doing channel %d of %d...\n", z+1, PNM_GetDepth( &ppm ) );
    if (copyChannelToVar( &ppm, z, &imageIn)) {
      fprintf( stderr, "Error copying channel %d to var", z+1 );
      exit(1);
    }
    /* if blue channel, chval all 255s to 254 in only the blue channel */
    /* this assumes it's either 1ch or 3ch, e.g. would fail if other number */
    if( pnmnanpass_flag == 0 ) {
      if ((PNM_GetDepth( &ppm ) > 1) && (z==(PNM_GetDepth( &ppm )-1)))
        chval(&imageIn,MyNaN,254); /* change all 255s to 254s in blue channel*/
      if (PNM_GetDepth( &ppm ) == 1)
        chval(&imageIn,MyNaN,254);/*change all 255s to 254s in greyscale image*/
    }
      pchirp2( &pout, &imageIn, &imageOut,
         /* Interp_order and antialias_method set to defaults
            for the moment */
	      1, -1, mOut, nOut, 0);
    if (!pngAction) {
      if (copyChannelFromVar(&imageOut, z, &ppmOut)) {
        fprintf( stderr, "Error copying channel %d from var", z+1);
        exit(1);
      }
    } else {
      // if png, copy to png buffer 
      int i,j,c,a;
      int doAlpha = ( PNM_GetDepth(&ppm)==z+1 );
  	if (g_to_ga) {
        for(i=0;i<mOut;i++) {
          for(j=0;j<nOut;j++) {
  	        c = ((double*)imageOut.image)[i*nOut+j];
  	        a = ((c<0)||(c>255)||(doAlpha&&(c>254)))?0:255;
  	        pngBufferGS[i*nOut+j] = ((c<<8)|a) ;
          }
  	    }
  	  } else if (rgb_to_rgba) {
         for(i=0;i<mOut;i++) {
          for(j=0;j<nOut;j++) {
  	        c = ((double*)imageOut.image)[i*nOut+j];
  	        a = ((c<0)||(c>255)||(doAlpha&&(c>254)))?c=0:255;
            pngBuffer[i*nOut+j] |= (c<<((z+1)*8))|a;
          }
  	}
      } else {
        perror("Non rgb or greyscale input not yet supported for PNG output.");
        return 1;
      }
    }
  } /* for */
  Var_Destructor( &imageIn );
  Var_Destructor( &imageOut );
  Var_Destructor( &pout );

  if (!pngAction) 
  {
    if (PNM_WriteToFile( &ppmOut,outFile) != PNM_SUCCESS ) {
      perror( "pchirp2nocrop: Error writing file" ) ;
      return 1;
    }
  } else {
    // if png, set alpha channel & write to disk
		retval = 0;
    if (g_to_ga) {
      printf("Exporting Gray_Alpha with reversed offset %g, %g\n",(double) ml*M, (double) nl*N);
		 	retval = simple_png_write_ga(
          outFile,
          pngBufferGS, nOut, mOut,
          "no comment at this time",
          /* write out in reversed, for the left hand coord. system of png */
//          nl*N, ml*M); /* reversed order */
          nl*N, ml*M);
		} 
		else if(rgb_to_rgba) {
      printf("Exporting RGB_Alpha with reversed offset ml*M=%g, nl*N=%g\n", (double) ml*M, (double) nl*N);
	  retval = simple_png_write_rgba(
          outFile,
          pngBuffer, nOut, mOut,
          "no comment at this time",
/*          ml*M, nl*N, 0); */ /* this 0 is for compression */
                          /* true (1) means uncompressed */
          /* write out in reversed, for the left hand coord. system of png */
          nl*N, ml*M, 0); /* reversed order */
		} else {
			retval = -1;
      perror( "pchirp2nocrop: Unknown conversion requested." ) ;
		}

		// free prior to returning
    free(pngBuffer);

		if (retval !=0) {
      perror( "pchirp2nocrop: Error writing PNG file" ) ;
      return 1;
		}
  }
  PNM_Destroy( &ppmOut );
  PNM_Destroy( &ppm );   
  return 0;
}

