#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <assert.h>
#include <getopt.h>
#include "cement.h"
#include "pnmutils.h"
#include "file.h"

//Need to add this to wrap jpeglib.h for C++
#ifdef __cplusplus
extern "C" {
#endif 

#include <jpeglib.h>

#ifdef __cplusplus
}
#endif



#define PLMNAN 128

//#define JPEG_QUALITY 90 (now defined in cement.h)

/* Functions need in the library and the executable */
static inline double createValue(double a,double p_inv);
double compute_gnf(double const *block,int heigth,int width,int nchans,double * largest_spacial_in);
static unsigned char binarySearch(double value);
double *read_plm_image( File *plmfp, struct image_params *plmparams);
double *read_block( double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *plmparams );
double *read_block2( double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *plmparams );
double *write_block( double *block, double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *pnmparams );
int hist_eq_block( double *block, int blockheight, int blockwidth, int nchans ,double stretch_factor,
                int use_hanning);
double *write_block_pixel( double *block, double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *pnmparams );
double calculate_block_mean( double *block, int blockheight,int blockwidth, 
                             int nchans, int channel,int newblock);
double calculate_block_var(double *block,int blockheight,int blockwidth,
                           int nchans,double mean,int use_new_block );
double calculate_image_mean( double *img, struct image_params *a_params, int nchans);
double calculate_image_var( double *img, struct image_params *a_params, int nchans, double mean);
double shift_block_to_mean( double *block, double mean_g, 
                            double block_mean, int blockwidth, 
                            int blockheight, int nchans);
double dilate_block_to_var( double *block, double var_g, 
                            double block_mean, int blockwidth, 
                            int blockheight, int nchans, double block_var, double max_var,int scaling);
int calculate_image_block_var(double *myimg, struct image_params *a_params, 
                              int nchans, int blockwidth, int blockheight,
                              double *max_var, double *min_var);

int write_status(int val, int max_val);
unsigned char *plm2pnm_array(double *a_pixel, int numelements, int nchans );
int write_pnm_image( double *imgarray, File *dest, struct image_params *pnmparams );
double hanning(int n, int N);

void initialize_img(double *img, int height, int width, 
                      int nchans, double* values );

static double inverseLookup[256];
static double k=1;
static int kinputflag=0;
static int histogrameqflag=0;
//static double scale_factor=1;
static int blocksize = 7;
static double mean_shift =2;
static double var_gamma = 0.1;
static double max_dilation = 4;
static double gnf_spacial_grad_thresh = .017;
static int show_nastiness = 0;
static int variable_blocksize = 0;
//static int flatregion=1000;


#ifndef OBJECT_COMPILE
/* Functions only needed in the executable */
void usage(char * string);
void parse_commandline(int argc, char ** argv,
                       char ** input_filename,
                       char ** output_filename,
                       char ** lut_filename);

void mystatus_lut(char *action, char *filename, char *lut_filename, double red, double green, double blue, double complete);
#endif

void readlut(char *lutfilename);
int plm2pnm(char *input_filename,
            char *dest_filename,
            char *command,
            void (*status)(char *action, char *filename, double R,double G, double B, double complete)
           )
{
  return plm2pnm_lut(input_filename, dest_filename, NULL, command, status, NULL);
}

/* set one of the status functions to NULL when calling */
int plm2pnm_lut(char *input_filename,
            char *dest_filename,
            char *lut_filename,
            char *command,
            void (*status)(char *action, char *filename, double R,double G, double B, double complete),
            void (*status_lut)(char *action, char *filename, char *lut, double R,double G, double B, double complete)
           )
{
    struct image_params a_params={NULL,0,0,0,0,0.0};
//    char *dest_filename=NULL;
    File *a, *dest;
    double p=EXPONENT;  /* exponent for pow */
    double        a_pixel[ARRAY_SIZE];
    unsigned char out_pixel[ARRAY_SIZE];
    int i,j=0,num_of_pixels_read=0;
    int channels_per_pixel;

    
    unsigned int pixels=0;
    unsigned int numberOfPixels;

    char string[MAX_STRING_LENGTH];
    int image_ptr;
    printf("show_nastiness = %d\n",show_nastiness);
    a_params.filename=input_filename;
    
    if ((a=fileopen(a_params.filename, "r"))==NULL)
    {
        fprintf(stderr, "Unable to open %s.\n", a_params.filename);
        exit(EXIT_FAILURE);
    }

    if ((dest=fileopen(dest_filename, "w"))==NULL)
    {
        fprintf(stderr, "Unable to open %s.\n", dest_filename);
        exit(EXIT_FAILURE);
    }

    get_image_type(a, &a_params);

    if (a_params.type=='A')
    {
        channels_per_pixel=3;
        sprintf(string,"P%c\n", '6');
        filewrite(string,sizeof(unsigned char),strlen(string),dest);
        get_image_params_preserve_comments(a,&a_params,dest);
    }
    else if (a_params.type=='9')
    {
        channels_per_pixel=1;
        sprintf(string,"P%c\n", '5');
        filewrite(string,sizeof(unsigned char),strlen(string),dest);
        get_image_params_preserve_comments(a,&a_params,dest);
    }
    else
    {
        fprintf(stderr, "Portable Lightspace map (plm) must be type 9 or A and it is type %c\n",a_params.type);
        exit(EXIT_FAILURE);
    }


    /* Use the exponent from the image file given, if there is one. */
    /* This determines the LUT file used. */
    /* Otherwise the default exponent and lut file are used. */
    /* If conflict between given LUT file and exponent, there will be a warning message. */
    if (a_params.exponent > 0) 
	p = (double) a_params.exponent;

    if (lut_filename == NULL) 
      {
	lut_filename = (char *) malloc (50*sizeof(char));
	if (lut_filename == NULL) {
	  printf("malloc failed in plm2pnm_lut\n");
	  exit(-1);
	}
	sprintf(lut_filename, "../lookuptables/Kodak_DCS260_response_function" );
      }    


    /* maali may 9 2000 */
    if(lut_filename !=NULL)
    {
        readlut(lut_filename);
    }


    /*    if ((a_params.exponent!=(float)EXPONENT))
    {
        fprintf(stderr, "Lightspace must have the same exponent.\n");
        exit(EXIT_FAILURE);
	}*/

    sprintf(string,"# %s %s -o %s \n",command,input_filename,dest_filename);
    filewrite(string,sizeof(unsigned char),strlen(string),dest);

    sprintf(string, "%d %d\n%d\n", a_params.width,
            a_params.height, a_params.max_val);
    filewrite(string,sizeof(unsigned char),strlen(string),dest);


    image_width=a_params.width;
    image_height=a_params.height;
    image_buffer=(JSAMPLE *)malloc(3*image_width*image_height*sizeof(unsigned char));
    if(image_buffer==NULL)
        printf("ERROR: Could not allocate buffer for image\n");
    image_ptr=0;

    if( k!=1 ) assert(kinputflag == 1);    

    /*  Work with individual blocks */
    if( histogrameqflag == 1&&blocksize!=1 ) {
      double *myblock;
      double *myimg = read_plm_image( a, &a_params);
      double *write_out_img;
      double max_var, min_var;
      double initial_canvas_vals[3];
      int i,j;
      int blockwidth=blocksize;
      int blockheight=blocksize;
      int nchans = 3;
      int use_new_block = 1;
      double mean_g = PLMNAN, var_g = PLMNAN, block_mean = PLMNAN, block_var=PLMNAN;
      double dilation_factor;
      int shrunken_amount;
      write_out_img = (double *)malloc(sizeof(double)*a_params.height*a_params.width*nchans);
  
      fprintf(stderr, "Initializing Image\n");
      initial_canvas_vals[0] = PLMNAN;
      initial_canvas_vals[1] = PLMNAN;
      initial_canvas_vals[2] = PLMNAN;
      initialize_img(write_out_img, a_params.height, a_params.width, 
                     3, initial_canvas_vals );
      fprintf(stderr, "Done\n");

      /* calculate some image statistics */
      mean_g = calculate_image_mean( myimg, &a_params, 3);      
      fprintf(stderr, "Image Mean = %g\n", mean_g);
      var_g = calculate_image_var( myimg, &a_params, 3, mean_g);      
      fprintf(stderr, "Image Var = %g\n", var_g);
      calculate_image_block_var(myimg, &a_params, 3, blockwidth, blockheight, &max_var, &min_var);
      fprintf(stderr, "Min Image Block Var = %g, Max Image Block Var = %g\n", 
              min_var, max_var);
      //max_var = 4*0.001;

      for( i=0; i<a_params.height-blockheight ; i+=1 ) { /* rows */
        write_status(i,a_params.height-blockheight);
        use_new_block = 1;
        for( j=0; j<a_params.width-blockwidth ; j+=1 ) { /*cols */
          double tmp_gnf;
          int gnf_okay = 0;
          int current_block_width = blocksize;
          int current_block_height = blocksize;
          double  *tmp_block;
          shrunken_amount = 0;
          myblock = read_block( myimg, j,i, nchans, blockwidth, blockheight, &a_params );
	  
          while (variable_blocksize && !gnf_okay && current_block_width>1 ){ 
             double tmp_largest_sp =0;
             tmp_block = read_block2( myimg, j,i, nchans, current_block_height,
                                   current_block_width, &a_params );
             tmp_gnf = compute_gnf(tmp_block,current_block_height,
                                   current_block_width,3,&tmp_largest_sp);
             if ((tmp_gnf < gnf_spacial_grad_thresh) || variable_blocksize==0){
                gnf_okay = 1;
             }
             else{
               shrunken_amount +=1;
              // // fprintf(stderr,"reducing block size by %d\n",shrunken_amount);
                current_block_height = current_block_height -1;
                current_block_width = current_block_width -1;
             }
            gnf_okay=1;
          }
          block_mean = calculate_block_mean( myblock, blockwidth, blockheight, 
                                             3,1,use_new_block);
          block_var = calculate_block_var( myblock, blockwidth, blockheight, 
                                           3, block_mean,use_new_block);
          use_new_block = 0;
          shift_block_to_mean( myblock, mean_g, block_mean, blockwidth, blockheight,3);
          dilation_factor=dilate_block_to_var(myblock, var_g, block_mean, 
                                              blockwidth, blockheight,3, block_var,
                                              max_var,0);
          write_block_pixel( myblock, write_out_img,j,i,nchans,blockwidth,blockheight,&a_params ); 
        }
      }
      write_pnm_image( write_out_img, dest, &a_params );
      exit(0);
    }
    write_status(1,1);
    numberOfPixels=a_params.width*a_params.height;
    /* Read int the lightspace image */        
    while (!fileeof(a))
    {
      num_of_pixels_read+=fileread(a_pixel, sizeof(double),CHANNELS_PER_PIXEL, a);
      fprintf(stderr,"plm2pnm reading %4.2f percent done\r       ",
             (double)num_of_pixels_read/numberOfPixels *100/CHANNELS_PER_PIXEL);
      if(fileeof(a)) break;
      
			if( isnan(a_pixel[j]) ) out_pixel[j] = 255; 
			else out_pixel[0] = binarySearch(a_pixel[0]*k);
			if( isnan(a_pixel[j+1]) ) out_pixel[j+1] = 255; 
			else out_pixel[1] = binarySearch(a_pixel[1]*k);
			if( isnan(a_pixel[j+2]) ) out_pixel[j+2] = 255; 
			else out_pixel[2] = binarySearch(a_pixel[2]*k);
      filewrite(out_pixel, sizeof(unsigned char), CHANNELS_PER_PIXEL, dest);
    }
    
    fileclose(a);
    fprintf(stderr,"\n"); 
    if(isJpegFile(dest)==1)
    {
        fileclose(dest);
        write_JPEG_file (dest_filename,JPEG_QUALITY);
    }
    else
        fileclose(dest);

    free(image_buffer);
    return(EXIT_SUCCESS);
}

static inline double createValue(double a, double p_inv)
{
    return(pow(a,p_inv));
}


static unsigned char binarySearch(double value)
{
    int low=0;
    int high=0x100;
    int middle=0x7f;
	if( isnan(value) ) return 255;
    
    while(1)
    {
        if(high==middle || low==middle)
            return(middle);

        if(value>=inverseLookup[middle])
        {
            low=middle;
            middle=((high-low)>>1)+middle;
        }
        else
            if(value<inverseLookup[middle])
            {
                high=middle;
                middle=((high-low)>>1)+low;
            }
    }
    return(middle);
}

#ifndef OBJECT_COMPILE
int main(int argc, char ** argv)
{
    char * dest_filename=NULL;
    char * input_filename=NULL;
    char * lut_filename=NULL;

    parse_commandline(argc, argv, &input_filename, &dest_filename,&lut_filename);
    return(plm2pnm_lut(input_filename,dest_filename,lut_filename,argv[0],NULL,mystatus_lut));
}

void mystatus_lut(char *action, char *filename, char *lut_filename, double R,double G, double B, double complete)
{
    if (lut_filename!=NULL) 
        printf("%s %s (%s) %d %d %d %d%%\r",
            action,filename,lut_filename,(int)R,(int)G,(int)B,(int)complete);
    else printf("%s %s %d %d %d %d%%\r",action,filename,(int)R,(int)G,(int)B,(int)complete);
    fflush(stdout);
    if((int)complete==100)
        printf("\n");
}



void usage(char * string)
{
  fprintf(stderr, "Use: %s [LightSpaceFile] -o [ImageSpaceFile] "
		  "[-l [LookUpTableFile]]\n",string);
  fprintf(stderr, "    [-s histogramScalingFactor] [-b histogramBlockfileSize]\n"); //fprintf(stderr, "Example: %s trowel_out.plm -o pork.jpg\n\n",string);
  fprintf(stderr, "Example: %s cemented.plm -o .pork --lut collected_data -k 1.00 -b 6" ,string);

  exit(EXIT_SUCCESS);
}


/*  Cleaned up by Corey Apr 13 2002 */
void parse_commandline(int argc, char ** argv,
                       char ** input_filename,
                       char ** output_filename,
                       char ** lut_filename)
{
   int c,i;
   int got_output_filename = 0; 

   for( i=1 ; i<argc ; i++ ) {
     if( !strcmp( "-lut", argv[i]) ) {
       fprintf(stderr, "Found %s, changing to ", argv[i]);
       argv[i] = strdup( "--lut" );
       fprintf(stderr, "%s. done.\n ", argv[i]);
     }
   }

   while(1){

      int option_index = 0;
      static struct option long_options[] = {
                   {"gain", 1, 0, 'k'},
                   {"lut", 1, 0, 'l'},
                   {"scale", 1, 0, 's'},
                   {"blocksize", 1, 0, 'b'},
                   {"mean-shift", 1, 0, 'm'},
                   {"output", 1, 0, 'o'},
                   {"dilation", 1, 0, 'd'},
                   {"vargamma", 1, 0, 'g'},
                   //{"help", 0, 0, 'h'},
                   {"histo", 0, 0, 'h'},
                   {0, 0, 0, 0}
               };

      c = getopt_long(argc,argv,"vwk:l:b:o:m:d:g:n:f:",long_options, &option_index); /*valid options*/

      /* Modified by Pouya 13/01/2003: Wrong file name was being reported in many cases.
       * The filename is not "cementinit," it's "plm2pnm" actually, but I will
       * use argv[0] in all the changed lines */

      if (c==-1) break; /* done parsing arguments */ 
      if (c=='?'){
         fprintf(stderr,"%s: input option given with no argument (-%c) \n",argv[0],optopt);
         usage(argv[0]);
      }

      switch (c){

      case 'k':
         fprintf(stderr,"%s: Using k (gain) value of `%s'\n",argv[0],optarg);
         sscanf(optarg,"%lg\n",&k);
         kinputflag = 1;
         break;

      case 'l':
      //case 'u':
      //case 't':
         fprintf(stderr,"%s: Using lookup table `%s'\n",argv[0],optarg);
         *lut_filename=(char *)strdup(optarg);
         break;
      case 'b':
         fprintf(stderr,"%s: Using blocksize factor `%s'\n",argv[0],optarg);
         sscanf(optarg,"%d\n",&blocksize);
         histogrameqflag = 1;
         break;
      case 'o':
         fprintf(stderr,"%s: Using output file `%s'\n",argv[0],optarg);
         *output_filename=(char *)strdup(optarg);
         got_output_filename = 1;
         break;
      case 'h':
         //usage(argv[0]); 
         histogrameqflag = 1;
         break;
      case 'm':
         fprintf(stderr,"%s: Using mean-scale factor `%s'\n",argv[0],optarg);
         sscanf(optarg,"%lg\n",&mean_shift);
         break;
      case 'd':
         sscanf(optarg,"%lg\n",&max_dilation);
         fprintf(stderr,"%s: Using max dilation = %g\n", argv[0],max_dilation);
         break;
      case 'g':
         sscanf(optarg,"%lg\n",&var_gamma);
         fprintf(stderr,"%s: Using variance gamma = %g\n", argv[0],var_gamma);
         break;
      case 'n':
         sscanf(optarg,"%lg\n",&gnf_spacial_grad_thresh);
         fprintf(stderr,"%s: Using GNF spacial gradient threshold = %g\n",argv[0],
                         gnf_spacial_grad_thresh);
         break;
      case 'w':
         fprintf(stderr,"%s: showing nasty pixels\n",argv[0]);
         show_nastiness =1;
         break;
      case 'v':
         fprintf(stderr,"%s: using variable blocksizes\n",argv[0]);
         variable_blocksize =1;
         break;

      }
      
   }

   /* options have now been parsed, just need to get input filename */
   if (argv[optind]==NULL){
      fprintf(stderr,"\n%s: No input file given\n\n",argv[0]);
      usage(argv[0]);
   } 
   else{
      *input_filename=(char *)strdup(argv[optind]);
   }
   if (!got_output_filename){
      fprintf(stderr,"\n%s: No output file given\n\n",argv[0]);
      usage(argv[0]);
   }

   fprintf(stderr,"%s: Using input file `%s'\n",argv[0],argv[optind]);
}


   #endif
/* ===================================================================== */
/* ===================================================================== */
/*                    BLOCK MANIPULATION FUNCTIONS                       */
/* ===================================================================== */
/* Readlut - reads a given lookup table. */

void readlut(char *lutfilename)
{

    // some variables used in reading a text file lut
    FILE *lutfile;
    int lutsize=256;
    float p;
    int i;

    if(lutfilename !=NULL) 
      {
        if ((lutfile = fopen(lutfilename, "r")) == NULL)
	  {
            fprintf(stderr, "Unable to open %s.\n", lutfilename);
	    fclose(lutfile);
            exit(EXIT_FAILURE);
	  } 
	else 
	  {
	    fscanf(lutfile, "%f", &p);
	    for (i=0; i<lutsize; i++)
	      fscanf(lutfile, "%lf", &(inverseLookup[i]));
		   
	    if (i != lutsize) {
	      fprintf(stderr, "Incomplete read of LUT. Numread: %d\n", i);
	      fclose(lutfile);
	      exit(1);
	    }
	  }
	fclose(lutfile);
      }
}

/* at this point we assume that the header has been read in, and the file
 * pointer now is at the head of the data for the plm so we can
 * call fileread properly */
double *read_plm_image( File *plmfp, struct image_params *plmparams) 
{
  double *imgarray, *imgarrayptr;
  int numchans = 3; /* HARDCODED to 3 for now.  TODO: add numchans into image_params struct and
                       use that or test on magic number */
  int num_of_elements;
  int elements_per_read = 1; /* elements read each time fileread is called */
  int elements_read=0;          /* elements read each time fileread is called (as returned from fileread) */
  int total_elements_read=0;    /* sum up the number of elements read to make sure we get the whole image */
  //char invalid;                 /* char just to make sure no remaining file info after we read */


  num_of_elements = plmparams->width * plmparams->height * numchans;
  if( (imgarray = (double *)malloc( sizeof(double) * num_of_elements)) == NULL) {
    fprintf(stderr, "plm2pnm error: Could not allocate requested %d bytes of memory\n",
                    (int)(sizeof(double) * num_of_elements) );
  }
  imgarrayptr = imgarray;

  while ( total_elements_read < num_of_elements ) { 
    elements_read = fileread( imgarrayptr, sizeof(double), elements_per_read, plmfp);
    imgarrayptr += elements_read;
    total_elements_read += elements_read;
  }
  fprintf(stderr, "plm2pnm: read %d elements of an expected %d elements\n",
                  total_elements_read, num_of_elements);

  /* make sure that there is no remaining information in the file */ 
  //assert( fread(&invalid ,1,1,plmfp->fp)==0);
  return imgarray; 
}

int write_pnm_image( double *imgarray, File *dest, struct image_params *pnmparams )
{
  unsigned char* imgarrayptr;
  int numchans = 3; /* HARDCODED */
  int num_of_elements;
  int elements_per_write = 1;
  int elements_written=0;
  int total_elements_written=0;
  unsigned char *pnm_array;

  num_of_elements = pnmparams->width * pnmparams->height * numchans;
  
  pnm_array = plm2pnm_array( imgarray, pnmparams->width*pnmparams->height, numchans);

  imgarrayptr = pnm_array;
  while ( total_elements_written < num_of_elements ) { 
    elements_written = filewrite( imgarrayptr, sizeof(unsigned char), elements_per_write, dest);
    imgarrayptr += elements_written;
    total_elements_written += elements_written;
  }
  assert(pnm_array != NULL );
  free( pnm_array );

  fprintf(stderr, "plm2pnm: wrote %d elements of an expected %d elements\n",
                  total_elements_written, num_of_elements);

  return total_elements_written;

}

/* e.g. send in double *array is 30 doubles long, then numelements is 10 and nchans is 3 for rgb */
/* otherwise, say double *array is 30 doubles long, then numelements is 30, nchans is 1 for gray */
unsigned char *plm2pnm_array(double *a_pixel, int numelements, int nchans ) 
{
  int i,z;
  /* just assume that our request for memory is a reasonable size */
  unsigned char *pnmarray = (unsigned char *)malloc( numelements*nchans*sizeof(unsigned char) );
  for( i=0 ; i<numelements*nchans ; i+=nchans ) {
    /* do the plm2pnm conversion using the lut, once per each channel rgb, (z) */
    for( z=0; z<nchans ; z++ ) {
 			if( isnan(a_pixel[i+z]) ) 
              pnmarray[i+z] = 255; 
			else 
              pnmarray[i+z]=binarySearch(a_pixel[i+z]*k);
    }
  }
  return pnmarray;
}

double *get_pixel( double *imgarray, int x, int y, int nchans, struct image_params *plmparams )
{
  return( &(imgarray[ nchans*(plmparams->width*y + x) ]) );
}

void write_pixel( double *imgarray, double *value, int x, int y, int nchans, struct image_params *pnmparams )
{
  int z;
  for( z=0 ; z<nchans; z++ ) {
    imgarray[ nchans*(pnmparams->width*y + x) + z ] = value[z];
  }
  return;
}

/* x and y are the location of the upper left hand corner of the desired block */
/* returns an array 7x7 which the caller will need to free() later */
/* the block memeory is reused , so after this is called the previous block is no longer*/
double *read_block( double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *plmparams )
{
  double *src;
  int i,j,z;
  static double *block = NULL;

  if( block == NULL ) {
    if( (block = (double *)malloc( sizeof(double)*blockwidth*blockheight*nchans )) == NULL ) {
      fprintf(stderr, "plm2pnm: error initializing block\n");
      exit(1);
    }
  }

  for( i=0 ; i<blockheight ; i++ ) { /* rows */
    for( j=0; j<blockwidth ; j++ ) { /* cols */
      src = get_pixel( imgarray, x+j, y+i, nchans,  plmparams ); /* returns ptr to rgb triplet */
      for( z=0; z<nchans; z++ ) { /* channels */
        block[nchans*(i*blockwidth + j) + z ] = src[z];
      }
    }
  } 
  return block;
}
double *read_block2( double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *plmparams )
{
  double *src;
  int i,j,z;
  static double *block = NULL;

  if( block == NULL ) {
    if( (block = (double *)malloc( sizeof(double)*blockwidth*blockheight*nchans )) == NULL ) {
      fprintf(stderr, "plm2pnm: error initializing block\n");
      exit(1);
    }
  }

  for( i=0 ; i<blockheight ; i++ ) { /* rows */
    for( j=0; j<blockwidth ; j++ ) { /* cols */
      src = get_pixel( imgarray, x+j, y+i, nchans,  plmparams ); /* returns ptr to rgb triplet */
      for( z=0; z<nchans; z++ ) { /* channels */
        block[nchans*(i*blockwidth + j) + z ] = src[z];
      }
    }
  } 
  return block;
}


/* writes a given blockheight  by blockwidth block into an array at location 
x,y (x,y) topleft corner of block position*/
double *write_block( double *block, double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *pnmparams )
{
  int i,j;

  for( i=0 ; i<blockheight ; i++ ) { /* rows */
    for( j=0; j<blockwidth; j++ ) { /* cols */
        write_pixel( imgarray, &(block[nchans*(i*blockwidth + j)]), x+j, y+i, nchans, pnmparams ); 
    }
  } 
  return block;
}
double *write_block_pixel( double *block, double *imgarray, int x, int y, int nchans,
                int blockwidth, int blockheight, struct image_params *pnmparams )
{
  int i,j;
    i = blockwidth/2;
    j = blockheight/2;
//  for( i=0 ; i<blockheight ; i++ ) { /* rows */
 //   for( j=0; j<blockwidth; j++ ) { /* cols */
      
        write_pixel( imgarray, &(block[nchans*(i*blockwidth + j)]), x+j, y+i, nchans, pnmparams ); 
//    }
 // } 
  return block;
}


/* this function takes a block, and actually changes the values inside the block */
int manip_block( double *block, int blockheight, int blockwidth, int nchans )
{
  int i,j,z;
  for( i=0 ; i<blockheight ; i++ ) {
    for( j=0 ; j<blockwidth ; j++ ) {
      for( z=0 ; z<nchans ; z++ ) {
        block[ nchans*(i*blockwidth + j) +z] = 0;
      }
    }
  }
  return 0;
}



/* this function takes a block, and actually changes the values inside the block */
int hist_eq_block( double *block, int blockheight, int blockwidth, int nchans ,double stretch_factor,
                   int use_hanning)
{
  int i,j;
  double dist_g;
  double mean_shift=0;
  double hannfactor = 1;
  double max_r =0, min_r = 1000000000;
  double max_g =0, min_g = 1000000000;
  double max_b =0, min_b = 1000000000;
  double  desired_max_r, desired_min_r;
  double  desired_max_g, desired_min_g;
  double  desired_max_b, desired_min_b;
  for( i=0 ; i<blockheight ; i++ ) {
    for( j=0 ; j<blockwidth ; j++ ) {
      if( use_hanning == 1 ) hannfactor = hanning(i,blockheight)*hanning(j,blockwidth);
         if (block[nchans*(i*blockwidth +j) ] > max_r)
            max_r =block[nchans*(i*blockwidth +j) ] *hannfactor;
         if (block[nchans*(i*blockwidth +j) ] < min_r)
            min_r =block[nchans*(i*blockwidth +j) ] *hannfactor ;
         if (block[nchans*(i*blockwidth +j) +1] > max_g)
            max_g =block[nchans*(i*blockwidth +j) +1]*hannfactor ;
         if (block[nchans*(i*blockwidth +j) +1] < min_g)
            min_g =block[nchans*(i*blockwidth +j) +1]*hannfactor ;
         if (block[nchans*(i*blockwidth +j) +2] > max_b)
            max_b =block[nchans*(i*blockwidth +j) +2]*hannfactor ;
         if (block[nchans*(i*blockwidth +j) +2] < min_b)
            min_b =block[nchans*(i*blockwidth +j) +2]*hannfactor ;
    }
  }
  desired_max_r = stretch_factor * max_r;
  desired_max_g = stretch_factor * max_g;
  desired_max_b = stretch_factor * max_b;

  desired_min_r = min_r*1/stretch_factor;
  desired_min_g = min_g*1/stretch_factor;
  desired_min_b = min_b*1/stretch_factor;

  desired_min_r = min_r-(desired_max_r-max_r);
  desired_min_b = min_b-(desired_max_b-max_b);
  desired_min_g = min_g-(desired_max_g-max_g);

  if (desired_min_r <0) desired_min_r = 0; 
  if (desired_min_g <0) { mean_shift = (0.0 - desired_min_g) ; desired_min_g = 0; }
  if (desired_min_b <0) desired_min_b = 0;

  if (desired_max_r >0.5) desired_max_r = 0.5;
  if (desired_max_g >0.5) { mean_shift = (0.5 - desired_max_g); desired_max_g = 0.5; }
  if (desired_max_b >0.5) desired_max_b = 0.5;
  if( ( (desired_min_g < 0 ) && (desired_max_g > 0.5) ) ) mean_shift = 0;
  mean_shift=0;


  for( i=0 ; i<blockheight ; i++ ) {
    for( j=0 ; j<blockwidth ; j++ ) {
       if( max_g == min_g ) dist_g = 0;
       else dist_g = (block[nchans*(blockwidth*i)+j +1]-min_g)/(max_g-min_g);
       block[ nchans*(blockwidth*i+j) +0]= mean_shift + desired_min_r+dist_g*(desired_max_r-desired_min_r);
       block[ nchans*(blockwidth*i+j) +1]= mean_shift + desired_min_g+dist_g*(desired_max_g-desired_min_g);
       block[ nchans*(blockwidth*i+j) +2]= mean_shift + desired_min_b+dist_g*(desired_max_b-desired_min_b);
    }
  }

  return 0;
}

double calculate_block_mean( double *block, int blockheight, int blockwidth, 
                             int nchans, int channel, int new_block )
{
  int i,j,rowpos;
  static double *mean;
  static int counter = 0;
  double sum = 0;
  int total_num_elements =0;
  static int *num_elements;
  if (mean == NULL){
     mean = (double*)malloc(sizeof(double)*blockwidth);
     num_elements = (int*)malloc(sizeof(int)*blockwidth);
  }
  if (new_block == 1){
  /* if new_block == 1, then we need to regenerate all of the column means */
    counter = 0;
    for (i=0;  i < blockwidth; i++){
      mean[i] = 0;
      num_elements[i] = 0;
    }
    for ( i=0 ; i<blockheight ; i++ ) {
      rowpos = nchans*blockwidth*i;
      for( j=0 ; j<blockwidth ; j++ ) {
        if( block[ rowpos+nchans*j +channel] == 128 ) continue;
        mean[j]+= block[ rowpos+nchans*j +channel];
        num_elements[j]++;
      }
    }
  }
  /* otherwise we just need to generate the mean for the new column */
  else{ 
  /* if new_block == 1, then we need to regenerate all of the column means */
    counter ++;
    if (counter == blockwidth){
       counter = 0; 
    }
    mean[counter]=0;
    num_elements[counter] =0;
    for (i= 0;i<blockheight;i++){
      rowpos = nchans*blockwidth*i;
      if(block[ rowpos+nchans*(blockwidth-1)+ channel] == PLMNAN) continue;
      mean[counter]+= block[rowpos+nchans*(blockwidth-1) +channel];
      num_elements[counter] = num_elements[counter]+1;
    }
  }
  for (i=0;i<blockwidth;i++){
     sum+=mean[i];
     total_num_elements+=num_elements[i];
  }
  return sum/(total_num_elements);

}

double calculate_block_var( double *img, int block_width, int block_height, 
                            int nchans, double mean,int use_new_block)
{
  int i,j,rowpos,total_num_elements=0;
  static double *var,total_var;
  static int counter = 0;
  static int *num_elements; 

  if (var == NULL){
     var = (double*)malloc(sizeof(double)*block_width);
     num_elements = (int *)malloc(sizeof(int)*block_width);
  }
  /* if the new_block == 1, then we need to regenerate all of the column vars */
  if (use_new_block == 1){
    counter =0;
    for (i=0; i<block_height; i++){
      /* clear out the arrays */
      var[i] =0;
      num_elements[i]=0;
    }
    for( i=0; i<block_height ; i+=1 ) { /* rows */
      rowpos = nchans*i*block_width;
      for( j=0; j<block_width ; j+=1 ){  /*cols */
        if( img[ rowpos+nchans*j +1 ] == PLMNAN ) continue;
        var[j]+=(img[ rowpos+nchans*j +1 ]-mean)*(img[ rowpos+nchans*j +1 ]-mean);
        num_elements[i]++;
      }
    }
  }
  /* otherwise we just need to generate the var of the new column */
  else{
    counter ++;
    if (counter == block_width){
      counter = 0;
    }
    var[counter] = 0;
    num_elements[counter] = 0;
    for(i = 0;i<block_height;i++){
      rowpos = nchans*block_width*i;
      if(img[ rowpos+nchans*(block_width-1)+ 1] == PLMNAN) continue;
      var[counter]+= (img[rowpos+nchans*(block_width-1) + 1]-mean)
          *(img[rowpos+nchans*(block_width-1)+1]-mean);
      num_elements[counter]++;
    }
  }
  total_var =0; 
  for (i=0;i<block_width;i++){
    total_var+=var[i];
    total_num_elements+=num_elements[i];
  }
    
  if( total_num_elements == 1 ) return 0;
  return total_var/(double)(total_num_elements-1);
}

/* calculates a block mean, with a hanning window  (like a blur) */
double calculate_block_mean_hanning( double *block, int blockheight, int blockwidth, int nchans, int channel)
{
  int i,j;
  double mean=0;
  int numelements = 0;
  for ( i=0 ; i<blockheight ; i++ ) {
    for( j=0 ; j<blockwidth ; j++ ) {
      if( block[ nchans*(blockwidth*i+j) +channel] == 128 ) continue;
      mean+= block[ nchans*(blockwidth*i+j) +channel] * hanning( i, blockwidth*8) ;
      numelements++;
    }
  }
  //fprintf(stderr, "Block mean=%g\n", mean/numelements );
  return mean/numelements;
}

/* from est_trans_fft2 */
double hanning(int n, int N) {
    /*printf("n = %d, N = %d, window = %g\n",n, N,(0.5 - 0.5*cos(2*M_PI*(((double)n)/((double)N)))));*/
    //return (0.5 - 0.5*cos(2*M_PI*(((double)n)/((double)N))));
    return (cos(2*M_PI*(((double)n)/((double)N))));
    printf("Hanning also scnanning\n");
}

double calculate_image_mean( double *img, struct image_params *a_params, int nchans)
{
  int i,j;
  double mean = 0;
  double nelements =0;
  for( i=0; i<a_params->height ; i+=1 ) { /* rows */
    for( j=0; j<a_params->width ; j+=1 ) { /*cols */
      if( img[ nchans*(i*a_params->width + j) +1 ] == PLMNAN ) continue;
      mean+=img[ nchans*(i*a_params->width + j) +1 ];   //green
      nelements++;
    }
  }
  return mean/(double)nelements;
}

double calculate_image_var( double *img, struct image_params *a_params, int nchans, double mean)
{
  int i,j,rowpos;
  double var = 0;
  double nelements =0;
  for( i=0; i<a_params->height ; i+=1 ) { /* rows */
    rowpos = nchans*(i*a_params->width);
    for( j=0; j<a_params->width ; j+=1 ) { /*cols */
      //if( img[ nchans*(i*a_params->width +j) +1 ] == PLMNAN ) continue;
      if( img[ rowpos+nchans*j +1 ] == PLMNAN ) continue;
      //var+=(img[ nchans*(i*a_params->width + j) +1 ]-mean)*(img[ nchans*(i*a_params->width + j) +1 ]-mean);
      var+=(img[ rowpos+nchans*j +1 ]-mean)*(img[ rowpos+nchans*j +1 ]-mean);
      nelements++;
    }
  }
  return var/(double)(nelements-1);
}

double shift_block_to_mean( double *block, double mean_g, 
                            double block_mean, int blockwidth, 
                            int blockheight, int nchans)
{
  int i,j,z;
  // if block_mean < sqrt(\eps???) test for 0*Inf condition...
  double shift_factor = mean_g/block_mean;
  //shift_factor = ow(shift_factor,2); //squaring essentially inverts image
  shift_factor = pow(shift_factor,1/(mean_shift));
  for( i=0 ; i<blockheight; i++ ) {
    for( j=0 ; j<blockwidth; j++ ) {
      for( z=0 ; z<nchans; z++ ) {
        if( block[ nchans*(blockwidth*i+j)+z] == PLMNAN ) continue;
        block[ nchans*(blockwidth*i+j)+z]*=shift_factor;
      }
    }
  }
  return shift_factor;
}

/*calculate the dilation factor based on global var_gamma and gbl max_dilation*/
inline double calc_dilation_factor( double var, double max_var)
{
  if( var > max_var/var_gamma ) return 1;  // 100000 
  				//looks really really close
  return (log((max_var)/var)-log(max_var/(max_var/var_gamma))+1)*(max_dilation);
}

double dilate_block_to_var( double *block, double var_g, 
                            double block_mean, int blockwidth, 
                            int blockheight, int nchans, double block_var,
                            double max_var,int scaling)
{
  int i,j,z,rowpos;
  // if block_mean < sqrt(\eps???) test for 0*Inf condition...
  double dilation_factor = var_g/block_var;
  double new_block_mean;
  dilation_factor = calc_dilation_factor( block_var, max_var ); 
  dilation_factor = (dilation_factor -1) *1/(1+scaling)+1;
   
  new_block_mean = calculate_block_mean(block,blockheight,blockwidth, nchans, 1,1);
  //new_block_mean = block_mean;
  for( i=0 ; i<blockheight; i++ ) {
    rowpos = nchans*blockwidth*i;
    for( j=0 ; j<blockwidth; j++ ) {
      for( z=0 ; z<nchans; z++ ) {
        if( block[ rowpos + nchans*j+z] == PLMNAN ) continue;
        block[rowpos+nchans*j+z] = (block[rowpos+nchans*j+z] -
                             new_block_mean)*(dilation_factor)+ new_block_mean;
      }
    }
  }
  return dilation_factor;
}


int write_status(int val, int max_val){
   double result;
   FILE * counter_file;
   int check;

   counter_file = fopen("/tmp/.counter","w+");
   result = ((double)val/(double)max_val)*100;
   fprintf(counter_file,"%2.0f\n",result);
   check =fclose(counter_file);
   return check;
} 

/* returns max and min block variances found for whole image */
/* uses green channel only */
int calculate_image_block_var(double *myimg, struct image_params *a_params, 
                              int nchans, int blockwidth, int blockheight,
                              double *max_var, double *min_var)
{
  int i,j;
  double *myblock;
  double block_mean, block_var;
  int use_new_block =1;
  *max_var =0;
  *min_var =1000000; 

  for(i=0; i<a_params->height-blockheight; i++) {
    use_new_block =1;
    fprintf(stderr, "Calculating image block variance line %d/%d\r", i, a_params->height-blockheight);
    for( j=0; j<a_params->width-blockwidth; j++ ) {
      myblock = read_block( myimg, j,i, nchans, blockwidth, blockheight, a_params);
      block_mean = calculate_block_mean( myblock, blockheight, blockwidth, nchans, 1,use_new_block);
      block_var = calculate_block_var( myblock, blockwidth, blockheight, 
                                       nchans, block_mean,use_new_block);
      use_new_block = 0;
      if( block_var > *max_var ) *max_var = block_var;  
      if( block_var < *min_var ) *min_var = block_var;
    }
  } 
  return 0;
}

void initialize_img(double *img, int height, int width, 
                      int nchans, double* values )
{
  int i,j,z,pos;

  for( i=0 ; i<height ; i++ ) {
    pos = nchans*(i*width);
    //write_status(i, height+1);
    for( j=0 ; j<width ; j++ ) {
      for( z=0 ; z<nchans ; z++ ) {
        img[ pos+j*nchans+z ] = values[z];
      }
    }
  }
   //write_status(1, height+1);
}
double compute_gnf(double const *block,int heigth,int width,int nchans,double * largest_spacial_in){
    /* compute nastiness in the x-direction */
    double x_grad=0,y_grad=0;
    double largest_spacial=0,temp1,temp2;
    int i,j = 0;
    for (i = 0;i<heigth;i++){
       for(j = 0;j<width-1;j++){
           if (block[ nchans*(width*i+j)+1] != PLMNAN&&
               block[ nchans*(width*i+j+1)+1] != PLMNAN){ 
           temp1=block[ nchans*(width*i+j)+1];
           temp2= block[ nchans*(width*i+j+1)+1];
           
           x_grad=fabs(temp1-temp2);
           
           if ( x_grad > largest_spacial) largest_spacial = x_grad;
           }
       }
    }
   for (i = 0;i<heigth-1;i++){
       for(j = 0;j<width;j++){
           if (block[ nchans*(width*i+j)+1] == PLMNAN ||
               block[ nchans*(width*(i+1)+j)+1] == PLMNAN){ 
               temp1= block[ nchans*(width*i+j)+1]; 
               temp2=block[ nchans*(width*(i+1)+j)+1];
               y_grad =fabs(temp1-temp2);
           if ( y_grad > largest_spacial) largest_spacial= y_grad;
           }
       }
    }
    return largest_spacial;

}
